/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
/*
 * Copyright (c) 1986 - 1991 by Sun Microsystems, Inc.
 */

/*
 * xdr.h, External Data Representation Serialization Routines.
 *
 */

#ifndef _RPC_XDR_H
#define	_RPC_XDR_H

#pragma ident	"@(#)xdr.h	1.23	94/04/25 SMI"

/*	xdr.h 1.26 88/12/17 SMI		*/

#include <sys/byteorder.h>	/* For all ntoh* and hton*() kind of macros */
#include <sys/types.h>	/* For all ntoh* and hton*() kind of macros */
#ifndef _KERNEL
#include <stdio.h> /* defines FILE *, used in ANSI C function prototypes */
#endif
#ifdef _KERNEL
#include <sys/stream.h>
#endif

#ifdef linux
typedef long long longlong_t;
typedef unsigned long long u_longlong_t;
#endif


#ifdef __cplusplus
extern "C" {
#endif

/*
 * XDR provides a conventional way for converting between C data
 * types and an external bit-string representation.  Library supplied
 * routines provide for the conversion on built-in C data types.  These
 * routines and utility routines defined here are used to help implement
 * a type encode/decode routine for each user-defined type.
 *
 * Each data type provides a single procedure which takes two arguments:
 *
 *	bool_t
 *	xdrproc(xdrs, argresp)
 *		XDR *xdrs;
 *		<type> *argresp;
 *
 * xdrs is an instance of a XDR handle, to which or from which the data
 * type is to be converted.  argresp is a pointer to the structure to be
 * converted.  The XDR handle contains an operation field which indicates
 * which of the operations (ENCODE, DECODE * or FREE) is to be performed.
 *
 * XDR_DECODE may allocate space if the pointer argresp is null.  This
 * data can be freed with the XDR_FREE operation.
 *
 * We write only one procedure per data type to make it easy
 * to keep the encode and decode procedures for a data type consistent.
 * In many cases the same code performs all operations on a user defined type,
 * because all the hard work is done in the component type routines.
 * decode as a series of calls on the nested data types.
 */

/*
 * Xdr operations.  XDR_ENCODE causes the type to be encoded into the
 * stream.  XDR_DECODE causes the type to be extracted from the stream.
 * XDR_FREE can be used to release the space allocated by an XDR_DECODE
 * request.
 */
enum xdr_op {
	XDR_ENCODE = 0,
	XDR_DECODE = 1,
	XDR_FREE = 2
};

/*
 * This is the number of bytes per unit of external data.
 */
#define	BYTES_PER_XDR_UNIT	(4)
#define	RNDUP(x)  ((((x) + BYTES_PER_XDR_UNIT - 1) / BYTES_PER_XDR_UNIT) \
		    * BYTES_PER_XDR_UNIT)

/*
 * The XDR handle.
 * Contains operation which is being applied to the stream,
 * an operations vector for the paticular implementation (e.g. see xdr_mem.c),
 * and two private fields for the use of the particular impelementation.
 */
typedef struct XDR {
	enum xdr_op	x_op;	/* operation; fast additional param */
	struct xdr_ops {
#ifdef __STDC__
		bool_t	(*x_getlong)(struct XDR *, long *);
		/* get a long from underlying stream */
		bool_t	(*x_putlong)(struct XDR *, long *);
		/* put a long to " */
		bool_t	(*x_getbytes)(struct XDR *, caddr_t, int);
		/* get some bytes from " */
		bool_t	(*x_putbytes)(struct XDR *, caddr_t, int);
		/* put some bytes to " */
		u_int	(*x_getpostn)(struct XDR *);
		/* returns bytes off from beginning */
		bool_t  (*x_setpostn)(struct XDR *, u_int);
		/* lets you reposition the stream */
		__u32 *	(*x_inline)(struct XDR *, int);
		/* buf quick ptr to buffered data */
		void	(*x_destroy)(struct XDR *);
		/* free privates of this xdr_stream */
		bool_t	(*x_control)(struct XDR *, int, void *);
#else
		bool_t	(*x_getlong)();	/* get a long from underlying stream */
		bool_t	(*x_putlong)();	/* put a long to " */
		bool_t	(*x_getbytes)(); /* get some bytes from " */
		bool_t	(*x_putbytes)(); /* put some bytes to " */
		u_int	(*x_getpostn)(); /* returns bytes off from beginning */
		bool_t  (*x_setpostn)(); /* lets you reposition the stream */
		__u32 *	(*x_inline)(); /* buf quick ptr to buffered data */
		void	(*x_destroy)();	/* free privates of this xdr_stream */
		bool_t	(*x_control)();
#endif
	} *x_ops;
	caddr_t 	x_public; /* users' data */
	caddr_t		x_private; /* pointer to private data */
	caddr_t 	x_base;	/* private used for position info */
	int		x_handy; /* extra private word */
} XDR;

/*
 * Operations defined on a XDR handle
 *
 * XDR		*xdrs;
 * long		*longp;
 * caddr_t	 addr;
 * u_int	 len;
 * u_int	 pos;
 */
#define	XDR_GETLONG(xdrs, longp)			\
	(*(xdrs)->x_ops->x_getlong)(xdrs, longp)
#define	xdr_getlong(xdrs, longp)			\
	(*(xdrs)->x_ops->x_getlong)(xdrs, longp)

#define	XDR_PUTLONG(xdrs, longp)			\
	(*(xdrs)->x_ops->x_putlong)(xdrs, longp)
#define	xdr_putlong(xdrs, longp)			\
	(*(xdrs)->x_ops->x_putlong)(xdrs, longp)

#define	XDR_GETBYTES(xdrs, addr, len)			\
	(*(xdrs)->x_ops->x_getbytes)(xdrs, addr, len)
#define	xdr_getbytes(xdrs, addr, len)			\
	(*(xdrs)->x_ops->x_getbytes)(xdrs, addr, len)

#define	XDR_PUTBYTES(xdrs, addr, len)			\
	(*(xdrs)->x_ops->x_putbytes)(xdrs, addr, len)
#define	xdr_putbytes(xdrs, addr, len)			\
	(*(xdrs)->x_ops->x_putbytes)(xdrs, addr, len)

#define	XDR_GETPOS(xdrs)				\
	(*(xdrs)->x_ops->x_getpostn)(xdrs)
#define	xdr_getpos(xdrs)				\
	(*(xdrs)->x_ops->x_getpostn)(xdrs)

#define	XDR_SETPOS(xdrs, pos)				\
	(*(xdrs)->x_ops->x_setpostn)(xdrs, pos)
#define	xdr_setpos(xdrs, pos)				\
	(*(xdrs)->x_ops->x_setpostn)(xdrs, pos)

#define	XDR_INLINE(xdrs, len)				\
	(*(xdrs)->x_ops->x_inline)(xdrs, len)
#define	xdr_inline(xdrs, len)				\
	(*(xdrs)->x_ops->x_inline)(xdrs, len)

#define	XDR_DESTROY(xdrs)				\
	(*(xdrs)->x_ops->x_destroy)(xdrs)
#define	xdr_destroy(xdrs) XDR_DESTROY(xdrs)


#define	XDR_CONTROL(xdrs, req, op)				\
	(*(xdrs)->x_ops->x_control)(xdrs, req, op)
#define	xdr_control(xdrs, req, op) XDR_CONTROL(xdrs, req, op)

/*
 * Support struct for discriminated unions.
 * You create an array of xdrdiscrim structures, terminated with
 * a entry with a null procedure pointer.  The xdr_union routine gets
 * the discriminant value and then searches the array of structures
 * for a matching value.  If a match is found the associated xdr routine
 * is called to handle that part of the union.  If there is
 * no match, then a default routine may be called.
 * If there is no match and no default routine it is an error.
 */


/*
 * A xdrproc_t exists for each data type which is to be encoded or decoded.
 *
 * The second argument to the xdrproc_t is a pointer to an opaque pointer.
 * The opaque pointer generally points to a structure of the data type
 * to be decoded.  If this pointer is 0, then the type routines should
 * allocate dynamic storage of the appropriate size and return it.
 * bool_t	(*xdrproc_t)(XDR *, void *);
 */
#ifdef __STDC__
typedef bool_t	(*xdrproc_t)(XDR *, void *); /* For Backward compatibility */
typedef bool_t	(*xdrproc2_t)(XDR *, void *, u_long);
#else
typedef	bool_t (*xdrproc_t)();
#endif

#define	NULL_xdrproc_t ((xdrproc_t)0)
#define	NULL_xdrproc2_t ((xdrproc2_t)0)
struct xdr_discrim {
	int	value;
	xdrproc2_t proc;
};

/*
 * In-line routines for fast encode/decode of primitve data types.
 * Caveat emptor: these use single memory cycles to get the
 * data from the underlying buffer, and will fail to operate
 * properly if the data is not aligned.  The standard way to use these
 * is to say:
 *	if ((buf = XDR_INLINE(xdrs, count)) == NULL)
 *		return (FALSE);
 *	<<< macro calls >>>
 * where ``count'' is the number of bytes of data occupied
 * by the primitive data types.
 *
 * N.B. and frozen for all time: each data type here uses 4 bytes
 * of external representation.
 */
#define	IXDR_GET_LONG(buf)		((long)ntohl(*((__u32 *)buf)++))
#define	IXDR_PUT_LONG(buf, v)		(*((__u32 *)(buf))++ = htonl((u_long)v))

#define	IXDR_GET_BOOL(buf)		((bool_t)IXDR_GET_LONG(buf))
#define	IXDR_GET_ENUM(buf, t)		((t)IXDR_GET_LONG(buf))
#define	IXDR_GET_U_LONG(buf)		((u_long)IXDR_GET_LONG(buf))
#define	IXDR_GET_SHORT(buf)		((short)IXDR_GET_LONG(buf))
#define	IXDR_GET_U_SHORT(buf)		((u_short)IXDR_GET_LONG(buf))

#define	IXDR_PUT_BOOL(buf, v)		IXDR_PUT_LONG((buf), ((long)(v)))
#define	IXDR_PUT_ENUM(buf, v)		IXDR_PUT_LONG((buf), ((long)(v)))
#define	IXDR_PUT_U_LONG(buf, v)		IXDR_PUT_LONG((buf), ((long)(v)))
#define	IXDR_PUT_SHORT(buf, v)		IXDR_PUT_LONG((buf), ((long)(v)))
#define	IXDR_PUT_U_SHORT(buf, v)	IXDR_PUT_LONG((buf), ((long)(v)))

/*
 * These are the "generic" xdr routines.
 */


/*
 * These are the "generic" xdr routines.
 */
#ifdef __STDC__
extern bool_t	xdr_void (void);
extern bool_t	xdr_int (XDR *, int *);
extern bool_t	xdr_u_int (XDR *, u_int *);
extern bool_t	xdr_long (XDR *, long *);
extern bool_t	xdr_u_long (XDR *, u_long *);
extern bool_t	xdr_short (XDR *, short *);
extern bool_t	xdr_u_short (XDR *, u_short *);
extern bool_t	xdr_bool (XDR *, bool_t *);
extern bool_t	xdr_enum (XDR *, enum_t *);
extern bool_t	xdr_array (XDR *, caddr_t *, u_int *, const u_int, const
u_int, const xdrproc_t);
extern bool_t	xdr_bytes (XDR *, char **, u_int *, const u_int);
extern bool_t	xdr_opaque (XDR *, caddr_t, const u_int);
extern bool_t	xdr_string (XDR *, char **, const u_int);
extern bool_t	xdr_union (XDR *, enum_t *, char *,
		const struct xdr_discrim *, const xdrproc2_t);
extern unsigned long	xdr_sizeof (xdrproc_t, void *);

extern bool_t   xdr_hyper (XDR *, longlong_t *);
extern bool_t   xdr_longlong_t (XDR *, longlong_t *);
extern bool_t   xdr_u_hyper (XDR *, u_longlong_t *);
extern bool_t   xdr_u_longlong_t (XDR *, u_longlong_t *);

extern bool_t	xdr_char (XDR *, char *);
extern bool_t	xdr_wrapstring (XDR *, char **);
extern bool_t	xdr_reference (XDR *, caddr_t *, u_int, const xdrproc2_t);
extern bool_t	xdr_pointer (XDR *, char **, u_int, const xdrproc_t);

#ifndef _KERNEL
extern bool_t	xdr_u_char (XDR *, u_char *);
extern bool_t	xdr_vector (XDR *, char *, const u_int, const u_int, const
xdrproc2_t);
extern bool_t	xdr_float (XDR *, float *);
extern bool_t	xdr_double (XDR *, double *);
extern bool_t	xdr_quadruple (XDR *, long double *);
extern void	xdr_free (xdrproc_t, char *);
#endif /* !_KERNEL */
#else
extern bool_t	xdr_void ();
extern bool_t	xdr_int ();
extern bool_t	xdr_u_int ();
extern bool_t	xdr_long ();
extern bool_t	xdr_u_long ();
extern bool_t	xdr_short ();
extern bool_t	xdr_u_short ();
extern bool_t	xdr_bool ();
extern bool_t	xdr_enum ();
extern bool_t	xdr_array ();
extern bool_t	xdr_bytes ();
extern bool_t	xdr_opaque ();
extern bool_t	xdr_string ();
extern bool_t	xdr_union ();

extern bool_t   xdr_hyper ();
extern bool_t   xdr_longlong_t ();
extern bool_t   xdr_u_hyper ();
extern bool_t   xdr_u_longlong_t ();
extern bool_t	xdr_char ();
extern bool_t	xdr_reference ();
extern bool_t	xdr_pointer ();
extern bool_t	xdr_wrapstring ();

#ifndef _KERNEL
extern bool_t	xdr_u_char ();
extern bool_t	xdr_vector ();
extern bool_t	xdr_float ();
extern bool_t	xdr_double ();
extern bool_t   xdr_quadruple();
extern void	xdr_free ();
#endif /* !_KERNEL */
#endif

/*
 * Common opaque bytes objects used by many rpc protocols;
 * declared here due to commonality.
 */
#define	MAX_NETOBJ_SZ 1024
struct netobj {
	u_int	n_len;
	char	*n_bytes;
};
typedef struct netobj netobj;

#ifdef __STDC__
extern bool_t   xdr_netobj(XDR *, netobj *);
#else
extern bool_t   xdr_netobj();
#endif

/*
 * These are XDR control operators
 */

#define	XDR_GET_BYTES_AVAIL 1

struct xdr_bytesrec{
	bool_t xc_is_last_record;
	size_t xc_num_avail;
	};

typedef struct xdr_bytesrec xdr_bytesrec;

#ifdef _KERNEL

#define	XDR_PEEK 2

#endif

/*
 * These are the public routines for the various implementations of
 * xdr streams.
 */
#ifndef _KERNEL
#ifdef __STDC__
extern void   xdrmem_create(XDR *, const caddr_t, const u_int, const enum
xdr_op);
	/* XDR using memory buffers */
extern void   xdrstdio_create(XDR *, FILE *, const enum xdr_op);
/* XDR using stdio library */
extern void   xdrrec_create(XDR *, const u_int, const u_int, const caddr_t,
int (*) (void *, caddr_t, int), int (*) (void *, caddr_t, int));
/* XDR pseudo records for tcp */
extern bool_t xdrrec_endofrecord(XDR *, bool_t);
/* make end of xdr record */
extern bool_t xdrrec_skiprecord(XDR *);
/* move to beginning of next record */
extern bool_t xdrrec_eof(XDR *);
extern u_int xdrrec_readbytes(XDR *, caddr_t, u_int);
/* true if no more input */

#ifdef __RPCLIB_INTERNAL__
int __xdrreq_setfirst(XDR *);
int __xdrreq_resetfirst(XDR *);
int __is_xdrreq_first(XDR *);
struct rpc_msg;
extern bool_t	xdr_callmsg(XDR *, struct rpc_msg *);
#endif
#else
extern void   xdrmem_create ();
extern void   xdrstdio_create ();
extern void   xdrrec_create ();
extern bool_t xdrrec_endofrecord ();
extern bool_t xdrrec_skiprecord ();
extern bool_t xdrrec_eof ();
extern u_int xdrrec_readbytes();
#endif
#else

extern void	xdrmem_create(XDR *, caddr_t, u_int, enum xdr_op);
extern void	xdrmblk_init(XDR *, mblk_t *, enum xdr_op, int);
extern bool_t	xdrmblk_getmblk(XDR *, mblk_t **, u_int *);
extern bool_t	xdrmblk_putmblk(XDR *, mblk_t *, u_int);

struct rpc_msg;
extern bool_t	xdr_callmsg(XDR *, struct rpc_msg *);
extern bool_t	xdr_replymsg_body(XDR *, struct rpc_msg *);
extern bool_t	xdr_replymsg_hdr(XDR *, struct rpc_msg *);

extern bool_t	xdrmblk_setpos(XDR *, u_int);

extern struct xdr_ops xdrmblk_ops;
#endif /* !_KERNEL */

#ifdef __cplusplus
}
#endif

#endif	/* !_RPC_XDR_H */
