/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#pragma ident "@(#)keylogin.c	1.7	92/04/21 SMI"

/*
 *	Copyright (c) 1988-1992 Sun Microsystems Inc
 *	All Rights Reserved.
 */

#pragma ident	"@(#)keylogin.c	1.10	93/07/05 SMI"

/*
 * Set secret key on local machine
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <string.h>

#include <rpc/keylib.h>


#define ROOTKEY_FILE "/etc/.rootkey"


static void
usage(char *cmd)
{
	fprintf(stderr, "usage: %s [-r]\n", cmd );
	exit(1);
}

int
main(int argc, char *argv[])
{
	char fullname[MAXNETNAMELEN + 1];
	struct key_netstarg_long netst;
	int mkrootkey, fd;

	if( argc == 1 )
		mkrootkey = 0;
	else if( argc == 2 && (strcmp( argv[1], "-r" ) == 0) ) {
		if (geteuid() != 0) {
			fprintf(stderr, "Must be root to use -r option.\n");
			exit(1);
		}
		mkrootkey = 1;
	} else
		usage( argv[0] );
		
	if( !getnetname(fullname) ) {
		fprintf(stderr, "Could not generate netname\n");
		exit(1);
	}

	if (!getsecretkey_long(fullname, &(netst.st_priv_key),
			   getpass("Password:"))){
		fprintf(stderr, "Could not find %s's secret key\n", fullname);
		exit(1);
	}

	if (netst.st_priv_key == NULL) {
		fprintf(stderr, "Password incorrect for %s\n", fullname);
		exit(1);
	}

	netst.st_pub_key = NULL;
	netst.st_netname = strdup(fullname);

	/* do actual key login */
	if (key_setnet_long(&netst) < 0 ) {
		fprintf(stderr, "Could not set %s's secret key\n", fullname);
		fprintf(stderr, "May be the keyserv is down?\n");
		if( mkrootkey == 0 )   /* nothing else to do */
			exit(1);
	}

	/* write unencrypted secret key into root key file */
	if( mkrootkey == 1 ) {
		unlink( ROOTKEY_FILE );
		if( (fd = open(ROOTKEY_FILE, O_WRONLY+O_CREAT, 0600)) != -1 ) {
			char * lf = "\n";
			write(fd, netst.st_priv_key, 
			      strlen(netst.st_priv_key));
			write(fd, lf, 1); 
			close(fd);
			fprintf(stderr, "Wrote secret key into %s\n",
				ROOTKEY_FILE);
		} else {
			fprintf(stderr, "Could not open %s for update\n",
				ROOTKEY_FILE );
			exit(1);
		}
	}

	exit(0);
}
