/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#pragma ident	"@(#)keyserv.c	1.15	94/04/25 SMI"

/*
 * Copyright (c) 1986 - 1991 by Sun Microsystems, Inc.
 */

/*
 * Keyserver
 * Store secret keys per uid. Do public key encryption and decryption
 * operations. Generate "random" keys.
 * Do not talk to anything but a local root
 * process on the local transport only
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <rpc/rpc.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <time.h>
#include <pwd.h>
#include <rpc/des_crypt.h>
#include <rpc/key_prot.h>
#include <rpc/svc_soc.h>
#ifdef linux
#include <netinet/in.h>
#include <sys/un.h>
#include <syslog.h>
#endif
#include <errno.h>


#ifdef KEYSERV_RANDOM
extern long random();
#endif
#ifndef NGROUPS
#define	NGROUPS 16
#endif

#include "keyserv.h"
#include <rpc/keylib.h>

static void randomize(des_block *master);
static void usage(void);
static int getrootkey(des_block *master, int prompt, int hexkeybytes);
static int root_auth(SVCXPRT *trans, struct svc_req *rqstp);

#ifdef DEBUG
static int debugging = 1;
#else
static int debugging = 0;
#endif

unsigned server_keybytes;
unsigned server_keybits;
unsigned server_hexkeybytes;

static int convkey_7bit;
static int sunos4_compat;

static void keyprogram(struct svc_req *rqstp, SVCXPRT *transp);
static des_block masterkey;
static char ROOTKEY[] = "/etc/.rootkey2";

#ifdef linux
#include <linux/unistd.h>

int setkeyserv(int arg)
{
  return syscall(__NR_init_module, arg);
}

void flush_kernel_uid(int uid)
{
	setkeyserv(-uid-1);
}

int make_kernel_connection(SVCXPRT *xprt)
{
	int fd;
	int res;
	struct sockaddr_un sun;

	if (sunos4_compat)
		return 0;

	fd = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (fd < 0)
		return -1;
	sun.sun_family = AF_UNIX;
	sun.sun_path[0] = 0;
	res = bind(fd, (struct sockaddr*)&sun, sizeof(short));
	if (res < 0) {
		syslog(LOG_ERR, "bind: %m");
		return -1;
	}
	memcpy(sun.sun_path+1, xprt->xp_ltaddr.buf, xprt->xp_ltaddr.len);
	res = connect(fd, (struct sockaddr*)&sun, xprt->xp_ltaddr.len+1+sizeof(short));
	if (res < 0)
	{
		syslog(LOG_ERR, "connect: %m");
		return -1;
	}
	res = setkeyserv(fd);
	close(fd);
	if (res < 0)
	{
		syslog(LOG_ERR, "setkeyserv: %m");
		return -1;
	}
}
#endif


/*
 * Hack to allow the keyserver to use AUTH_DES (for authenticated
 * NIS+ calls, for example).  The only functions that get called
 * are key_encryptsession_pk, key_decryptsession_pk, and key_gendes.
 *
 * The approach is to have the keyserver fill in pointers to local
 * implementations of these functions, and to call those in key_call().
 */

extern cryptkeyres *(*__key_encryptsession_pk_LOCAL)(uid_t, cryptkeyarg2*);
extern cryptkeyres *(*__key_decryptsession_pk_LOCAL)(uid_t, cryptkeyarg2*);
extern des_block *(*__key_gendes_LOCAL)(void *v, struct svc_req *s);

static cryptkeyres * key_encrypt_pk_2_svc_uid(uid_t uid, cryptkeyarg2 *arg);
static cryptkeyres * key_decrypt_pk_2_svc_uid(uid_t uid, cryptkeyarg2 *arg);

static void keyprogram_old(struct svc_req * rqstp, SVCXPRT * transp);


int
main(int argc, char *argv[])
{
	int nflag = 0;
	extern int optind;
	int c;

	__key_encryptsession_pk_LOCAL = &key_encrypt_pk_2_svc_uid;
	__key_decryptsession_pk_LOCAL = &key_decrypt_pk_2_svc_uid;
	__key_gendes_LOCAL = &key_gen_1_svc;

	while ((c = getopt(argc, argv, "4ndDkc:")) != -1)
		switch (c) {
		case 'n':
			nflag++;
			break;
		case '4':
			sunos4_compat++;
			break;
		case 'd':
			pk_nodefaultkeys();
			break;
		case 'k':
			pk_nodiskkeystore();
			break;
		case 'c':
			convkey_7bit = 1;
			break;
		case 'D':
			debugging = 1;
			break;
		default:
			usage();
		}

	if (optind != argc) {
		usage();
	}

	/*
	 * Initialize
	 */
	(void) umask(066);	/* paranoia */
	if (geteuid() != 0) {
		(void) fprintf(stderr, "%s must be run as root\n", argv[0]);
		exit(1);
	}

	setmodulus(DEFAULT_HEXMODULUS);

	openstore(server_keybytes);
	getrootkey(&masterkey, nflag, server_hexkeybytes);
	readkeys(server_keybytes);

	if (svc_create_local_service(keyprogram, KEY_PROG, KEY_VERS,
		"netpath", "keyserv") == 0) {
		(void) fprintf(stderr,
			"%s: unable to create service\n", argv[0]);
		exit(1);
	}

	if (svc_create_local_service(keyprogram, KEY_PROG, KEY_VERS2,
		"netpath", "keyserv") == 0) {
		(void) fprintf(stderr,
			"%s: unable to create service\n", argv[0]);
		exit(1);
	}
	if (sunos4_compat) {
		SVCXPRT * transp;

#ifdef linux
{
	int sock;
	int csock;
	struct sockaddr_in caddr;
	struct sockaddr_in addr;
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("socket");
		exit(1);
	}
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = 0 /*htonl(INADDR_LOOPBACK)*/;
	addr.sin_port = htons(1000);

again:
	if (bind(sock, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
		if (errno == EADDRINUSE && (ntohs(addr.sin_port) > 1)) {
			addr.sin_port = htons(ntohs(addr.sin_port) - 1);
			goto again;
		}
		perror("bind");
		exit(1);
	}
	
	csock = socket(AF_INET, SOCK_DGRAM, 0);
	if (csock < 0) {
		perror("socket");
		exit(1);
	}
	caddr.sin_family = AF_INET;
	caddr.sin_addr.s_addr = 0;
	caddr.sin_port = htons(999);

again1:
	if (bind(csock, (struct sockaddr*)&caddr, sizeof(caddr)) < 0) {
		if (errno == EADDRINUSE && (ntohs(caddr.sin_port) > 1)) {
			caddr.sin_port = htons(ntohs(caddr.sin_port) - 1);
			goto again1;
		}
		perror("cbind");
		exit(1);
	}
	if (connect(csock, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
		perror("cconnect");
		exit(1);
	}
	if (setkeyserv(csock))
		perror("sys_setkeyserv");
	close(csock);
	if (t_sync(sock) < 0) {
		t_error("t_sync");
		exit(1);
	}
	transp = svcudp_create(sock);
	if (!transp)
		exit(1);
}
#else
	transp = svcudp_create(RPC_ANYSOCK);
#endif
	if (transp) {
	  pmap_unset(KEY_PROG, KEY_VERS);
			if (svc_register(transp, KEY_PROG, KEY_VERS, keyprogram_old,
					 IPPROTO_UDP) == FALSE) {
				exit(2);
			}
	}



	}
	if (!debugging) {
		detachfromtty();
	}
	svc_run();
	abort();
	/* NOTREACHED */
}

/*
 * In the event that we don't get a root password, we try to
 * randomize the master key the best we can
 */
static void
randomize(des_block *master)
{
	int i;
	int seed;
	struct timeval tv;
	int shift;

	seed = 0;
	for (i = 0; i < 1024; i++) {
		(void) gettimeofday(&tv, (struct timezone *) NULL);
		shift = i % 8 * sizeof (int);
		seed ^= (tv.tv_usec << shift) | (tv.tv_usec >> (32 - shift));
	}
#ifdef KEYSERV_RANDOM
	srandom(seed);
	master->key.low = random();
	master->key.high = random();
	srandom(seed);
#else
	/* use stupid dangerous bad rand() */
	srand(seed);
	master->key.low = rand();
	master->key.high = rand();
	srand(seed);
#endif
}

/*
 * Try to get root's secret key, by prompting if terminal is a tty, else trying
 * from standard input.
 * Returns 1 on success.
 */
static int
getrootkey(des_block *master, int prompt, int hexkeybytes)
{
	char *passwd;
	char name[MAXNETNAMELEN + 1];
	char *secret;
	key_netstarg_long netstore;
	int fd;

	if (!prompt) {
		/*
		 * Read secret key out of ROOTKEY
		 */
		fd = open(ROOTKEY, O_RDONLY, 0);
		if (fd < 0) {
			randomize(master);
			return (0);
		}
		secret = (char*)malloc(MAX_HEXKEYBYTES+1);
		if (!secret) {
			randomize(master);
			close(fd);
			return (0);
		}
		if (read(fd, secret, MAX_HEXKEYBYTES) < 0) {
			randomize(master);
			free(secret);
			(void) close(fd);
			return (0);
		}
		(void) close(fd);
		if (secret[hexkeybytes] != '\n') {
			(void) fprintf(stderr, "Invalid %s\n", ROOTKEY);
			randomize(master);
			free(secret);
			return (0);
		}
		secret[hexkeybytes] = 0;
		if (!getnetname(name)) {
		    (void) fprintf(stderr, "keyserv: \
failed to generate host's netname when establishing root's key.\n");
		    free(secret);
		    return (0);
		}
		netstore.st_priv_key = secret;
		netstore.st_pub_key = NULL;
		netstore.st_netname = name;
		if (pk_netput_long(0, &netstore) != KEY_SUCCESS) {
			free(secret);
			(void) fprintf(stderr,
				       "keyserv: could not set root's key and netname.\n");
			return (0);
		}
		free(secret);
		return (1);
	}
	/*
	 * Decrypt yellow pages publickey entry to get secret key
	 */
	getnetname(name);
	passwd = getpass("root password:");
#ifndef SECURE_PUBLICKEY
	xpasswd2des(passwd, (char*)master, server_keybytes == DEFAULT_KEYBYTES);
#else
	xxpasswd2des(passwd, name, (char*)master, server_keybytes == DEFAULT_KEYBYTES);
#endif
	if (!getsecretkey_long(name, &secret, passwd)) {
		(void) fprintf(stderr,
		"Can't find %s's secret key\n", name);
		return (0);
	}
	if (!secret) {
		(void) fprintf(stderr,
	"Password does not decrypt secret key for %s\n", name);
		return (0);
	}
	(void) pk_setkey_long(0, secret);
	/*
	 * Store it for future use in $ROOTKEY, if possible
	 */
	fd = open(ROOTKEY, O_WRONLY|O_TRUNC|O_CREAT, 0);
	if (fd > 0) {
		char newline = '\n';

		write(fd, secret, strlen(secret));
		write(fd, &newline, sizeof (newline));
		close(fd);
	}
	free(secret);
	return (1);
}

void
writerootkey(keybuf_long secret)
{
	char newline = '\n';
	int fd;

	fd = open(ROOTKEY, O_WRONLY|O_TRUNC|O_CREAT, 0);
	if (fd < 0) {
		perror(ROOTKEY);
	} else {
		if (write(fd, secret, strlen(secret)) < 0 ||
		    write(fd, &newline, sizeof(newline)) < 0) {
			(void)fprintf(stderr, "%s: ", ROOTKEY);
			perror("write");
		}
		close(fd);
	}
}


/*
 * Procedures to implement RPC service
 */
char *
strstatus(keystatus status)
{
	switch (status) {
	case KEY_SUCCESS:
		return ("KEY_SUCCESS");
	case KEY_NOSECRET:
		return ("KEY_NOSECRET");
	case KEY_UNKNOWN:
		return ("KEY_UNKNOWN");
	case KEY_SYSTEMERR:
		return ("KEY_SYSTEMERR");
	default:
		return ("(bad result code)");
	}
}

keystatus *
key_set_1_svc_uid(uid_t uid, keybuf key)
{
	static keystatus status;
	char keybf[DEFAULT_HEXKEYBYTES+1];

	if (server_hexkeybytes != DEFAULT_HEXKEYBYTES) {
		status = KEY_SYSTEMERR;
		return &status;
	}

	memcpy(keybf, key, DEFAULT_HEXKEYBYTES);
	keybf[DEFAULT_HEXKEYBYTES] = 0;

	if (debugging) {
		(void) fprintf(stderr, "set(%d, %s) = ", uid, keybf);
	}
	status = pk_setkey_long(uid, (keybuf_long)keybf);
	if (debugging) {
		(void) fprintf(stderr, "%s\n", strstatus(status));
		(void) fflush(stderr);
	}
	return (&status);
}

keystatus *
key_setlong_1_svc_uid(uid_t uid, keybuf_long *key)
{
	static keystatus status;

	if (debugging) {
		(void) fprintf(stderr, "setlong(%d, %s) = ", uid, *key ? : "null");
	}
	status = pk_setkey_long(uid, *key);
	if (debugging) {
		(void) fprintf(stderr, "%s\n", strstatus(status));
		(void) fflush(stderr);
	}
	return (&status);
}


cryptkeyres *
key_encrypt_pk_2_svc_uid(uid_t uid, cryptkeyarg2 *arg)
{
	static cryptkeyres res;

	if (debugging) {
		(void) fprintf(stderr, "encrypt(%d, %s, %08x%08x) = ", uid,
				arg->remotename, arg->deskey.key.high,
				arg->deskey.key.low);
	}
	res.cryptkeyres_u.deskey = arg->deskey;
	res.status = pk_encrypt(uid, arg->remotename, &(arg->remotekey),
				&res.cryptkeyres_u.deskey);
	if (debugging) {
		if (res.status == KEY_SUCCESS) {
			(void) fprintf(stderr, "%08x%08x\n",
					res.cryptkeyres_u.deskey.key.high,
					res.cryptkeyres_u.deskey.key.low);
		} else {
			(void) fprintf(stderr, "%s\n", strstatus(res.status));
		}
		(void) fflush(stderr);
	}
	return (&res);
}

cryptkeyres *
key_decrypt_pk_2_svc_uid(uid_t uid, cryptkeyarg2 *arg)
{
	static cryptkeyres res;

	if (debugging) {
		(void) fprintf(stderr, "decrypt(%d, %s, %08x%08x) = ", uid,
				arg->remotename, arg->deskey.key.high,
				arg->deskey.key.low);
	}
	res.cryptkeyres_u.deskey = arg->deskey;
	res.status = pk_decrypt(uid, arg->remotename, &(arg->remotekey),
				&res.cryptkeyres_u.deskey);
	if (debugging) {
		if (res.status == KEY_SUCCESS) {
			(void) fprintf(stderr, "%08x%08x\n",
					res.cryptkeyres_u.deskey.key.high,
					res.cryptkeyres_u.deskey.key.low);
		} else {
			(void) fprintf(stderr, "%s\n", strstatus(res.status));
		}
		(void) fflush(stderr);
	}
	return (&res);
}

keystatus *
key_net_put_2_svc_uid(uid_t uid, key_netstarg *arg)
{
	static keystatus status;
	char   pkeybf[DEFAULT_HEXKEYBYTES+1];
	char   skeybf[DEFAULT_HEXKEYBYTES+1];
	key_netstarg_long larg;

	if (debugging) {
		(void) fprintf(stderr, "net_put(%s, %.*s, %.*s) = ",
			arg->st_netname, sizeof (arg->st_pub_key),
			arg->st_pub_key, sizeof (arg->st_priv_key),
			arg->st_priv_key);
	};

	if (server_hexkeybytes != DEFAULT_HEXKEYBYTES) {
		status = KEY_SYSTEMERR;
		return &status;
	}

	memset(&larg, 0, sizeof(larg));
	memcpy(skeybf, arg->st_priv_key, DEFAULT_HEXKEYBYTES);
	skeybf[DEFAULT_HEXKEYBYTES] = 0;
	memcpy(pkeybf, arg->st_pub_key, DEFAULT_HEXKEYBYTES);
	pkeybf[DEFAULT_HEXKEYBYTES] = 0;
	if (arg->st_pub_key[0])
		larg.st_pub_key = pkeybf;
	if (arg->st_priv_key[0])
		larg.st_priv_key = skeybf;

	status = pk_netput_long(uid, &larg);

	if (debugging) {
		(void) fprintf(stderr, "%s\n", strstatus(status));
		(void) fflush(stderr);
	}

	return (&status);
}

keystatus *
key_net_putlong_2_svc_uid(uid_t uid, key_netstarg_long *arg)
{
	static keystatus status;

	if (debugging) {
		(void) fprintf(stderr, "net_putlong(%s, %s, %s) = ",
			arg->st_netname ? : "null",
			arg->st_pub_key ? : "null",
			arg->st_priv_key ? : "null");
	};

	status = pk_netput_long(uid, arg);

	if (debugging) {
		(void) fprintf(stderr, "%s\n", strstatus(status));
		(void) fflush(stderr);
	}

	return (&status);
}


key_netstres *
key_net_get_2_svc_uid(uid_t uid, void *arg)
{
	static key_netstres keynetname;
	key_netstres_long keynetname_long;

	if (debugging)
		(void) fprintf(stderr, "net_get(%d) = ", uid);

	memset(&keynetname_long, 0, sizeof(keynetname_long));
	memset(&keynetname, 0, sizeof(keynetname));

	keynetname.status = pk_netget_long(uid, &keynetname_long.key_netstres_long_u.knet);

	if (keynetname.status == KEY_SUCCESS) {
		if (keynetname_long.key_netstres_long_u.knet.st_pub_key)
			strncpy(keynetname.key_netstres_u.knet.st_pub_key,
				keynetname_long.key_netstres_long_u.knet.st_pub_key,
				DEFAULT_HEXKEYBYTES);
		if (keynetname_long.key_netstres_long_u.knet.st_priv_key)
			strncpy(keynetname.key_netstres_u.knet.st_priv_key,
				keynetname_long.key_netstres_long_u.knet.st_priv_key,
				DEFAULT_HEXKEYBYTES);
		keynetname.key_netstres_u.knet.st_netname =
			keynetname_long.key_netstres_long_u.knet.st_netname;
	}

	if (debugging) {
		if (keynetname.status == KEY_SUCCESS) {
			fprintf(stderr, "<%s, %.*s, %.*s>\n",
			keynetname.key_netstres_u.knet.st_netname  ? : "null",
			sizeof (keynetname.key_netstres_u.knet.st_pub_key),
			keynetname.key_netstres_u.knet.st_pub_key,
			sizeof (keynetname.key_netstres_u.knet.st_priv_key),
			keynetname.key_netstres_u.knet.st_priv_key);
		} else {
			(void) fprintf(stderr, "NOT FOUND\n");
		}
		(void) fflush(stderr);
	}

	return (&keynetname);
}

key_netstres_long *
key_net_getlong_2_svc_uid(uid_t uid, void *arg)
{
	static key_netstres_long keynetname;

	if (debugging)
		(void) fprintf(stderr, "net_getlong(%d) = ", uid);

	memset(&keynetname, 0, sizeof(keynetname));

	keynetname.status = pk_netget_long(uid, &keynetname.key_netstres_long_u.knet);
	if (debugging) {
		if (keynetname.status == KEY_SUCCESS) {
			fprintf(stderr, "<%s, %s, %s>\n",
			keynetname.key_netstres_long_u.knet.st_netname ? : "null",
			keynetname.key_netstres_long_u.knet.st_pub_key ? : "null",
			keynetname.key_netstres_long_u.knet.st_priv_key ? : "null");
		} else {
			(void) fprintf(stderr, "NOT FOUND\n");
		}
		(void) fflush(stderr);
	}

	return (&keynetname);
}


cryptkeyres *
key_get_conv_2_svc_uid(uid_t uid, keybuf arg)
{
	static cryptkeyres  res;
	char keybf[DEFAULT_HEXKEYBYTES+1];


	if (debugging)
		(void) fprintf(stderr, "get_conv(%d, %.*s) = ", uid,
			sizeof (arg), arg);

	if (server_hexkeybytes != DEFAULT_HEXKEYBYTES) {
		res.status = KEY_SYSTEMERR;
		return &res;
	}

	memcpy(keybf, arg, DEFAULT_HEXKEYBYTES);
	keybf[DEFAULT_HEXKEYBYTES] = 0;

	res.status = pk_get_convlong_key(uid, (keybuf_long)keybf, &res);

	if (debugging) {
		if (res.status == KEY_SUCCESS) {
			(void) fprintf(stderr, "%08x%08x\n",
				res.cryptkeyres_u.deskey.key.high,
				res.cryptkeyres_u.deskey.key.low);
		} else {
			(void) fprintf(stderr, "%s\n", strstatus(res.status));
		}
		(void) fflush(stderr);
	}
	return (&res);
}

cryptkeyres *
key_get_convlong_2_svc_uid(uid_t uid, keybuf_long *arg)
{
	static cryptkeyres  res;

	if (debugging)
		(void) fprintf(stderr, "get_convlong(%d, %s) = ", uid, *arg ? : "null");

	res.status = pk_get_convlong_key(uid, *arg, &res);

	if (debugging) {
		if (res.status == KEY_SUCCESS) {
			(void) fprintf(stderr, "%08x%08x\n",
				res.cryptkeyres_u.deskey.key.high,
				res.cryptkeyres_u.deskey.key.low);
		} else {
			(void) fprintf(stderr, "%s\n", strstatus(res.status));
		}
		(void) fflush(stderr);
	}
	return (&res);
}



cryptkeyres *
key_encrypt_1_svc_uid(uid_t uid, cryptkeyarg *arg)
{
	static cryptkeyres res;

	if (debugging) {
		(void) fprintf(stderr, "encrypt(%d, %s, %08x%08x) = ", uid,
				arg->remotename, arg->deskey.key.high,
				arg->deskey.key.low);
	}
	res.cryptkeyres_u.deskey = arg->deskey;
	res.status = pk_encrypt(uid, arg->remotename, NULL,
				&res.cryptkeyres_u.deskey);
	if (debugging) {
		if (res.status == KEY_SUCCESS) {
			(void) fprintf(stderr, "%08x%08x\n",
					res.cryptkeyres_u.deskey.key.high,
					res.cryptkeyres_u.deskey.key.low);
		} else {
			(void) fprintf(stderr, "%s\n", strstatus(res.status));
		}
		(void) fflush(stderr);
	}
	return (&res);
}

cryptkeyres *
key_decrypt_1_svc_uid(uid_t uid, cryptkeyarg *arg)
{
	static cryptkeyres res;

	if (debugging) {
		(void) fprintf(stderr, "decrypt(%d, %s, %08x%08x) = ", uid,
				arg->remotename, arg->deskey.key.high,
				arg->deskey.key.low);
	}
	res.cryptkeyres_u.deskey = arg->deskey;
	res.status = pk_decrypt(uid, arg->remotename, NULL,
				&res.cryptkeyres_u.deskey);
	if (debugging) {
		if (res.status == KEY_SUCCESS) {
			(void) fprintf(stderr, "%08x%08x\n",
					res.cryptkeyres_u.deskey.key.high,
					res.cryptkeyres_u.deskey.key.low);
		} else {
			(void) fprintf(stderr, "%s\n", strstatus(res.status));
		}
		(void) fflush(stderr);
	}
	return (&res);
}

/* ARGSUSED */
des_block *
key_gen_1_svc(void *v, struct svc_req *s)
{
	struct timeval time;
	static des_block keygen;
	static des_block key;

	(void) gettimeofday(&time, (struct timezone *) NULL);
	keygen.key.high += (time.tv_sec ^ time.tv_usec);
	keygen.key.low += (time.tv_sec ^ time.tv_usec);
	ecb_crypt((char *)&masterkey, (char *)&keygen, sizeof (keygen),
		DES_ENCRYPT | DES_HW);
	key = keygen;
	sesskey_des_setparity((char *)&key);
	if (debugging) {
		(void) fprintf(stderr, "gen() = %08x%08x\n", key.key.high,
					key.key.low);
		(void) fflush(stderr);
	}
	return (&key);
}

getcredres *
key_getcred_1_svc_uid(uid_t uid, netnamestr *name)
{
	static getcredres res;
	static u_int gids[NGROUPS];
	struct unixcred *cred;

	cred = &res.getcredres_u.cred;
	cred->gids.gids_val = gids;
	if (!netname2user(*name, (uid_t *) &cred->uid, (gid_t *) &cred->gid,
			(int *)&cred->gids.gids_len, (gid_t *)gids)) {
		res.status = KEY_UNKNOWN;
	} else {
		res.status = KEY_SUCCESS;
	}
	if (debugging) {
		(void) fprintf(stderr, "getcred(%s) = ", *name);
		if (res.status == KEY_SUCCESS) {
			(void) fprintf(stderr, "uid=%d, gid=%d, grouplen=%d\n",
				cred->uid, cred->gid, cred->gids.gids_len);
		} else {
			(void) fprintf(stderr, "%s\n", strstatus(res.status));
		}
		(void) fflush(stderr);
	}
	return (&res);
}

/*
 * RPC boilerplate
 */
static void
keyprogram(struct svc_req *rqstp, SVCXPRT *transp)
{
	union {
		keybuf key_set_1_arg;
		keybuf_long key_setlong_1_arg;
		cryptkeyarg key_encrypt_1_arg;
		cryptkeyarg key_decrypt_1_arg;
		netnamestr key_getcred_1_arg;

		cryptkeyarg key_encrypt_2_arg;
		cryptkeyarg key_decrypt_2_arg;
		netnamestr key_getcred_2_arg;
		cryptkeyarg2 key_encrypt_pk_2_arg;
		cryptkeyarg2 key_decrypt_pk_2_arg;
		key_netstarg key_net_put_2_arg;
		key_netstarg_long key_net_putlong_2_arg;
		keybuf  key_get_conv_2_arg;
		keybuf_long  key_get_convlong_2_arg;
	} argument;
	char *result;
	xdrproc_t xdr_argument, xdr_result;
	char *(*local) (uid_t, caddr_t);
	uid_t uid;
	int check_auth;

	switch (rqstp->rq_proc) {
	case NULLPROC:
		svc_sendreply(transp, (xdrproc_t)xdr_void, (char *)NULL);
		return;

	case KEY_SET:
		xdr_argument = (xdrproc_t)xdr_keybuf;
		xdr_result = (xdrproc_t)xdr_int;
		local = (char *(*)(uid_t, caddr_t)) key_set_1_svc_uid;
		check_auth = 1;
		break;
	case KEY_SETLONG:
		xdr_argument = (xdrproc_t)xdr_keybuf_long;
		xdr_result = (xdrproc_t)xdr_int;
		local = (char *(*)(uid_t, caddr_t)) key_setlong_1_svc_uid;
		check_auth = 1;
		break;

	case KEY_ENCRYPT:
		xdr_argument = (xdrproc_t)xdr_cryptkeyarg;
		xdr_result = (xdrproc_t)xdr_cryptkeyres;
		local = (char *(*)(uid_t, caddr_t)) key_encrypt_1_svc_uid;
		check_auth = 1;
		break;

	case KEY_DECRYPT:
		xdr_argument = (xdrproc_t)xdr_cryptkeyarg;
		xdr_result = (xdrproc_t)xdr_cryptkeyres;
		local = (char *(*)(uid_t, caddr_t)) key_decrypt_1_svc_uid;
		check_auth = 1;
		break;

	case KEY_GEN:
		xdr_argument = (xdrproc_t)xdr_void;
		xdr_result = (xdrproc_t)xdr_des_block;
		local = (char *(*)(uid_t, caddr_t)) key_gen_1_svc;
		check_auth = 0;
		break;

	case KEY_GETCRED:
		xdr_argument = (xdrproc_t)xdr_netnamestr;
		xdr_result = (xdrproc_t)xdr_getcredres;
		local = (char *(*)(uid_t, caddr_t)) key_getcred_1_svc_uid;
		check_auth = 0;
		break;

	case KEY_ENCRYPT_PK:
		xdr_argument = (xdrproc_t)xdr_cryptkeyarg2;
		xdr_result = (xdrproc_t)xdr_cryptkeyres;
		local = (char *(*)(uid_t, caddr_t)) key_encrypt_pk_2_svc_uid;
		check_auth = 1;
		break;

	case KEY_DECRYPT_PK:
		xdr_argument = (xdrproc_t)xdr_cryptkeyarg2;
		xdr_result = (xdrproc_t)xdr_cryptkeyres;
		local = (char *(*)(uid_t, caddr_t)) key_decrypt_pk_2_svc_uid;
		check_auth = 1;
		break;


	case KEY_NET_PUT:
		xdr_argument = (xdrproc_t)xdr_key_netstarg;
		xdr_result = (xdrproc_t)xdr_keystatus;
		local = (char *(*)(uid_t, caddr_t)) key_net_put_2_svc_uid;
		check_auth = 1;
		break;
	case KEY_NET_PUTLONG:
		xdr_argument = (xdrproc_t)xdr_key_netstarg_long;
		xdr_result = (xdrproc_t)xdr_keystatus;
		local = (char *(*)(uid_t, caddr_t)) key_net_putlong_2_svc_uid;
		check_auth = 1;
		break;

	case KEY_NET_GET:
		xdr_argument = (xdrproc_t) xdr_void;
		xdr_result = (xdrproc_t)xdr_key_netstres;
		local = (char *(*)(uid_t, caddr_t)) key_net_get_2_svc_uid;
		check_auth = 1;
		break;
	case KEY_NET_GETLONG:
		xdr_argument = (xdrproc_t) xdr_void;
		xdr_result = (xdrproc_t)xdr_key_netstres_long;
		local = (char *(*)(uid_t, caddr_t)) key_net_getlong_2_svc_uid;
		check_auth = 1;
		break;

	case KEY_GET_CONV:
		xdr_argument = (xdrproc_t) xdr_keybuf;
		xdr_result = (xdrproc_t)xdr_cryptkeyres;
		local = (char *(*)(uid_t, caddr_t)) key_get_conv_2_svc_uid;
		check_auth = 1;
		break;
	case KEY_GET_CONVLONG:
		xdr_argument = (xdrproc_t) xdr_keybuf_long;
		xdr_result = (xdrproc_t)xdr_cryptkeyres;
		local = (char *(*)(uid_t, caddr_t)) key_get_convlong_2_svc_uid;
		check_auth = 1;
		break;

	default:
		svcerr_noproc(transp);
		return;
	}
	if (check_auth) {
		if (root_auth(transp, rqstp) == 0) {
			if (debugging) {
				(void) fprintf(stderr,
				"not local privileged process\n");
			}
			svcerr_weakauth(transp);
			return;
		}
		if (rqstp->rq_cred.oa_flavor != AUTH_SYS) {
			if (debugging) {
				(void) fprintf(stderr,
				"not unix authentication\n");
			}
			svcerr_weakauth(transp);
			return;
		}
		uid = ((struct authsys_parms *)rqstp->rq_clntcred)->aup_uid;
	}

	memset((char *) &argument, 0, sizeof (argument));
	if (!svc_getargs(transp, xdr_argument, (caddr_t)&argument)) {
		svcerr_decode(transp);
		return;
	}
	result = (*local) (uid, (caddr_t)&argument);
	if (!svc_sendreply(transp, xdr_result, (char *) result)) {
		if (debugging)
			(void) fprintf(stderr, "unable to reply\n");
		svcerr_systemerr(transp);
	}
	if (!svc_freeargs(transp, xdr_argument, (caddr_t)&argument)) {
		if (debugging)
			(void) fprintf(stderr,
			"unable to free arguments\n");
		exit(1);
	}
	return;
}

static int
root_auth(SVCXPRT *trans, struct svc_req *rqstp)
{
	uid_t uid;

	if (__rpc_get_local_uid(trans, &uid) < 0) {
		if (debugging)
			fprintf(stderr, "__rpc_get_local_uid failed %s %s\n",
				trans->xp_netid, trans->xp_tp);
		return (0);
	}
	if (debugging)
		fprintf(stderr, "local_uid  %d\n", uid);
	if (uid == 0)
		return (1);
	if (rqstp->rq_cred.oa_flavor == AUTH_SYS) {
		if (((uid_t) ((struct authunix_parms *)
			rqstp->rq_clntcred)->aup_uid)
			== uid) {
			return (1);
		} else {
			if (debugging)
				fprintf(stderr,
			"local_uid  %d mismatches auth %d\n", uid,
((uid_t) ((struct authunix_parms *)rqstp->rq_clntcred)->aup_uid));
			return (0);
		}
	} else {
		if (debugging)
			fprintf(stderr, "Not auth sys\n");
		return (0);
	}
}

static void
keyprogram_old(struct svc_req * rqstp, SVCXPRT * transp)
{
	union {
		cryptkeyarg key_encrypt_1_arg;
		cryptkeyarg key_decrypt_1_arg;
		des_block key_gen_1_arg;
	} argument;
	char *result;

	xdrproc_t xdr_argument, xdr_result;
	char *(*local) (uid_t, caddr_t);
	struct sockaddr_in *remote;
	struct netbuf *nb;
	int uid;
	int check_auth;

	switch (rqstp->rq_proc) {
	case NULLPROC:
		svc_sendreply(transp, (xdrproc_t)xdr_void, (char *) NULL);
		return;

	case KEY_ENCRYPT:
		xdr_argument = (xdrproc_t)xdr_cryptkeyarg;
		xdr_result = (xdrproc_t)xdr_cryptkeyres;
		local = (char *(*)(uid_t, caddr_t)) key_encrypt_1_svc_uid;
		check_auth = 1;
		break;

	case KEY_DECRYPT:
		xdr_argument = (xdrproc_t)xdr_cryptkeyarg;
		xdr_result = (xdrproc_t)xdr_cryptkeyres;
		local = (char *(*)(uid_t, caddr_t)) key_decrypt_1_svc_uid;
		check_auth = 1;
		break;

	case KEY_GEN:
		xdr_argument = (xdrproc_t)xdr_void;
		xdr_result = (xdrproc_t)xdr_des_block;
		local = (char *(*)(uid_t, caddr_t)) key_gen_1_svc;
		check_auth = 0;
		break;

#ifdef linux
	case KEY_SETLONG:
		xdr_argument = (xdrproc_t)xdr_keybuf_long;
		xdr_result = (xdrproc_t)xdr_int;
		local = (char *(*)(uid_t, caddr_t)) key_setlong_1_svc_uid;
		check_auth = 1;
		syslog(LOG_ERR, "this guy uses obsolete key_setlong");
		break;
#endif

	default:
		svcerr_noproc(transp);
		return;
	}
	syslog(LOG_ERR, "this guy uses obsolete INET keyserv");
	if (check_auth) {
		nb = svc_getrpccaller(transp);
		remote = (struct sockaddr_in*)nb->buf;
		if (nb->len != sizeof(struct sockaddr_in) ||
		    ntohs(remote->sin_port) >= IPPORT_RESERVED ||
		    ntohl(remote->sin_addr.s_addr) != INADDR_LOOPBACK) {
			if (debugging) {
				(void) fprintf(stderr,
					      "not local privileged process\n");
			}
			svcerr_weakauth(transp);
			return;
		}
		if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
			if (debugging) {
				(void) fprintf(stderr,
					       "not unix authentication\n");
			}
			svcerr_weakauth(transp);
			return;
		}
		uid = ((struct authunix_parms *) rqstp->rq_clntcred)->aup_uid;
	}
	memset((char *) &argument, 0, sizeof(argument));
	if (!svc_getargs(transp, xdr_argument, (char *)&argument)) {
		svcerr_decode(transp);
		return;
	}
	result = (*local) (uid, (caddr_t)&argument);
	if (!svc_sendreply(transp, xdr_result, (char *) result)) {
		(void) fprintf(stderr, "unable to reply\n");
		svcerr_systemerr(transp);
	}
	if (!svc_freeargs(transp, xdr_argument, (char *)&argument)) {
		(void) fprintf(stderr, "unable to free arguments\n");
		exit(1);
	}
}


static void
usage()
{
	(void) fprintf(stderr, "usage: keyserv [-n] [-D] [-d] [-k] [-c conv | dbase]\n");
	(void) fprintf(stderr, "-d disables the use of default keys\n");
	(void) fprintf(stderr, "-k disables the use of disk key store\n");
	exit(1);
}

void
sesskey_des_setparity(unsigned char *p)
{
	key_des_setparity(p, 0);
}

void
convkey_des_setparity(unsigned char *p)
{
	key_des_setparity(p, convkey_7bit);
}
