#include <stdio.h>
#include <pwd.h>
#include <string.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>

int
secretkey_decrypt(const char *netname, char *p, const char * passwd)
{
  int klen;

  klen = strlen(p) - KEYCHECKSUMSIZE;
  if (klen <= 0)
    return(0);
  if (!xxdecrypt(p, passwd, netname))
    return(0);

  if (bcmp(p, p + klen, KEYCHECKSUMSIZE) != 0) {
    p[0] = 0;
    return(1);
  }
  p[klen] = 0;
  return(1);
}

/*
 * Get somebody's encrypted secret key from the database, using
 * the given passwd to decrypt it.
 * _ANK_: hexkeylen is maximal 
 */
int
getsecretkey_long(const char * netname, keybuf_long *secretkey,
		  const char * passwd)
{
  char *p = NULL;

  if (getpublicsecretkey(netname, NULL, &p) == 0)
    return 0;

  if (p == NULL) {
    *secretkey = NULL;
    return 0;
  }

  if (secretkey_decrypt(netname, p, passwd) == 0) {
    *secretkey = NULL;
    free(p);
    return 1;
  }
  *secretkey = p;
  return(1);
}
