/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#pragma ident	"@(#)setkey.c	1.11	94/04/25 SMI"

/*
 * Copyright (c) 1986 - 1991 by Sun Microsystems, Inc.
 */

/*
 * Do the real work of the keyserver.
 * Store secret keys. Compute common keys,
 * and use them to decrypt and encrypt DES keys.
 * Cache the common keys, so the expensive computation is avoided.
 */
#include <stdio.h>
#include <mp.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpc/des_crypt.h>
#include <sys/errno.h>
#include <string.h>
#include <malloc.h>

#include "keyserv.h"
#include <rpc/keylib.h>

static MINT *MODULUS;
static int nodefaultkeys = 0;
static int nodiskkeystore = 0;

static keystatus
pk_crypt(uid_t uid, netnamestr remote_name, netobj * remote_key,
	 des_block *key, int mode);

static int
storesecretkey_long(uid_t uid, keybuf_long key);
static keybuf_long fetchsecretkey_long(uid_t uid);

static void extractdeskey(MINT *ck, des_block *deskey);

static int fetch_netname_long(uid_t uid, struct key_netstarg_long *key_netst);
static int store_netname_long(uid_t uid, key_netstarg_long * netstore);


static void seekitem(int item);
static int writeitem(uid_t uid, keybuf_long key);
static int readitem(uid_t * uidp, keybuf_long * keyp);


static void
writecache(keybuf_long pub, keybuf_long sec, des_block *deskey);
static int
readcache(keybuf_long pub, keybuf_long sec, des_block * deskey);


/*
 * prohibit the nobody key on this machine k (the -d flag)
 */
void
pk_nodefaultkeys()
{
	nodefaultkeys = 1;
}

void
pk_nodiskkeystore()
{
	nodiskkeystore = 1;
}

/*
 * Set the modulus for all our Diffie-Hellman operations 
 */
void
setmodulus(keybuf_long modx)
{
	FILE * fp = fopen("/etc/modulus", "r");
	if (fp) {
		char buf[1024];
		if (fgets(buf, sizeof(buf), fp) != NULL) {
			char * p = strrchr(buf, '\n');
			unsigned hexkeybytes;

			if (p)
				*p = 0;
			p = strchr(buf, ':');
			if (p)
				*p++ = 0;

			hexkeybytes = strlen(p);
			if (!(hexkeybytes%16) && hexkeybytes <= MAX_HEXKEYBYTES) {
				server_keybytes = hexkeybytes >> 1;
				server_hexkeybytes = hexkeybytes;
				server_keybits = server_keybytes << 3;
				MODULUS = xtom(p);
				fclose(fp);
				return;
			}
		}
		fclose(fp);
	}
	server_keybytes = DEFAULT_KEYBYTES;
	server_hexkeybytes = server_keybytes << 1;
	server_keybits = server_keybytes << 3;
	MODULUS = xtom(modx);
}

/*
 * Set the secretkey key for this uid
 */
keystatus
pk_setkey_long(uid_t uid, keybuf_long skey)
{
	if (!storesecretkey_long(uid, skey))
		return (KEY_SYSTEMERR);
	return (KEY_SUCCESS);
}

/*
 * Encrypt the key using the public key associated with remote_name and the
 * secret key associated with uid.
 */
keystatus
pk_encrypt(uid_t uid, netnamestr remote_name,
	   netobj * remote_key, des_block *key)
{
	return (pk_crypt(uid, remote_name, remote_key, key, DES_ENCRYPT));
}

/*
 * Decrypt the key using the public key associated with remote_name and the
 * secret key associated with uid.
 */
keystatus
pk_decrypt(uid_t uid, netnamestr remote_name,
	   netobj * remote_key, des_block *key)
{
	return (pk_crypt(uid, remote_name, remote_key, key, DES_DECRYPT));
}

keystatus
pk_netput_long(uid_t uid, key_netstarg_long *netstore)
{
	if (!store_netname_long(uid, netstore)) {
		return (KEY_SYSTEMERR);
	}
	return (KEY_SUCCESS);
}


keystatus
pk_netget_long(uid_t uid, key_netstarg_long * netstore)
{
	if (!fetch_netname_long(uid, netstore)) {
		return (KEY_SYSTEMERR);
	}
	return (KEY_SUCCESS);
}


/*
 * Do the work of pk_encrypt && pk_decrypt
 */
static keystatus
pk_crypt(uid_t uid, netnamestr remote_name,
	 netobj * remote_key, des_block *key, int mode)
{
	char *xsecret;
	char *xpublic;
	des_block deskey;
	int err;
	MINT *public;
	MINT *secret;
	MINT *common;
	char zero[8];
	int  free_xsecret=0;
	int  free_xpublic=0;
	char xpublic_buf[MAX_HEXKEYBYTES+1];

	xsecret = fetchsecretkey_long(uid);

	if (xsecret == NULL) {
		memset(zero, 0, sizeof (zero));
		if (nodefaultkeys)
			return (KEY_NOSECRET);
		if (!getsecretkey_long("nobody", &xsecret, zero) || !xsecret)
			return (KEY_NOSECRET);
		free_xsecret = 1;
	}
	if (remote_key) {
		xpublic = xpublic_buf;
		memcpy(xpublic_buf, remote_key->n_bytes, remote_key->n_len);
		xpublic[remote_key->n_len] = 0;
	} else {
		if (!getpublickey_long(remote_name, &xpublic) || !xpublic) {
			if (nodefaultkeys ||
			    !getpublickey_long("nobody", &xpublic) || !xpublic) {
				if (free_xsecret)
					free(xsecret);
				return (KEY_UNKNOWN);
			}
		}
		free_xpublic = 1;
	}

	if (!readcache(xpublic, xsecret, &deskey)) {
		public = xtom(xpublic);
		secret = xtom(xsecret);
		/* Sanity Check on public and private keys */
		if (!public || !secret) {
			if (free_xsecret)
				free(xsecret);
			if (free_xpublic)
				free(xpublic);
			return (KEY_SYSTEMERR);
		}

		common = itom(0);
		pow(public, secret, MODULUS, common);
		extractdeskey(common, &deskey);
		writecache(xpublic, xsecret, &deskey);
		mfree(secret);
		mfree(public);
		mfree(common);
	}
	if (free_xsecret)
		free(xsecret);
	if (free_xpublic)
		free(xpublic);
	err = ecb_crypt((char *)&deskey, (char *)key, sizeof (des_block),
		DES_HW | mode);
	if (DES_FAILED(err)) {
		return (KEY_SYSTEMERR);
	}
	return (KEY_SUCCESS);
}

keystatus
pk_get_convlong_key(uid_t uid, keybuf_long xpublic, cryptkeyres * result)
{
	char *xsecret;
	MINT *public;
	MINT *secret;
	MINT *common;
	char zero[8];
	int  free_xsecret = 0;

	if (!xpublic)
		return KEY_UNKNOWN;

	xsecret = fetchsecretkey_long(uid);

	if (xsecret == NULL) {
		memset(zero, 0, sizeof (zero));
		if (nodefaultkeys)
			return (KEY_NOSECRET);
		if (!getsecretkey_long("nobody", &xsecret, zero) || !xsecret)
			return (KEY_NOSECRET);
		free_xsecret = 0;
	}

	if (!readcache(xpublic, xsecret, &result->cryptkeyres_u.deskey)) {
		public = xtom(xpublic);
		secret = xtom(xsecret);
		/* Sanity Check on public and private keys */
		if (!public || !secret) {
			if (free_xsecret)
				free(xsecret);
			return (KEY_SYSTEMERR);
		}

		common = itom(0);
		pow(public, secret, MODULUS, common);
		extractdeskey(common, &result->cryptkeyres_u.deskey);
		writecache(xpublic, xsecret, &result->cryptkeyres_u.deskey);
		mfree(secret);
		mfree(public);
		mfree(common);
	}
	if (free_xsecret)
		free(xsecret);

	return (KEY_SUCCESS);
}

/*
 * Choose middle 64 bits of the common key to use as our des key, possibly
 * overwriting the lower order bits by setting parity.
 */
static void
extractdeskey(MINT *ck, des_block *deskey)
{
	MINT *a;
	short r;
	int i;
	short base = (1 << 8);
	char *k;

	a = itom(0);
#ifndef  __GNU_MP__
	_mp_move(ck, a);
#else
	move(ck, a);
#endif
	for (i = 0; i < ((server_keybits - 64) / 2) / 8; i++) {
		sdiv(a, base, a, &r);
	}
	k = deskey->c;
	for (i = 0; i < 8; i++) {
		sdiv(a, base, a, &r);
		*k++ = r;
	}
	mfree(a);
	convkey_des_setparity((char *)deskey);
}

/*
 * Key storage management
 */

#define	KEY_ONLY 0
#define	KEY_NAME 1

struct secretkey_netname_list {
	uid_t uid;
	key_netstarg_long keynetdata;
	u_char sc_flag;
	struct secretkey_netname_list *next;
};



static struct secretkey_netname_list *g_secretkey_netname;

/*
 * Store the keys and netname for this uid
 */
static int
store_netname_long(uid_t uid, key_netstarg_long * netstore)
{
	int nitems;
	struct secretkey_netname_list *new;
	struct secretkey_netname_list **l;

	nitems = 0;
	for (l = &g_secretkey_netname; *l != NULL && (*l)->uid != uid;
			l = &(*l)->next) {
		nitems++;
	}
	if (*l == NULL) {
		new = (struct secretkey_netname_list *)malloc(sizeof (*new));
		if (new == NULL) {
			return (0);
		}
		memset(new, 0, sizeof(*new));
		new->uid = uid;
		new->next = NULL;
		*l = new;
	} else {
		new = *l;
		if (new->keynetdata.st_priv_key)
			(void) free (new->keynetdata.st_priv_key);
		if (new->keynetdata.st_pub_key)
			(void) free (new->keynetdata.st_pub_key);
		if (new->keynetdata.st_netname)
			(void) free (new->keynetdata.st_netname);
		memset(&new->keynetdata, 0, sizeof(key_netstarg_long));
	}
	if (netstore->st_priv_key)
		new->keynetdata.st_priv_key = strdup(netstore->st_priv_key);
	if (netstore->st_pub_key)
		new->keynetdata.st_pub_key = strdup(netstore->st_pub_key);
	if (netstore->st_netname)
		new->keynetdata.st_netname = strdup(netstore->st_netname);
	new->sc_flag = KEY_NAME;
	seekitem(nitems);
	writeitem(uid, new->keynetdata.st_priv_key);
#ifdef linux
	flush_kernel_uid(uid);
#endif	
	return (1);
}

/*
 * Fetch the keys and netname for this uid
 */

static int fetch_netname_long(uid_t uid, struct key_netstarg_long *key_netst)
{
	struct secretkey_netname_list *l;

	for (l = g_secretkey_netname; l != NULL; l = l->next) {
		if (l->uid == uid) {
			memset(key_netst, 0, sizeof(struct key_netstarg_long));
			if (l->keynetdata.st_priv_key)
				key_netst->st_priv_key = l->keynetdata.st_priv_key;
			if (l->keynetdata.st_pub_key)
				key_netst->st_pub_key = l->keynetdata.st_pub_key;
			if (l->sc_flag == KEY_NAME)
				key_netst->st_netname = l->keynetdata.st_netname;
			else {
				static char buf[MAXNETNAMELEN+1];
				if (user2netname(buf, uid, NULL) == 0)
					continue;
				key_netst->st_netname = buf;
			}
			return 1;
		}
	}

	return 0;
}

static keybuf_long
fetchsecretkey_long(uid_t uid)
{
	struct secretkey_netname_list *l;

	for (l = g_secretkey_netname; l != NULL; l = l->next) {
		if (l->uid == uid) {
			return (l->keynetdata.st_priv_key);
		}
	}
	return (NULL);
}

/*
 * Store the secretkey for this uid
 */
static int
storesecretkey_long(uid_t uid, keybuf_long key)
{
	int nitems;
	struct secretkey_netname_list *new;
	struct secretkey_netname_list **l;

	nitems = 0;
	for (l = &g_secretkey_netname; *l != NULL && (*l)->uid != uid;
			l = &(*l)->next) {
		nitems++;
	}
	if (*l == NULL) {
		new = (struct secretkey_netname_list *) malloc(sizeof (*new));
		if (new == NULL) {
			return (0);
		}
		memset(new, 0, sizeof(*new));
		new->uid = uid;
		new->sc_flag = KEY_ONLY;
		new->next = NULL;
		*l = new;
	} else {
		new = *l;
		if (new->keynetdata.st_priv_key)
			free(new->keynetdata.st_priv_key);
	}

	if (key)
		new->keynetdata.st_priv_key = strdup(key);
	seekitem(nitems);
	writeitem(uid, new->keynetdata.st_priv_key);
#ifdef linux
	flush_kernel_uid(uid);
#endif
	return (1);
}

static int
hexdigit(int val)
{
	return ("0123456789abcdef"[val]);
}

static void
bin2hex(unsigned char * bin, unsigned char * hex, int size)
{
	int i;

	for (i = 0; i < size; i++) {
		*hex++ = hexdigit(*bin >> 4);
		*hex++ = hexdigit(*bin++ & 0xf);
	}
}

static int
hexval(char dig)
{
	if ('0' <= dig && dig <= '9') {
		return (dig - '0');
	} else if ('a' <= dig && dig <= 'f') {
		return (dig - 'a' + 10);
	} else if ('A' <= dig && dig <= 'F') {
		return (dig - 'A' + 10);
	} else {
		return (-1);
	}
}

static void
hex2bin(unsigned char * hex, unsigned char * bin, int size)
{
	int i;

	for (i = 0; i < size; i++) {
		*bin = hexval(*hex++) << 4;
		*bin++ |= hexval(*hex++);
	}
}

/*
 * Disk keystore management.
 */

static const char KEYSTORE[] = "/etc/keystore2";
FILE * kf;

static unsigned char * rootkey;

unsigned stored_keylen;

struct storedkey {
	int uid;
	char crypt[0];
};


int
openstore(unsigned keybytes)
{
	char keystore_name[128]; 
	unsigned ubuf;

	if (nodiskkeystore)
		return 0;

	sprintf(keystore_name, "%s.%d", KEYSTORE, keybytes << 3);

	kf = fopen(keystore_name, "r+");
	if (kf != NULL) {
		setbuf(kf, NULL);
		if (fread(&ubuf, sizeof(ubuf), 1, kf) != 1 || ubuf != keybytes) {
			fclose(kf);
			kf = NULL;
		}
	}
	if (kf == NULL) {
		kf = fopen(keystore_name, "w+");
		if (kf == NULL) {
			return (0);
		}
		setbuf(kf, NULL);
		ubuf = keybytes;
		if (fwrite(&ubuf, sizeof(ubuf), 1, kf) != 1) {
			fclose(kf);
			kf = NULL;
			return 0;
		}
	}
	stored_keylen = keybytes;
	
	return (1);
}


void
readkeys(unsigned keybytes)
{
	struct secretkey_netname_list *node;
	struct secretkey_netname_list **l;
	uid_t uid;
	char  *secretkey;

	if (kf == NULL) {
		return;
	}
	l = &g_secretkey_netname;
	seekitem(0);
	while (readitem(&uid, &secretkey)) {
		node = (struct secretkey_netname_list *)malloc(sizeof(*node));
		if (node == NULL) {
			return;
		}
		memset(node, 0, sizeof(*node));
		node->uid = uid;
		node->sc_flag = KEY_ONLY;
		node->next = NULL;
		node->keynetdata.st_priv_key = secretkey;
		*l = node;
		l = &node->next;
	}
}

void
writekeys(void)
{
	struct secretkey_netname_list *k;

	seekitem(0);
	for (k = g_secretkey_netname; k != NULL; k = k->next) {
		writeitem(k->uid, k->keynetdata.st_priv_key);
	}
}

void
seekitem(int item)
{
	if (kf != NULL) {
		fseek(kf, sizeof(unsigned) +
		      item * (sizeof(struct storedkey) + stored_keylen), 0);
	}
}

int
writeitem(uid_t uid, keybuf_long key)
{
	char buf[sizeof(struct storedkey) + stored_keylen];
	struct storedkey * item = (struct storedkey*)buf;
	char rootkey_tmp[MAX_KEYBYTES];
	int reencrypt;
	
	if (kf == NULL || !key) {
		return (1);
	}
	if (uid == 0) {
		writerootkey(key);
		hex2bin(key, rootkey_tmp, server_keybytes);
		reencrypt = (rootkey &&
			     memcmp(rootkey, rootkey_tmp, server_keybytes) != 0);
		if (!rootkey)
			rootkey = malloc(server_keybytes);
		if (!rootkey)
			return 0;
		memcpy(rootkey, rootkey_tmp, server_keybytes);
		if (reencrypt) {
			writekeys();
			return (1);
		}
	}
	if (!rootkey) {
		return (1);
	}
	item->uid = uid;
	hex2bin(key, item->crypt, stored_keylen);
	ecb_crypt(rootkey, item->crypt, stored_keylen, DES_ENCRYPT|DES_HW);
	return (fwrite(item, sizeof(struct storedkey) + stored_keylen, 1, kf) >= 0);
}

static int
readitem(uid_t * uidp, keybuf_long * keyp)
{
	char buf[sizeof(struct storedkey) + stored_keylen];
	struct storedkey * item = (struct storedkey*)buf;

	if (!rootkey || kf == NULL) {
		return (0);
	}
	if (fread(&buf, sizeof(buf), 1, kf) != 1) {
		return (0);
	}
	*keyp = malloc(server_hexkeybytes+1);
	if (!*keyp)
		return 0;
	*uidp = (uid_t)item->uid;
	ecb_crypt(rootkey, item->crypt, stored_keylen, DES_DECRYPT|DES_HW);
	bin2hex(item->crypt, *keyp, server_keybytes);
	(*keyp)[server_hexkeybytes] = 0;
	return (1);
}

/*
 * Exponential caching management
 */

struct cachekey_list {
	keybuf_long secret;
	keybuf_long public;
	des_block deskey;
	struct cachekey_list *next;
};
static struct cachekey_list *g_cachedkeys;


/*
 * cache result of expensive multiple precision exponential operation 
 */
static void
writecache(keybuf_long pub, keybuf_long sec, des_block *deskey)
{
	struct cachekey_list *new;

	new = (struct cachekey_list *) malloc(sizeof(struct cachekey_list));
	if (new == NULL) {
		return;
	}
	new->public = strdup(pub);
	new->secret = strdup(sec);
	new->deskey = *deskey;
	new->next = g_cachedkeys;
	g_cachedkeys = new;
}

/*
 * Try to find the common key in the cache 
 */
static int
readcache(keybuf_long pub, keybuf_long sec, des_block * deskey)
{
	struct cachekey_list *found;
	register struct cachekey_list **l;

#define cachehit(pub, sec, list)	\
		(strcmp(pub, (list)->public) == 0 && \
		strcmp(sec, (list)->secret) == 0)

	for (l = &g_cachedkeys;
	     (*l) != NULL && !cachehit(pub, sec, *l);
	     l = &(*l)->next);
	if ((*l) == NULL) {
		return (0);
	}
	found = *l;
	(*l) = (*l)->next;
	found->next = g_cachedkeys;
	g_cachedkeys = found;
	*deskey = found->deskey;
	return (1);
}

