#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <syslog.h>
#include <netinet/in.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netdb.h>
#include <stddef.h>
#include <rpc/rpc.h>

#include "inetd.h"

static FILE * fconfig = NULL;
static char   line[256];

static char * 
newstr(char * cp)
{
	if ((cp = strdup(cp ? cp : "")) != NULL)
		return(cp);
	logerr("strdup: %m");
	cleanup(-1);
}

static char * 
nextline(void)
{
	char *cp;

	if (fgets(line, sizeof (line), fconfig) == NULL)
		return (NULL);
	cp = index(line, '\n');
	if (cp)
		*cp = '\0';
	return line;
}


static char * 
skip(char ** cpp)
{
	char *cp = *cpp;
	char *start;

again:
	while (*cp == ' ' || *cp == '\t')
		cp++;
	if (*cp == '\0') {
		int c;
		c = getc(fconfig);
		(void) ungetc(c, fconfig);
		if (c == ' ' || c == '\t')
			if ((cp = nextline()) != NULL)
				goto again;
		*cpp = NULL;
		return (NULL);
	}
	start = cp;
	while (*cp && *cp != ' ' && *cp != '\t')
		cp++;
	if (*cp != '\0')
		*cp++ = '\0';
	*cpp = cp;
	return start;
}

#ifdef not_yet
static int
parse_rpcprog(char * arg, struct port * port)
{
	char * progptr = arg;
	char * versptr;
	char * nextprogptr;
	u_long rpcprognum;
	struct pmaplist * pml = port->pml;
	int vers, ver_lo, ver_hi;

	do {
		if ((nextprogptr = strchr(progptr, '|')) != NULL)
			*nextprogptr++ = 0;
		if ((versptr = strchr(progptr,':')) == NULL) {
			logerr("missing version \"%s\"", arg);
			return 1;
		}
    *versptr++ = 0;
    if (progptr[0] >= '0' && progptr[0] <= '9')
      rpcprognum = atoi(progptr);
    else {
      struct rpcent * rpcent = NULL;
      rpcent = getrpcbyname(progptr);
      if (!rpcent) {
        logerr("invalid rpc program name \"%s\"\n", arg);
        return 1;
      }
      rpcprognum = rpcent->r_number;
    }
    ver_lo = ver_hi = atoi(versptr);
    if ((versptr = strchr(versptr,'-')) != NULL)
      ver_hi = atoi(versptr+1);
    for (vers = ver_lo; vers <= ver_hi; vers++) {
      struct pmaplist * tpml;
      tpml = (struct pmaplist *)calloc(sizeof(struct pmaplist), 1);
      tpml->pml_map.pm_prog = rpcprognum;
      tpml->pml_map.pm_vers = vers;
      tpml->pml_next = NULL;
      if (pml)
	pml->pml_next = tpml;
      else
        port->pml = tpml;
      pml = tpml;
    }
    progptr = nextprogptr; 
  } while (progptr);
  return 0;
}
#endif

void nc_init_port(struct port *port)
{
	struct nd_hostserv nd;
	struct nd_addlist *na;

	nd.h_host = HOST_SELF;
	nd.h_serv = port->service;
	if (netdir_getbyname(port->nconf, &nd, &na) < 0)
		return;

	if (na->n_cnt) {
		port->bnd.addr = na->n_addrs[0];
		port->bnd.qcnt = 8;
		na->n_addrs[0].buf = NULL;
		na->n_addrs[0].len = 0;
		na->n_addrs[0].maxlen = 0;
		port->uaddr = taddr2uaddr(port->nconf, &port->bnd.addr);
	}
	netdir_free(na, ND_ADDRLIST);
}

struct servtab * 
getconfigent(void)
{
	register struct servtab *sep;
	char *cp, *arg;

	sep = (struct servtab*)malloc(sizeof(struct servtab));

more:
	memset(sep, 0, sizeof(struct servtab));

	while ((cp = nextline()) && (*cp == '#' || *cp == '\0'))
		/*NOTHING*/;

	if (cp == NULL) {
		free(sep);
		return NULL;
	}

	arg = skip(&cp);
        
	while (arg) {
		struct port * port = (struct port *)calloc(sizeof(struct port), 1);
#ifdef not_yet
		struct rpcbinding * rpcbl;
#endif
		struct netconfig *nconf;
		char service[64];
		char nettype[128];
		char * ptr;

#ifdef not_yet
		if ((ptr = strchr(arg,'/')) != NULL) {
			*ptr++ = 0;
			parse_rpcprog(arg, port);
			arg = ptr;
		}
#endif
		strncpy(service, arg, sizeof(service)-1);
		arg = skip(&cp);
		strncpy(nettype, arg, sizeof(nettype)-1);
		arg = skip(&cp);
		port->fd = -1;
#ifdef not_yet
		for (rpcbl = port->rpcbl; rpcbl; rpcbl = rpcbl->pml_next) {
			pml->pml_map.pm_prot = port->proto;
			pml->pml_map.pm_port = ntohs(IN_PORT(port));
		}
		if (port->pml && !IN_PORT(port))
			port->autorpc = 1;
#endif
		arg = skip(&cp);
		while (arg) {
			if (strcmp(arg,"accept") == 0)
				port->accept = 1;
			else if (strcmp(arg,"noaccept") == 0)
				port->accept = 0;
			else if (strcmp(arg,"reuse") == 0)
				port->reuse = 1;
			else if (strcmp(arg,"debug") == 0)
				port->debug = 1;
			else if (strcmp(arg,"keepalive") == 0)
				port->keepalive = 1;
			else if (strcmp(arg,"log") == 0)
				port->log = 1;
			else if (strcmp(arg,"downgrade") == 0)
				port->downgrade = 1;
			else if (strcmp(arg,"rfc931") == 0) {
				port->rfc931 = 1;
				port->log = 1;
			} else
				break;
			arg = skip(&cp);
		}

		for (nconf = first_netconf(nettype); nconf;
		     nconf = next_netconf(nettype, nconf)) {
			struct port *port1 = port_clone(port);
			if (port1) {
				port1->nconf = nconf;
				nc_init_port(port1);
				add_to_port_list(&sep->port, port1);
			}
		}
		free_port(port);

		if (strcmp(arg,"&"))
			break;
		arg = skip(&cp);
	}

	while (arg) {
		if (strcmp(arg,"wait") == 0)
			sep->wait = 1;
		else if (strcmp(arg,"nowait") == 0) 
			sep->wait = 0;
		else
			break;
		arg = skip(&cp);
	}

	sep->user = newstr(arg);
	arg = skip(&cp);
	sep->server = newstr(arg);

	if (sep->port == NULL) {
		logerr("service [%s] have no ports\n");
		goto more;
	}

	if (strcmp(sep->server, "internal") == 0) {
		struct biltin *bi;

		if (sep->port->next) {
			logerr("internal service may listen only one port\n");
			free_port_list(&sep->port);
			goto more;
		}

		for (bi = biltins; bi->servicename; bi++) {
			if (bi->semantics == sep->port->nconf->nc_semantics &&
			    strcmp(bi->servicename, sep->port->service) == 0 &&
			    sep->wait == bi->wait)
				break;
		}
 
		if (bi->servicename == 0) {
			logerr("internal service [%s/%s] is unknown\n",
			       sep->port->service, sep->port->nconf->nc_netid);
			free_port_list(&sep->port);
			goto more;
		}

		free(sep->server);
		sep->server = newstr(bi->servicename);
		sep->bi = bi;
	} else
		sep->bi = NULL;

	_BLOCK_BEGIN
	int argc = 0;

	for (arg = skip(&cp); cp; arg = skip(&cp)) {
		if (argc < MAXARGV)
			sep->argv[argc++] = newstr(arg);
	}
	while (argc <= MAXARGV)
		sep->argv[argc++] = NULL;
	_BLOCK_END
	return (sep);
}

int
setconfig(char * name)
{
	if (fconfig != NULL) {
		fseek(fconfig, 0L, SEEK_SET);
		return 1;
	}
	fconfig = fopen(name, "r");
	return (fconfig != NULL);
}

void 
endconfig(void)
{
	if (fconfig) {
		fclose(fconfig);
		fconfig = NULL;
	}
}
