#ifndef _INETD_H_
#define _INETD_H_ 1

#define _BLOCK_BEGIN {
#define _BLOCK_END   }

void cleanup(int code) __attribute__((noreturn));

#include <sys/time.h>
#include <sys/tiuser.h>
#include <sys/netconfig.h>
#include <netdir.h>


struct  port
{
	struct port	*next, *prev;
	int		fd;
	struct netconfig *nconf;
	struct proto_ops *ops;
	struct t_bind	bnd;
	char		*uaddr;
	char		*service;
	struct rpcbinding *rpcbl;
	time_t		last_use;
	unsigned char accept   : 1,
	        debug    : 1,
                log      : 1,
	        downgrade: 1,
                rfc931   : 1,
                autorpc  : 1,
                keepalive: 1,
	        reuse    : 1;
};

void add_to_port_list(struct port ** pport, struct port * port);
void free_port_list(struct port ** pport);

/* server tables */

#define	MAXARGV 16

struct	servtab
{
	char	*user;		        /* user name to run as */
	char    *server;
	struct biltin * bi;		/* if built-in, description */
	char	*argv[MAXARGV+1];	/* program arguments */
	int     pid;
	int	count;		        /* number started since start_time */
	struct  timeval start_time;
	struct  servtab * next, *prev;
	struct  port * port;
	unsigned char obsolete : 1,	
	              wait     : 1;
};

struct proto_ops
{
	char *netid;
	int  (downgrade)(struct servtab, struct port*, int);
};

extern struct servtab * servtab;


/* Builtin server  definitions */

typedef void builtin_fn_t(int fd, struct servtab * sep);		

struct biltin
{
	char	* servicename;		/* internally provided service name */
	unsigned long semantics;
	unsigned char forking : 1,
	              wait    : 1;
	builtin_fn_t * bi_fn;		/* function which performs it */
};

extern struct biltin biltins[];


/* inetd.conf.c */

struct servtab * getconfigent(void);

int  setconfig(char *);
void endconfig(void);
void discard_servtab(struct servtab * sep);
void set_port_options(struct servtab * sep, struct port * port);
void activate_fd(int fd);

#define logerr(fmt...) ({ syslog(LOG_ERR, fmt); /* if (!debug) debug++; */ })
#define loginfo(fmt...) syslog(LOG_INFO, fmt)

#ifdef DEBUG
extern int debug;

#define dprintf(level, fmt...) ({ if ((level) <= debug) \
				    syslog(LOG_DEBUG, fmt); })

#else
#define dprintf(level, fmt...) (0)
#endif

#endif /*_INETD_H_*/
