#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <malloc.h>
#include <tiuser.h>
#include <netdir.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>

static char thishostname[64];

struct nd_addrlist*
inet_prep_broadcasts(unsigned short port)
{
	struct ifconf ifc;
        struct ifreq ifreq, *ifr;
	struct sockaddr_in *sin;
        int n, i;
	char	buf[4096];
	int nif;
	struct nd_addrlist *nd;
	struct netbuf *nbl;
	int sock;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0)
		return NULL;

        ifc.ifc_len = sizeof(buf);
        ifc.ifc_buf = buf;
        if (ioctl(sock, SIOCGIFCONF, (char *)&ifc) < 0) {
		close(sock);
                return NULL;
	}

        ifr = ifc.ifc_req;
	nif = 0;
        for (i = 0, n = ifc.ifc_len/sizeof (struct ifreq); n > 0; n--, ifr++) {
		if (ifr->ifr_addr.sa_family != AF_INET)
			continue;
		nif++;
	}

	if (!nif) {
		close(sock);
		return NULL;
	}

	nbl = malloc(sizeof(struct netbuf)*nif);
	if (!nbl) {
		close(sock);
		return NULL;
	}
	memset(nbl, 0, sizeof(struct netbuf)*nif);

	nd = malloc(sizeof(*nd));
	if (!nd) {
		close(sock);
		free(nbl);
		return NULL;
	}
	nd->n_cnt = 0;
	nd->n_addrs = nbl;

        ifr = ifc.ifc_req;
	nif = 0;
        for (i = 0, n = ifc.ifc_len/sizeof (struct ifreq); n > 0; n--, ifr++) {
                ifreq = *ifr;
                if (ioctl(sock, SIOCGIFFLAGS, (char *)&ifreq) < 0)
                        continue;
		if (!(ifreq.ifr_flags & IFF_UP) ||
		    (ifreq.ifr_flags & IFF_LOOPBACK))
			continue;
                if (ifreq.ifr_flags & IFF_BROADCAST) {
			if (ioctl(sock, SIOCGIFBRDADDR, &ifreq) < 0)
				continue;
		} else if (ifreq.ifr_flags & IFF_POINTOPOINT) {
			if (ioctl(sock, SIOCGIFDSTADDR, (char *)&ifreq) < 0)
				continue;
		} else
			continue;

		nbl[nif].buf = malloc(sizeof(struct sockaddr_in));
		if (!nbl[nif].buf)
			continue;
		nbl[nif].maxlen = sizeof(struct sockaddr_in);
		nbl[nif].len = sizeof(struct sockaddr_in);
		((struct sockaddr_in*)&ifreq.ifr_addr)->sin_port = port;
		memcpy(nbl[nif].buf, &ifreq.ifr_addr, sizeof(struct sockaddr_in));
		nif++;
	}
	close(sock);

	if (nif == 0) {
		free(nbl);
		free(nd);
		return NULL;
	}
	nd->n_cnt = nif;

	return nd;
}


struct nd_addrlist*
_netdir_getbyname(const struct netconfig* nconf,
		  const struct nd_hostserv* hserv)
{
	struct nd_addrlist *nd;
	struct hostent *he;
	struct servent *se;
	struct sockaddr_in sin;
	struct netbuf *nb;
	char   *host = hserv->h_host;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;

	if (hserv->h_serv) {
		se = getservbyname(hserv->h_serv, nconf->nc_proto);
		if (!se) {
			_nderror = ND_NOSERV;
			return NULL;
		}
		sin.sin_port = se->s_port;
	}
	if (host) {
		if (strcmp(host, HOST_SELF)==0) {
			sin.sin_addr.s_addr = 0;
			goto resolved;
		} else if (strcmp(host, HOST_SELF_CONNECT)==0) {
			sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
			goto resolved;
		} else if (strcmp(host, HOST_ANY)==0) {
/* ? */
			if (!thishostname[0])
				gethostname(thishostname, sizeof(thishostname));
			host = thishostname;
		} else if (strcmp(host, HOST_BROADCAST)==0) {
			if (!(nconf->nc_flag & NC_BROADCAST)) {
				_nderror = ND_NOHOST;
				return NULL;
			}
			return inet_prep_broadcasts(sin.sin_port);
		} else if (strcmp(host, HOST_ALLNODES)==0) {
			if (!(nconf->nc_flag & NC_MULTICAST)) {
				_nderror = ND_NOHOST;
				return NULL;
			}
			sin.sin_addr.s_addr = htonl(0xE0000001);
			goto resolved;
		} else if (strcmp(host, HOST_RPCNODES)==0) {
			if (!(nconf->nc_flag & NC_MULTICAST)) {
				_nderror = ND_NOHOST;
				return NULL;
			}
			sin.sin_addr.s_addr = htonl(0xE0000202);
			goto resolved;
		}
		he = gethostbyname(host);
		if (!he) {
			_nderror = ND_NOHOST;
			return NULL;
		}
		sin.sin_addr.s_addr = *(__u32*)he->h_addr;
	}

resolved:
	nd = malloc(sizeof(*nd));
	nd->n_cnt = 1;
	nb = nd->n_addrs = malloc(sizeof(*nb));
	nb->buf = malloc(sizeof(sin));
	nb->len = nb->maxlen = sizeof(sin);
	memcpy(nb->buf, &sin, sizeof(sin));
	return nd;
}

struct nd_hostservlist*
_netdir_getbyaddr(const struct netconfig* nconf,
		  const struct netbuf *addr)
{
	struct nd_hostservlist* nd;
	struct hostent *he;
	struct servent *se;
	struct sockaddr_in *sin = (struct sockaddr_in*)addr->buf;

	if (addr->len != sizeof(*sin) || sin->sin_family != AF_INET) {
		_nderror = ND_BADARG;
		return NULL;
	}

	if (sin->sin_port) {
		se = getservbyport(sin->sin_port, nconf->nc_proto);
		if (!se) {
			_nderror = ND_NOSERV;
			return NULL;
		}
	}
	if (sin->sin_addr.s_addr) {
		he = gethostbyaddr((char*)&sin->sin_addr, AF_INET, 4);
		if (!he) {
			_nderror = ND_NOHOST;
			return NULL;
		}
	}

	nd = malloc(sizeof(*nd));
	nd->h_cnt = 1;
	nd->h_hostservs = malloc(sizeof(*(nd->h_hostservs)));
	
	if (sin->sin_addr.s_addr)
		nd->h_hostservs->h_host = strdup(he->h_name);
	else
		nd->h_hostservs->h_host = NULL;
	if (sin->sin_port)
		nd->h_hostservs->h_serv = strdup(se->s_name);
	else
		nd->h_hostservs->h_serv = NULL;
	return nd;
}


#define MAX_UADDR (4*(3+1)+2*(3+1))

static int
_inet_mergeaddr(const struct netconfig* nconf, struct nd_mergearg* ma)
{
	char *p;
	if (memcmp("0.0.0.0.", ma->s_uaddr, 8) != 0) {
		ma->m_uaddr = strdup(ma->s_uaddr);
		return 0;
	}
	ma->m_uaddr = malloc(MAX_UADDR+1);
	strncpy(ma->m_uaddr, ma->c_uaddr, MAX_UADDR);
	ma->m_uaddr[MAX_UADDR] = 0;
	p = strrchr(ma->m_uaddr, '.');
	if (!p)
		goto fail;
	*p = 0;

	p = strrchr(ma->m_uaddr, '.');
	if (!p)
		goto fail;
	*p = 0;
	strcat(ma->m_uaddr, ma->s_uaddr+7);
	return 0;

fail:
	_nderror = ND_BADARG;
	free(ma->m_uaddr);
	ma->m_uaddr = NULL;
	return -1;
}


int
_netdir_options(const struct netconfig* nconf,
		int option, int fd, const char* par)
{
	int on;
	struct netbuf n;
	struct sockaddr_in sa;
	struct netbuf *nb;
	struct sockaddr_in *sin;
	unsigned short *pp;
	int res;

	switch (option) {
	case ND_SET_BROADCAST:
		if (!(nconf->nc_flag&NC_BROADCAST))
			return -1;
		on = 1;
		return setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on));

	case ND_SET_RESERVEDPORT:
		nb = (struct netbuf*)par;
		if (!par) {
			n.buf = (char*)&sa;
			sa.sin_family = AF_INET;
			sa.sin_addr.s_addr = 0;
			n.maxlen = n.len = sizeof(sa);
			nb = &n;
		}
		if (nb->len != sizeof(*sin))
			return -1;
		sin = (struct sockaddr_in*)nb->buf;
		if (sin->sin_family != AF_INET)
			return -1;
		pp = &sin->sin_port;
		*pp = htons(1023);
		while ((res = bind(fd, (struct sockaddr*)sin, sizeof(sin))) < 0
		       && errno == EADDRINUSE)
			*pp = htons(htohs(*pp) - 1);
		t_sync(fd);
		return res;

	case ND_CHECK_RESERVEDPORT:
		nb = (struct netbuf*)par;
		if (nb->len != sizeof(*sin))
			return -1;
		sin = (struct sockaddr_in*)nb->buf;
		if (sin->sin_family != AF_INET)
			return -1;
		if (htohs(sin->sin_port) < 1024)
			return 0;
		return -1;
	case ND_MERGEADDR:
		return _inet_mergeaddr(nconf, (struct nd_mergearg*)par);
	case ND_JOIN_MULTICAST:
	case ND_LEAVE_MULTICAST:
	{
		struct ip_mreq imr;
		nb = (struct netbuf*)par;
		if (nb->len != sizeof(*sin))
			return -1;
		sin = (struct sockaddr_in*)nb->buf;
		if (sin->sin_family != AF_INET)
			return -1;
		if (!(nconf->nc_flag&NC_MULTICAST))
			return -1;
		memset(&imr, 0, sizeof(imr));
		imr.imr_multiaddr = sin->sin_addr;
		return setsockopt(fd, SOL_IP,
				  option == ND_JOIN_MULTICAST ? IP_ADD_MEMBERSHIP : IP_DROP_MEMBERSHIP,
				  &imr, sizeof(imr));
	}
	default:
		return -1;
	}
}

char *
_taddr2uaddr(const struct netconfig* nconf,
		    const struct netbuf *taddr)
{
	char *up, *a, *p;
	struct sockaddr_in *sin = (struct sockaddr_in *)taddr->buf;
	if (taddr->len != sizeof(*sin))
		return NULL;
	if (sin->sin_family != AF_INET)
		return NULL;
	up = (char*)malloc(MAX_UADDR+1);
	if (!up)
		return NULL;
	a = (unsigned char*)&sin->sin_addr;
	p = (unsigned char*)&sin->sin_port;
	sprintf(up, "%d.%d.%d.%d.%d.%d", a[0], a[1], a[2], a[3], p[0], p[1]);
	return up;
}

struct netbuf*
_uaddr2taddr(const struct netconfig* nconf,
		    const char *uaddr)
{
	struct netbuf *nb;
	struct sockaddr_in *sin;
	int    buf[6];
#if 0
	char tbuf[3+1+3+1+3+1+3+1+3+1+3+1];
#endif
	char *a, *p;

#if 0
	strncpy(tbuf, uaddr, 3+1+3+1+3+1+3+1+3+1+3);
	tbuf[3+1+3+1+3+1+3+1+3+1+3] = 0;
#endif

	nb = malloc(sizeof(struct netbuf));
	nb->maxlen = nb->len = sizeof(struct sockaddr_in);
	nb->buf = malloc(sizeof(struct sockaddr_in));
	sin = (struct sockaddr_in*)nb->buf;
	sin->sin_family = AF_INET;
	a = (unsigned char*)&sin->sin_addr;
	p = (unsigned char*)&sin->sin_port;
	sscanf(uaddr, "%d.%d.%d.%d.%d.%d", buf, buf+1, buf+2,
	       buf+3, buf+4, buf+5);
	a[0] = buf[0];
	a[1] = buf[1];
	a[2] = buf[2];
	a[3] = buf[3];
	p[0] = buf[4];
	p[1] = buf[5];
	
	return nb;
}
