#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <malloc.h>
#include <tiuser.h>
#include <netdir.h>
#include <netdb.h>
#include <sys/socket.h>
#include <linux/in6.h>

static char thishostname[64];

struct nd_addrlist*
_netdir_getbyname(const struct netconfig* nconf,
		  const struct nd_hostserv* hserv)
{
	struct nd_addrlist *nd;
	struct hostent *he;
	struct servent *se;
	struct sockaddr_in6 sin;
	struct netbuf *nb;
	char   *host = hserv->h_host;

	memset(&sin, 0, sizeof(sin));
	sin.sin6_family = AF_INET6;

	if (hserv->h_serv) {
		if (strcmp(hserv->h_serv, "rpcbind") == 0)
			se = getservbyname("rpcbind6", nconf->nc_proto);
		else
			se = getservbyname(hserv->h_serv, nconf->nc_proto);
		if (!se) {
			_nderror = ND_NOSERV;
			return NULL;
		}
		sin.sin6_port = se->s_port;
	}
	if (host) {
		if (strcmp(host, HOST_SELF) == 0) {
			sin.sin6_addr.s6_addr32[0] = 0;
			sin.sin6_addr.s6_addr32[1] = 0;
			sin.sin6_addr.s6_addr32[2] = 0;
			sin.sin6_addr.s6_addr32[3] = 0;
			goto resolved;
		} else if (strcmp(host, HOST_SELF_CONNECT) == 0) {
			sin.sin6_addr.s6_addr32[0] = 0;
			sin.sin6_addr.s6_addr32[1] = 0;
			sin.sin6_addr.s6_addr32[2] = 0;
			sin.sin6_addr.s6_addr32[3] = htonl(1);
			goto resolved;
		} else if (strcmp(host, HOST_ANY) == 0) {
/* ? */
			if (!thishostname[0])
				gethostname(thishostname, sizeof(thishostname));
			host = thishostname;
		} else if (strcmp(host, HOST_BROADCAST) == 0) {
			if (nconf->nc_semantics != NC_TPI_CLTS) {
				_nderror = ND_NOHOST;
				return NULL;
			}

			sin.sin6_addr.s6_addr32[0] = htonl(0xFF020000);
			sin.sin6_addr.s6_addr32[1] = 0;
			sin.sin6_addr.s6_addr32[2] = 0;
			sin.sin6_addr.s6_addr32[3] = htonl(0x1);

			goto resolved;
		}
		he = gethostbyname2(host, AF_INET6);
		if (!he) {
			_nderror = ND_NOHOST;
			return NULL;
		}
		memcpy(&sin.sin6_addr, he->h_addr, sizeof(struct in6_addr));
	}

resolved:
	nd = malloc(sizeof(*nd));
	nd->n_cnt = 1;
	nb = nd->n_addrs = malloc(sizeof(*nb));
	nb->buf = malloc(sizeof(sin));
	nb->len = nb->maxlen = sizeof(sin);
	memcpy(nb->buf, &sin, sizeof(sin));
	return nd;
}

static const struct in6_addr in6addr_any;

struct nd_hostservlist*
_netdir_getbyaddr(const struct netconfig* nconf,
		  const struct netbuf *addr)
{
	struct nd_hostservlist* nd;
	struct hostent *he;
	struct servent *se;
	struct sockaddr_in6 *sin = (struct sockaddr_in6*)addr->buf;

	if (addr->len != sizeof(*sin) || sin->sin6_family != AF_INET6) {
		_nderror = ND_BADARG;
		return NULL;
	}

	if (sin->sin6_port) {
		se = getservbyport(sin->sin6_port, nconf->nc_proto);
		if (!se) {
			_nderror = ND_NOSERV;
			return NULL;
		}
		if (strcmp(se->s_name, "rpcbind6") == 0) {
			se = getservbyname("rpcbind", nconf->nc_proto);
			if (!se) {
				_nderror = ND_NOSERV;
				return NULL;
			}
		}
	}
	if (memcmp(&sin->sin6_addr, &in6addr_any, sizeof(*sin))) {
		he = gethostbyaddr((char*)&sin->sin6_addr, AF_INET6, sizeof(sin->sin6_addr));
		if (!he) {
			_nderror = ND_NOHOST;
			return NULL;
		}
	}

	nd = malloc(sizeof(*nd));
	nd->h_cnt = 1;
	nd->h_hostservs = malloc(sizeof(*(nd->h_hostservs)));
	
	if (memcmp(&sin->sin6_addr, &in6addr_any, sizeof(in6addr_any)))
		nd->h_hostservs->h_host = strdup(he->h_name);
	else
		nd->h_hostservs->h_host = NULL;
	if (sin->sin6_port)
		nd->h_hostservs->h_serv = strdup(se->s_name);
	else
		nd->h_hostservs->h_serv = NULL;
	return nd;
}

#define MAX_UADDR6 (8*(4+1)+2*(3+1))

static int
_inet_mergeaddr(const struct netconfig* nconf, struct nd_mergearg* ma)
{
	char *p;
	char *tbuf = NULL;
	struct in6_addr addr;
	char *sport;

	tbuf = malloc(MAX_UADDR6+1);
	strncpy(tbuf, ma->s_uaddr, MAX_UADDR6);
	tbuf[MAX_UADDR6] = 0;

	p = strrchr(tbuf, '.');
	if (!p)
		goto fail;
	p = strrchr(--p, '.');
	if (!p)
		goto fail;
	*p = 0;
	sport = ma->s_uaddr + (p-tbuf);

	if (inet_pton(AF_INET6, tbuf, &addr) != 1)
		goto fail;

	if (memcmp(&addr, &in6addr_any, sizeof(addr)) != 0) {
		*p = '.';
		ma->m_uaddr = tbuf;
		return 0;
	}

	strncpy(tbuf, ma->c_uaddr, MAX_UADDR6);

	p = strrchr(tbuf, '.');
	if (!p)
		goto fail;
	p = strrchr(--p, '.');
	if (!p)
		goto fail;
	strcpy(p, sport);
	ma->s_uaddr = tbuf;
	return 0;

fail:
	_nderror = ND_BADARG;
	if (tbuf)
		free(tbuf);
	return -1;
}


int
_netdir_options(const struct netconfig* nconf,
		int option, int fd, const char* par)
{
	int on;
	struct netbuf n;
	struct sockaddr_in6 sa;
	struct netbuf *nb;
	struct sockaddr_in6 *sin;
	unsigned short *pp;
	int res;

	switch (option) {
	case ND_SET_BROADCAST:
		return -1;
	case ND_SET_RESERVEDPORT:
		nb = (struct netbuf*)par;
		if (!par)
		{
			n.buf = (char*)&sa;
			sa.sin6_family = AF_INET6;
			memset(&sa.sin6_addr, 0, sizeof(sa.sin6_addr));
			n.maxlen = n.len = sizeof(sa);
			nb = &n;
		}
		if (nb->len != sizeof(*sin))
			return -1;
		sin = (struct sockaddr_in6*)nb->buf;
		if (sin->sin6_family != AF_INET6)
			return -1;
		pp = &sin->sin6_port;
		*pp = htons(1023);
		while ((res = bind(fd, (struct sockaddr*)sin, sizeof(sin))) < 0
		       && errno == EADDRINUSE)
			*pp = htons(htohs(*pp) - 1);
		t_sync(fd);
		return res;
	case ND_CHECK_RESERVEDPORT:
		nb = (struct netbuf*)par;
		if (nb->len != sizeof(*sin))
			return -1;
		sin = (struct sockaddr_in6*)nb->buf;
		if (sin->sin6_family != AF_INET6)
			return -1;
		if (htohs(sin->sin6_port) < 1024)
			return 0;
		return -1;
	case ND_MERGEADDR:
		return _inet6_mergeaddr(nconf, (struct nd_mergearg*)par);
	case ND_JOIN_MULTICAST:
	case ND_LEAVE_MULTICAST:
	{
		struct ipv6_mreq imr;
		nb = (struct netbuf*)par;
		if (nb->len != sizeof(*sin))
			return -1;
		sin = (struct sockaddr_in6*)nb->buf;
		if (sin->sin6_family != AF_INET6)
			return -1;
		if (!(nconf->nc_flag&NC_MULTICAST))
			return -1;
		memset(&imr, 0, sizeof(imr));
		memcpy(&imr.ipv6mr_multiaddr, &sin->sin6_addr, sizeof(struct in6_addr));
		return setsockopt(fd, SOL_IPV6,
				  option == ND_JOIN_MULTICAST ? IPV6_ADD_MEMBERSHIP : IPV6_DROP_MEMBERSHIP,
				  &imr, sizeof(imr));
	}
	default:
		return -1;
	}
}


char *
_taddr2uaddr(const struct netconfig* nconf, const struct netbuf *taddr)
{
	char *up, *p;
	struct sockaddr_in6 *sin = (struct sockaddr_in6 *)taddr->buf;

	if (taddr->len != sizeof(*sin))
		return NULL;
	if (sin->sin6_family != AF_INET6)
		return NULL;
	up = (char*)malloc(MAX_UADDR6+1);
	if (!up)
		return NULL;
	if (inet_ntop(AF_INET6, &sin->sin6_addr, up, MAX_UADDR6) == NULL) {
		free(up);
		return NULL;
	}
	p = (unsigned char*)&sin->sin6_port;
	sprintf(up+strlen(up), ".%d.%d", p[0], p[1]);
	return up;
}

struct netbuf*
_uaddr2taddr(const struct netconfig* nconf,
		    const char *uaddr)
{
	struct netbuf *nb;
	struct sockaddr_in6 *sin;
	char tbuf[MAX_UADDR6+1];
	char *a, *p;

	strncpy(tbuf, uaddr, MAX_UADDR6);
	tbuf[MAX_UADDR6] = 0;

	a = strrchr(tbuf, '.');
	if (!a)
		return NULL;
	a = strrchr(--a, '.');
	if (!a)
		return NULL;
	*a++ = 0;

	nb = malloc(sizeof(struct netbuf));
	nb->maxlen = nb->len = sizeof(struct sockaddr_in6);
	nb->buf = malloc(sizeof(struct sockaddr_in6));
	sin = (struct sockaddr_in6*)nb->buf;
	sin->sin6_family = AF_INET6;

	p = (unsigned char*)&sin->sin6_port;

	if (inet_pton(AF_INET6, tbuf, &sin->sin6_addr) != 1 ||
	    sscanf(a, "%d.%d", p[0], p[1]) != 2) {
		free(nb->buf);
		free(nb);
		return NULL;
	}

	return nb;
}
