#include <unistd.h>
#include <sys/socket.h>
#include <time.h>

#include <netinet/in.h>
#include <syslog.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netdb.h>
#include <stddef.h>

#include "inetd.h"

#ifdef FAKE_SOMEBODY
static void setproctitle(char * name, int s) {
  int size;
  register char *cp;
  struct sockaddr_in sin;
  char buf[80];

  cp = Argv[0];
  size = sizeof(sin);
  if (getpeername(s, (struct sockaddr *)&sin, &size) == 0)
    sprintf(buf, "-%s [%s]", name, inet_ntoa(sin.sin_addr)); 
  else
    sprintf(buf, "-%s", name); 
  strncpy(cp, buf, LastArg - cp);
  cp += strlen(cp);
  while (cp < LastArg)
  *cp++ = ' ';
}
#else
#define setproctitle(x,y)
#endif

/*
 * Internet services provided internally by inetd:
 */
#define	BUFSIZE	8192

/* Echo service -- echo data back */

static void 
echo_stream(int s, struct servtab * sep) {
  char buffer[BUFSIZE];
  int i;

  s = 0;

  setproctitle(sep->port->service, 0);
  while ((i = read(0, buffer, sizeof(buffer))) > 0 &&
	  write(0, buffer, i) > 0)
    /*NOTHING*/;
  exit(0);
}

/* Echo service -- echo data back */

static void
echo_dg(int s, struct servtab * sep) {
  char buffer[BUFSIZE];
  int i, size;
  struct sockaddr sa;

  size = sizeof(sa);
  if ((i = recvfrom(s, buffer, sizeof(buffer), 0, &sa, &size)) < 0)
    return;
  sendto(s, buffer, i, 0, &sa, sizeof(sa));
}

/* Discard service -- ignore data */

static void
discard_stream(int s, struct servtab * sep) {
  int ret;
  char buffer[BUFSIZE];

  s = 0;     

  setproctitle(sep->port->service, s);
  while (1) {
    while ((ret = read(s, buffer, sizeof(buffer))) > 0)
      /*NOTHING*/;
  if (ret == 0 || errno != EINTR)
    break;
  }
  exit(0);
}

/* Discard service -- ignore data */

static void
discard_dg(int s, struct servtab * sep) {
  char buffer[BUFSIZE];

  (void) read(s, buffer, sizeof(buffer));
}

#include <ctype.h>

#define LINESIZ 72
static char ring[128];
static char *endring;

static void 
initring(void) {
  register int i;

  endring = ring;

  for (i = 0; i <= 128; ++i)
    if (isprint(i))
      *endring++ = i;
}

/* Character generator */

static void
chargen_stream(int s, struct servtab * sep) {
  register char *rs;
  int len;
  char text[LINESIZ+2];

  s = 0;

  setproctitle(sep->port->service, s);

  if (!endring) {
    initring();
    rs = ring;
  }

  text[LINESIZ] = '\r';
  text[LINESIZ + 1] = '\n';
  for (rs = ring;;) {
    if ((len = endring - rs) >= LINESIZ)
      bcopy(rs, text, LINESIZ);
    else {
      bcopy(rs, text, len);
      bcopy(ring, text + len, LINESIZ - len);
    }
    if (++rs == endring)
      rs = ring;
    if (write(s, text, sizeof(text)) != sizeof(text))
      break;
  }
  exit(0);
}

/* Character generator */

static 
void
chargen_dg(int s, struct servtab * sep) {
  struct sockaddr sa;
  static char *rs;
  int len, size;
  char text[LINESIZ+2];

  if (endring == 0) {
    initring();
    rs = ring;
  }

  size = sizeof(sa);
  if (recvfrom(s, text, sizeof(text), 0, &sa, &size) < 0)
    return;

  if ((len = endring - rs) >= LINESIZ)
    bcopy(rs, text, LINESIZ);
  else {
    bcopy(rs, text, len);
    bcopy(ring, text + len, LINESIZ - len);
  }
  if (++rs == endring)
    rs = ring;
  text[LINESIZ] = '\r';
  text[LINESIZ + 1] = '\n';
  sendto(s, text, sizeof(text), 0, &sa, sizeof(sa));
}

/*
 * Return a machine readable date and time, in the form of the
 * number of seconds since midnight, Jan 1, 1900.  Since gettimeofday
 * returns the number of seconds since midnight, Jan 1, 1970,
 * we must add 2208988800 seconds to this figure to make up for
 * some seventy years Bell Labs was asleep.
 */

static 
long
machtime(void) {
  struct timeval tv;

  if (gettimeofday(&tv, (struct timezone *)0) < 0) {
    fprintf(stderr, "Unable to get time of day\n");
    return (0L);
  }
  return (htonl((long)tv.tv_sec + 2208988800U));
}

static 
void
machtime_stream(int s, struct servtab * sep) {
  long result;
  result = machtime();
  write(s, (char *) &result, sizeof(result));
}

static 
void
machtime_dg(int s, struct servtab * sep) {
  long result;
  struct sockaddr sa;
  int size;

  size = sizeof(sa);
  if (recvfrom(s, (char *)&result, sizeof(result), 0, &sa, &size) < 0)
    return;
  result = machtime();
  sendto(s, (char *) &result, sizeof(result), 0, &sa, sizeof(sa));
}

/* Return human-readable time of day */

static 
void
daytime_stream(int s, struct servtab * sep) {
  char buffer[256];
  time_t time(), clock;

  clock = time((time_t *) 0);
  sprintf(buffer, "%.24s\r\n", ctime(&clock));
  write(s, buffer, strlen(buffer));
}

/* Return human-readable time of day */

static 
void
daytime_dg(int s, struct servtab * sep) {
  char buffer[256];
  time_t time(), clock;
  struct sockaddr sa;
  int size;

  clock = time((time_t *) 0);

  size = sizeof(sa);
  if (recvfrom(s, buffer, sizeof(buffer), 0, &sa, &size) < 0)
    return;
  sprintf(buffer, "%.24s\r\n", ctime(&clock));
  sendto(s, buffer, strlen(buffer), 0, &sa, sizeof(sa));
}

/*
struct icmphdr {
  unsigned char		type;
  unsigned char		code;
  unsigned short	checksum;
  union {
	struct {
		unsigned short	id;
		unsigned short	sequence;
	} echo;
	unsigned long gateway;
  } un;
};
*/

static 
void
broadcast_dg(int s, struct servtab * sep) {
  char buffer[256];
  struct sockaddr sa;
  int size;
/*
  int sock_raw;
  char packet[sizeof(struct icmphdr)+8];
  struct icmphdr * ic = (struct icmphdr *)&packet;
*/

  openlog("udpbrd", LOG_PID, LOG_DAEMON);
  size = sizeof(sa);
  if (recvfrom(s, buffer, sizeof(buffer), 0, &sa, &size) < 0)
    return;
  loginfo("UNKNOWN UDP REQUEST FROM %08x:%i",
          ntohl(((struct sockaddr_in*)(&sa))->sin_addr.s_addr),
          ntohs(((struct sockaddr_in*)(&sa))->sin_port)
  );
/*
  sock_raw = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP); 
  if (sock_raw < 0)  {
    logerr("opening raw socket: %m");
    return;
  }
  ic->type = ICMP_DEST_UNREACH;
  ic->code = ICMP_PORT_UNREACH;
  ic->checksum = 0;
  ic->un.gateway = 0;
  if (sendto(sock_raw, packet, sizeof(packet), 0, 
      &sa, sizeof(struct sockaddr_in)) < 0) {
    logerr("sending PORT_UNREACH: %m");
  }
  close(sock_raw);
*/
  return;
}

static void 
broadcast_stream(int s, struct servtab * sep) {
  struct sockaddr_in si;
  struct sockaddr_in peer;
  int size = sizeof(si);

  s = 0;

  openlog("tcpbrd", LOG_PID, LOG_DAEMON);
  setproctitle(sep->port->service, 0);
  if (getsockname(0, (struct sockaddr *)&si, &size)) {
    logerr("cannot getsockname : %m");
    shutdown(0,2);
    exit(1);
  }
  loginfo("UNKNOWN TCPREQUEST FOR %08x:%i",
          ntohl(((struct sockaddr_in*)(&si))->sin_addr.s_addr),
          ntohs(((struct sockaddr_in*)(&si))->sin_port));
  size = sizeof(peer);
  if (getpeername(0, (struct sockaddr *)&peer, &size)) {
    logerr("cannot getsockname : %m");
    shutdown(0,2);
    exit(1);
  }
  loginfo("                  FROM %08x:%i",
          ntohl(((struct sockaddr_in*)(&peer))->sin_addr.s_addr),
          ntohs(((struct sockaddr_in*)(&peer))->sin_port));
  shutdown(0,2);
  exit(0);
}



struct biltin biltins[] = {
{       "tcpbrd",       SOCK_STREAM,    1, 0,   broadcast_stream },
{       "udpbrd",       SOCK_DGRAM,     0, 1,   broadcast_dg     },
	/* Echo received data */
{	"echo",		SOCK_STREAM,	1, 0,	echo_stream,     },
{	"echo",		SOCK_DGRAM,	0, 1,	echo_dg,         },

	/* Internet /dev/null */
{	"discard",	SOCK_STREAM,	1, 0,	discard_stream,  },
{	"discard",	SOCK_DGRAM,	0, 1,	discard_dg,      },

	/* Return 32 bit time since 1970 */
{	"time",		SOCK_STREAM,	0, 0,	machtime_stream, },
{	"time",		SOCK_DGRAM,	0, 1,	machtime_dg,     },

	/* Return human-readable time */
{	"daytime",	SOCK_STREAM,	0, 0,	daytime_stream,  },
{	"daytime",	SOCK_DGRAM,	0, 1,	daytime_dg,      },

	/* Familiar character generator */
{	"chargen",	SOCK_STREAM,	1, 0,	chargen_stream,  },
{	"chargen",	SOCK_DGRAM,	0, 1,	chargen_dg,      },
{	NULL, }
};
