
#ifndef _INETD_H_
#define _INETD_H_ 1

#define _BLOCK_BEGIN {
#define _BLOCK_END   }


void cleanup(int code) __attribute__((noreturn));


inline static int
FDSET_OR(fd_set * f1, fd_set * f2) {
  int i;
  int maxi = -1;
  for (i = 0; i<OPEN_MAX ; i++) {
    if (FD_ISSET(i, f2)) {
      FD_SET(i, f1);
      maxi = i;
    }
  }
  return maxi;
}

inline static void
FDSET_NOTAND(fd_set * f1, fd_set * f2, int nr) {
  int i;
  for (i=(nr > 0 ? nr-1 : OPEN_MAX); i >= 0; i--) {
    if (FD_ISSET(i, f2))
      FD_CLR(i, f1);
  }
}

inline static void
FDSET_AND(fd_set * f1, fd_set * f2) {
  int i;
  for (i=0; i<OPEN_MAX; i++) {
    if (!FD_ISSET(i, f2))
      FD_CLR(i, f1);
  }
}


/* ports */

#include <sys/socket.h>
#include <sys/time.h>

struct  port {
  int	socktype;
  int	proto;
  struct pmaplist * pml;
  struct sockaddr addr;
  int	fd;
  struct  port *next, *prev;
  time_t  last_use;
  unsigned char accept   : 1,
	        debug    : 1,
                log      : 1,
                rfc931   : 1,
                autorpc  : 1,
                broadcast: 1,
                keepalive: 1,
	        reuse    : 1;
};

#define FAMILY(port)  ((port)->addr.sa_family)
#define IN_ADDR(port) (((struct sockaddr_in*)&((port)->addr))->sin_addr.s_addr)
#define IN_PORT(port) (((struct sockaddr_in*)&((port)->addr))->sin_port)

void add_to_port_list(struct port ** pport, struct port * port);
void free_port_list(struct port ** pport);

/* server tables */

#define	MAXARGV 16

struct	servtab {
  char	* user;		        /* user name to run as */
  char  * server;
  struct biltin * bi;		/* if built-in, description */
  char	* argv[MAXARGV+1];	/* program arguments */
  int     pid;
  int	  count;		        /* number started since se_time */
  struct  timeval start_time;
  struct  servtab * next, *prev;
  struct  port * port;
  unsigned char obsolete : 1,	
	        wait     : 1;
};

extern struct servtab * servtab;


/* Builtin server  definitions */

typedef void builtin_fn_t(int fd, struct servtab * sep);		

struct biltin {
  char	* servicename;		/* internally provided service name */
  int	  socktype;		/* type of socket supported */
  unsigned char forking : 1,
	        wait    : 1;
  builtin_fn_t * bi_fn;		/* function which performs it */
};

extern struct biltin biltins[];


/* inetd.conf.c */

struct servtab * getconfigent(void);

int  setconfig(char *);
void endconfig(void);
void discard_servtab(struct servtab * sep);
void set_port_options(struct servtab * sep, struct port * port);
void activate_fd(int fd);
struct pmaplist * find_rpc_service(u_long, u_long, u_long, 
	     struct servtab **,
	     struct port **portp);
void shutdown_portmap(void);
int  resetup_portmap(struct servtab *);
int  reap_rpc_callit(pid_t);
int  socktypestr2num(const char *);
int  protostr2num(const char *);
int  portstr2num(const char *, int);
int  bind_rpc_service(struct servtab * sep);

#define logerr(fmt...) ({ syslog(LOG_ERR, fmt); /* if (!debug) debug++; */ })
#define loginfo(fmt...) syslog(LOG_INFO, fmt)

#ifdef DEBUG
extern int debug;

#define dprintf(level, fmt...) ({ if ((level) <= debug) \
				    syslog(LOG_DEBUG, fmt); })

#else
#define dprintf(level, fmt...) (0)
#endif



#endif /*_INETD_H_*/
