#include <stdio.h>
#include <malloc.h>
#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <limits.h>
#include <rpc/pmap_clnt.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <syslog.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/stat.h>
#include <rpc/svc_soc.h>

#include "inetd.h"

#undef DO_3_4

#ifdef DO_3_4
#define	RPCBPROG ((unsigned long)(100000))
#define	RPCBVERS ((unsigned long)(3))

#define	RPCBPROC_NULL ((unsigned long)(0))
#define	RPCBPROC_SET ((unsigned long)(1))
#define	RPCBPROC_UNSET ((unsigned long)(2))
#define	RPCBPROC_GETADDR ((unsigned long)(3))
#define	RPCBPROC_DUMP ((unsigned long)(4))
#define	RPCBPROC_CALLIT ((unsigned long)(5))
#define	RPCBPROC_GETTIME ((unsigned long)(6))
#define	RPCBPROC_UADDR2TADDR ((unsigned long)(7))
#define	RPCBPROC_TADDR2UADDR ((unsigned long)(8))

#define	RPCBVERS4 ((unsigned long)(4))

#define	RPCBPROC_BCAST ((unsigned long)(RPCBPROC_CALLIT))
#define	RPCBPROC_GETVERSADDR ((unsigned long)(9))
#define	RPCBPROC_INDIRECT ((unsigned long)(10))
#define	RPCBPROC_GETADDRLIST ((unsigned long)(11))
#define	RPCBPROC_GETSTAT ((unsigned long)(12))
#endif

static void reg_service(struct svc_req*, SVCXPRT*);
#ifdef DO_3_4
static void reg_service_3(struct svc_req*, SVCXPRT*);
#endif
static void pmapproc_call(struct svc_req*, SVCXPRT*);

#define MAX_CALLITS 16

static pid_t callit_childs[MAX_CALLITS];
static int   callit_cnt = 0;

struct pmaplist * pm_list = NULL;

struct pmaplist *
find_rpc_service(u_long prog, u_long vers, u_long prot, 
	     struct servtab ** sepp,
	     struct port **portp) {
	struct pmaplist *pml = NULL;
	struct servtab  *sep;
	struct port     *port;
	struct pmaplist *fnd = NULL;
	u_long highest_vers = 0;

	if (sepp)
	  *sepp = NULL;
	if (portp)
	  *portp = NULL;

	for (sep = servtab; sep; sep=sep->next)
	  for (port = sep->port; port; port=port->next)
	    for (pml = port->pml; pml; pml = pml->pml_next)
		if ((pml->pml_map.pm_prog == prog) &&
		    (pml->pml_map.pm_prot == prot)) {
		      if (pml->pml_map.pm_vers >= highest_vers) {
		        highest_vers = pml->pml_map.pm_vers;
		        fnd = pml;
	                if (sepp)
	                  *sepp = sep;
	                if (portp)
	                  *portp = port;
		      }
		      if (pml->pml_map.pm_vers == vers)
		        return pml;
		  }
	return fnd;
}

static struct pmaplist *
register_rpc_service_1(struct pmap * p) {
  struct pmaplist * pml, *fnd;
  pml = (struct pmaplist *)malloc((u_int)sizeof(struct pmaplist));
  pml->pml_map  = *p;
  pml->pml_next = 0;
  if (!pm_list)
    pm_list = pml;
  else {
      for (fnd= pm_list; fnd->pml_next; fnd = fnd->pml_next);
      fnd->pml_next = pml;
      fnd = pml;
  }
  return pml;
}


static struct pmaplist *
find_rpc_service_1(u_long prog, u_long vers, u_long prot) {
	struct pmaplist *pml = NULL;
	struct pmaplist *fnd = NULL;
	u_long highest_vers = 0;

        for (pml = pm_list; pml; pml = pml->pml_next)
		if ((pml->pml_map.pm_prog == prog) &&
		    (pml->pml_map.pm_prot == prot)) {
		      if (pml->pml_map.pm_vers >= highest_vers) {
		        highest_vers = pml->pml_map.pm_vers;
		        fnd = pml;
		      }
		      if (pml->pml_map.pm_vers == vers)
		        return pml;
		}
	return fnd;
}



static int
unregister_rpc_service_1(u_long prog, u_long vers, u_long prot) {
  struct pmaplist *pml, *prevpml;
  void * t;
  int ans = 0;

  for (prevpml = NULL, pml = pm_list; pml; ) {
    if ((pml->pml_map.pm_prog != prog) ||
  	(pml->pml_map.pm_vers != vers)) {
  	/* both pml & prevpml move forwards */
      prevpml = pml;
      pml = pml->pml_next;
      continue;
    }
    /* found it; pml moves forward, prevpml stays */
    ans = 1;
    t = (caddr_t)pml;
    pml = pml->pml_next;
    if (prevpml == NULL)
      pm_list = pml;
    else
      prevpml->pml_next = pml;
    free(t);
  }
  return ans;
}


SVCXPRT * udp_transp = NULL;
SVCXPRT * tcp_transp = NULL;

void
shutdown_portmap() {
  if (udp_transp)
    svc_destroy(udp_transp);
  if (tcp_transp)
    svc_destroy(tcp_transp);
  udp_transp = tcp_transp = NULL;
}


int
resetup_portmap(struct servtab * sep) {
  SVCXPRT * transp;
  struct port * port = sep->port;

#ifdef PARANOID
  if (!port || !port->pml ||
    port->pml->pml_map.pm_prog != PMAPPROG) {
    logerr("par200");
    cleanup(1);
  }
#endif

  if (udp_transp || tcp_transp) {
    logerr("portmap already running?");
    discard_servtab(sep);
    return -1;
  }

  for (; port; port = port->next) {
    int proto = 0;
    transp = NULL;
    if (port->socktype == SOCK_DGRAM && !udp_transp) {
      transp = udp_transp = svcudp_create(port->fd);
      proto = IPPROTO_UDP; 
    } else if (port->socktype == SOCK_STREAM && !tcp_transp) {
      transp = tcp_transp = svctcp_create(port->fd, RPCSMALLMSGSIZE, RPCSMALLMSGSIZE);
      proto = IPPROTO_TCP; 
    }
    if (!transp) {
      logerr("pmap server on %i not created", ntohs(IN_PORT(port)));
      continue;
    }
    if (!svc_register(transp, PMAPPROG, PMAPVERS, reg_service, FALSE)) {
      logerr("unable to register(PMAPPROG, PMAPVERS, %d)", port->proto);
      continue;
    }
#ifdef DO_3_4
    if (!svc_register(transp, RPCBPROG, RPCBVERS, reg_service_3, FALSE)) {
      logerr("unable to register(RPCBPROG, RPCBVERS, %d)", port->proto);
      continue;
    }
#endif
  }
  return 0;
}



static void 
pmapproc_null(struct svc_req * rqstp, SVCXPRT * xprt) {
  dprintf(1,"NULL(void)");
  if (!svc_sendreply(xprt, xdr_void, NULL)) 
    logerr("pmapproc_null: svc_sendreply failure");
}

static void 
pmapproc_set(struct svc_req * rqstp, SVCXPRT * xprt)
{
  struct pmap reg;
  struct pmaplist * fnd;
  struct servtab  * sep;
  int ans = 0;
  struct sockaddr_in *sin = svc_getcaller(rqstp->rq_xprt);

  if (!svc_getargs(xprt, xdr_pmap, &reg)) {
    logerr("pmapproc_set: svc_getargs failure");
    svcerr_decode(xprt);
    return;
  }

  dprintf(1,"SET(%i v%i, port=(%i,%i))",
            reg.pm_prog, reg.pm_vers, 
            reg.pm_port, reg.pm_prot);

  if (sin->sin_addr.s_addr != htonl (INADDR_LOOPBACK)) {
	  logerr("attempt of non-local SET from %08x", sin->sin_addr.s_addr);
	  ans = 0;
	  goto fail;
  }
	if (ntohs(sin->sin_port) >= 1024) {
		logerr("attempt of SET from %08x:%04x", sin->sin_addr.s_addr, sin->sin_port);
		ans = 0;
		goto fail;
	}


  fnd = find_rpc_service(reg.pm_prog, reg.pm_vers, reg.pm_prot, &sep, NULL);
  if (!fnd) {
    fnd = find_rpc_service_1(reg.pm_prog, reg.pm_vers, reg.pm_prot);
    if (!fnd || fnd->pml_map.pm_vers != reg.pm_vers) {
      register_rpc_service_1(&reg);
      ans = 1;
    }
  } else if (fnd->pml_map.pm_vers != reg.pm_vers)
    logerr("rpc %d(%d)/%d @%d version mismatch %d", reg.pm_prog, reg.pm_vers,
	   reg.pm_prot, reg.pm_port, fnd->pml_map.pm_vers);
  else if (!sep->pid)
    logerr("rpc %d(%d)/%d @%d not running", reg.pm_prog, reg.pm_vers,
	   reg.pm_prot, reg.pm_port);
  else if (fnd->pml_map.pm_port != reg.pm_port)
    logerr("rpc %d(%d)/%d @%d port mismatch %d", reg.pm_prog, reg.pm_vers,
	   reg.pm_prot, reg.pm_port, fnd->pml_map.pm_port);
  else
      ans = 1;
  
fail:
  if (!svc_sendreply(xprt, xdr_long, (caddr_t)&ans)) {
    logerr("pmapproc_set: svc_sendreply failure");
    return;
  }
  dprintf(1,"ans=%i",ans);
}

static void 
pmapproc_unset(struct svc_req * rqstp, SVCXPRT * xprt)
{
  struct pmap reg;
  struct pmaplist *pml;
  struct servtab * sep;
  int ans = 0;
  struct sockaddr_in *sin = svc_getcaller(rqstp->rq_xprt);

  if (!svc_getargs(xprt, xdr_pmap, &reg)) {
    logerr("pmapproc_unset: svc_getargs failure");
    svcerr_decode(xprt);
    return;
  }

  dprintf(1,"UNSET(%i v%i)", reg.pm_prog, reg.pm_vers);

  if (sin->sin_addr.s_addr != htonl (INADDR_LOOPBACK)) {
	  logerr("attempt of non-local UNSET from %08x", sin->sin_addr.s_addr);
	  ans = 0;
	  goto fail;
  }
  if (ntohs(sin->sin_port) >= 1024) {
	  logerr("attempt of SET from %08x:%04x", sin->sin_addr.s_addr, sin->sin_port);
	  ans = 0;
	  goto fail;
  }

  pml = find_rpc_service(reg.pm_prog, reg.pm_vers, IPPROTO_UDP, &sep, NULL);
  if (pml) {
    if (sep->pid)
      ans = 1;
  } else
    ans = ans || unregister_rpc_service_1(reg.pm_prog, reg.pm_vers, IPPROTO_UDP);

  pml = find_rpc_service(reg.pm_prog, reg.pm_vers, IPPROTO_TCP, &sep, NULL);
  if (pml) {
    if (sep->pid)
      ans = 1;
  } else
    ans = ans || unregister_rpc_service_1(reg.pm_prog, reg.pm_vers, IPPROTO_TCP);

fail:
  if (!svc_sendreply(xprt, xdr_long, (caddr_t)&ans)) {
    logerr("pmapproc_unset: svc_sendreply failure");
    return;
  }
  dprintf(1,"ans=%i",ans);
}

static void 
pmapproc_get(struct svc_req * rqstp, SVCXPRT * xprt)
{
  struct pmap reg;
  struct pmaplist *fnd;
  struct servtab  *sep;
  u_long ans = 0;

  if (!svc_getargs(xprt, xdr_pmap, &reg)) {
	  logerr("pmapproc_get: svc_getargs failure");
	  svcerr_decode(xprt);
	  return;
  }

  fnd = find_rpc_service(reg.pm_prog, reg.pm_vers, reg.pm_prot, &sep, NULL);

  if (fnd) {
    if (sep)
      bind_rpc_service(sep);
    ans = fnd->pml_map.pm_port;
  } else {
    fnd = find_rpc_service_1(reg.pm_prog, reg.pm_vers, reg.pm_prot);
    if (fnd)
      ans = fnd->pml_map.pm_port;
  }

  dprintf(1,"GETPORT(%i v%i, %i)", reg.pm_prog, reg.pm_vers, reg.pm_prot);

  if (!svc_sendreply(xprt, xdr_long, (caddr_t)&ans)) {
    logerr("pmapproc_get: svc_sendreply");
    return;
  }
  dprintf(1,"ans=%i", ans);
}

static void 
pmapproc_dump(struct svc_req * rqstp, SVCXPRT * xprt) {
  time_t curtime;
  time(&curtime);

  if (!svc_getargs(xprt, xdr_void, NULL)) {
    logerr("pmapproc_dump: svc_getargs failure");
    svcerr_decode(xprt);
    return;
  }

  dprintf(1,"DUMP(void)");

  _BLOCK_BEGIN
  struct servtab * sep;
  struct port    * port;
  struct pmaplist * pml;
  int cnt = 0;

  for (sep = servtab; sep; sep=sep->next)
    for (port = sep->port; port; port=port->next)
      for (pml = port->pml; pml; pml = pml->pml_next)
        cnt++;

  _BLOCK_BEGIN
  struct pmaplist dump[cnt];
  struct pmaplist * dumplist;
  dumplist = (--cnt < 0 ? NULL : &dump[cnt]);
  for (sep = servtab; sep; sep=sep->next)
    for (port = sep->port; port; port=port->next)
      for (pml = port->pml; pml; pml = pml->pml_next) {
        if (cnt < 0)
	  goto done;
	port->last_use = curtime;
	dump[cnt] = *pml;
	dump[cnt].pml_next = &dump[cnt-1];
	cnt--;
      }
done:
  dump[cnt+1].pml_next = pm_list;
  if (!svc_sendreply(xprt, xdr_pmaplist, (caddr_t)&dumplist)) {
    logerr("pmapproc_dump: svc_sendreply");
    return;
  }
  _BLOCK_END
  _BLOCK_END
}

/* 
 * 1 OK, 0 not
 */
void
reg_service(struct svc_req * rqstp, SVCXPRT * xprt)
{
  switch (rqstp->rq_proc) {
  case PMAPPROC_NULL:
	pmapproc_null(rqstp, xprt);  
	break;  

  case PMAPPROC_SET:
	pmapproc_set(rqstp, xprt);  
	break;  

  case PMAPPROC_UNSET:
	pmapproc_unset(rqstp, xprt);  
	break;  

  case PMAPPROC_GETPORT:
	pmapproc_get(rqstp, xprt);
        break;

  case PMAPPROC_DUMP:
	pmapproc_dump(rqstp, xprt);
        break;

  case PMAPPROC_CALLIT:
	pmapproc_call(rqstp, xprt);
	break;

  default:
	logerr("pmapproc_[%i] called.", rqstp->rq_proc);
	svcerr_noproc(xprt);
	break;
  }
}

#ifdef DO_3_4

struct rpcb {
	u_long r_prog;
	u_long r_vers;
	char *r_netid;
	char *r_addr;
	char *r_owner;
};
typedef struct rpcb rpcb;

typedef rpcb RPCB;


bool_t
xdr_rpcb(XDR *xdrs, RPCB *objp)
{
	if (!xdr_u_long(xdrs, &objp->r_prog))
		return (FALSE);
	if (!xdr_u_long(xdrs, &objp->r_vers))
		return (FALSE);
	if (!xdr_string(xdrs, &objp->r_netid, ~0))
		return (FALSE);
	if (!xdr_string(xdrs, &objp->r_addr, ~0))
		return (FALSE);
	if (!xdr_string(xdrs, &objp->r_owner, ~0))
		return (FALSE);
	return (TRUE);
}


static u_long *
rpcbproc_gettime_com()
{
	static time_t curtime;

	(void) time(&curtime);
	return ((u_long *)&curtime);
}

static char *
mergeaddr(unsigned long port, char *r_addr)
{
	static char addrb[32];
	if (r_addr && r_addr[0])
		strcpy(addrb, r_addr);
	else
		strcpy(addrb, "0.0.0.0.");
	sprintf (addrb + strlen(addrb), "%d.%d", (port>>8)&0xFF, port&0xFF);
	return (char*)addrb;
}

static char **
rpcbproc_getaddr_3(RPCB *r, struct svc_req *rqstp, SVCXPRT *transp)
{
	unsigned long prot;
	struct pmaplist *fnd;
	struct servtab  *sep;
	u_long ans = 0;
	static char * uaddr;
	int socktype; 
	int ssize = sizeof (socktype);

	/*dprintf(1,*/ logerr( "GETADDR_3(%i v%i, %s, %s)", r->r_prog, r->r_vers,
		r->r_netid ?: "null", r->r_addr ?: "null");

#if 0
	if (uaddr && uaddr[0])
		free(uaddr);
#endif

	uaddr = "";

	if (getsockopt(transp->xp_sock, SOL_SOCKET, SO_TYPE,
		       (char *) &socktype, &ssize) < 0)
		return &uaddr;
	
	if (socktype == SOCK_DGRAM)
		prot = IPPROTO_UDP;
	else if (socktype == SOCK_STREAM)
		prot = IPPROTO_TCP;
	else
		return &uaddr;

	fnd = find_rpc_service(r->r_prog, r->r_vers, prot, &sep, NULL);

	if (fnd) {
		if (sep)
			bind_rpc_service(sep);
		ans = fnd->pml_map.pm_port;
	} else {
		fnd = find_rpc_service_1(r->r_prog, r->r_vers, prot);
		if (fnd)
			ans = fnd->pml_map.pm_port;
	}

	uaddr = mergeaddr(ans, r->r_addr);

	return &uaddr;
}


/* 
 * 1 OK, 0 not
 */
void
reg_service_3(struct svc_req * rqstp, SVCXPRT * xprt)
{
	union {
#if 0
		RPCB rpcbproc_set_3_arg;
		RPCB rpcbproc_unset_3_arg;
#endif
		RPCB rpcbproc_getaddr_3_arg;
#if 0
		struct rpcb_rmtcallargs rpcbproc_callit_3_arg;
		char *rpcbproc_uaddr2taddr_3_arg;
		struct netbuf rpcbproc_taddr2uaddr_3_arg;
#endif
	} argument;
	char *result;
	bool_t (*xdr_argument)(), (*xdr_result)();
	char *(*local)();

	switch (rqstp->rq_proc) {

	case RPCBPROC_NULL:
		dprintf(1,"RPCBPROC_NULL(void)");
		if (!svc_sendreply(xprt, xdr_void, NULL)) 
			logerr("rpcbproc_null: svc_sendreply failure");
		return;

	case RPCBPROC_GETADDR:
		logerr("RPCBPROC_GETADDR");
		xdr_argument = xdr_rpcb;
		xdr_result = xdr_wrapstring;
		local = (char *(*)()) rpcbproc_getaddr_3;
		break;

        case RPCBPROC_GETTIME:
		logerr("RPCBPROC_GETTIME");
		xdr_argument = xdr_void;
		xdr_result = xdr_u_long;
		local = (char *(*)()) rpcbproc_gettime_com;
		break;

	default:
		logerr("rpcbproc_[%i] v3 called.", rqstp->rq_proc);
		svcerr_progvers(xprt, PMAPVERS, PMAPVERS);
		return;
	}

	(void) memset((char *)&argument, 0, sizeof (argument));
	if (!svc_getargs(xprt, (xdrproc_t) xdr_argument, (char *) &argument)) {
		svcerr_decode(xprt);
		logerr("rpcbind: could not decode\n");
		return;
	}
	result = (*local)(&argument, rqstp, xprt, RPCBVERS);
	if (result != NULL && !svc_sendreply(xprt, (xdrproc_t)xdr_result,
					     result)) {
		svcerr_systemerr(xprt);
		logerr("rpcbind: svc_sendreply\n");
	}
	if (!svc_freeargs(xprt, (xdrproc_t)xdr_argument, (char *)
			  &argument)) {
		logerr("rpcbind: unable to free arguments\n");
	}
}
#endif


/*
 * Stuff for the rmtcall service
 */
#define ARGSIZE 9000

struct encap_parms {
	u_int arglen;
	char *args;
};

static bool_t
xdr_encap_parms(XDR * xdrs, struct encap_parms * epp) {

	return (xdr_bytes(xdrs, &(epp->args), &(epp->arglen), ARGSIZE));
}

struct my_rmtcallargs {
	u_long	rmt_prog;
	u_long	rmt_vers;
	u_long	rmt_port;
	u_long	rmt_proc;
	struct encap_parms rmt_args;
};

static bool_t
xdr_rmtcall_args(XDR * xdrs, struct my_rmtcallargs * cap) {
	/* does not get a port number */
  if (xdr_u_long(xdrs, &(cap->rmt_prog)) &&
      xdr_u_long(xdrs, &(cap->rmt_vers)) &&
      xdr_u_long(xdrs, &(cap->rmt_proc))) 
    return (xdr_encap_parms(xdrs, &(cap->rmt_args)));
  return (FALSE);
}

static bool_t
xdr_rmtcall_result(XDR * xdrs, struct my_rmtcallargs * cap) {
  if (xdr_u_long(xdrs, &(cap->rmt_port)))
    return (xdr_encap_parms(xdrs, &(cap->rmt_args)));
  return (FALSE);
}

/*
 * only worries about the struct encap_parms part of struct my_rmtcallargs.
 * The arglen must already be set!!
 */
static bool_t
xdr_opaque_parms(XDR * xdrs, struct my_rmtcallargs * cap) {
  return (xdr_opaque(xdrs, cap->rmt_args.args, cap->rmt_args.arglen));
}

/*
 * This routine finds and sets the length of incoming opaque paraters
 * and then calls xdr_opaque_parms.
 */
static bool_t
xdr_len_opaque_parms(XDR * xdrs, struct my_rmtcallargs * cap) {
  register u_int beginpos, lowpos, highpos, currpos, pos;

  beginpos = lowpos = pos = xdr_getpos(xdrs);
  highpos = lowpos + ARGSIZE;
  while ((int)(highpos - lowpos) >= 0) {
    currpos = (lowpos + highpos) / 2;
    if (xdr_setpos(xdrs, currpos)) {
      pos = currpos;
      lowpos = currpos + 1;
    } else 
      highpos = currpos - 1;
  }
  xdr_setpos(xdrs, beginpos);
  cap->rmt_args.arglen = pos - beginpos;
  return (xdr_opaque_parms(xdrs, cap));
}

/*
 * Call a remote procedure service
 * This procedure is very quiet when things go wrong.
 * The proc is written to support broadcast rpc.  In the broadcast case,
 * a machine should shut-up instead of complain, less the requestor be
 * overrun with complaints at the expense of not hearing a valid reply ...
 *
 * This now forks so that the program & process that it calls can call 
 * back to the portmapper.
 */

static
void
pmapproc_call(struct svc_req * rqstp, SVCXPRT *xprt) {
  struct my_rmtcallargs a;
  struct pmaplist *pml;
  struct servtab  *sep;
  u_short port;
  u_long  vers;
  struct sockaddr_in me;
  int pid, socket = -1;
  CLIENT *client;
  struct authunix_parms *au = (struct authunix_parms *)rqstp->rq_clntcred;
  struct timeval timeout;
  char buf[ARGSIZE];
  int  res;
  int  slot;

  timeout.tv_sec = 5;
  timeout.tv_usec = 0;
  a.rmt_args.args = buf;
  if (!svc_getargs(xprt, xdr_rmtcall_args, &a)) {
    logerr("pmapproc_call: svc_getargs failure");
    svcerr_decode(xprt);
    return;
  }

  dprintf(1,"CALLIT(%i v%i, proc=%i)",
            a.rmt_prog, a.rmt_vers, a.rmt_proc);

  sep = NULL;

  if ((pml = find_rpc_service(a.rmt_prog, a.rmt_vers, IPPROTO_UDP, &sep, NULL)) == NULL) {
    if ((pml = find_rpc_service_1(a.rmt_prog, a.rmt_vers, IPPROTO_UDP)) == NULL) {
      dprintf(1,"not registered");
      return;
    }
  }

  if (sep)
    bind_rpc_service(sep);

  port = pml->pml_map.pm_port;
  vers = pml->pml_map.pm_vers;

  if (!port) {
    logerr("callit: port not bound");
    return;
  }

  for (slot = MAX_CALLITS-1; slot >= 0; slot--)
    if (!callit_childs[slot])
      break;

  if (slot<0) {
    logerr("no more callit slots");
    return;
  }

  pid = fork();

  if (pid < 0) {
    logerr("fork: %m");
    return;
  }

  if (pid) {
    callit_childs[slot] = pid;
    callit_cnt++;
    return;
  }

  setuid((uid_t)-2);
  setgid((gid_t)-2);
  setgroups(0, NULL);

  closelog();
  openlog("pmap_callit",LOG_PID,LOG_DAEMON);

  get_myaddress(&me);
  me.sin_port = htons(port);

  client = clntudp_create(&me, a.rmt_prog, a.rmt_vers, timeout, &socket);

  if (client) {
    if (rqstp->rq_cred.oa_flavor == AUTH_UNIX) {
  	client->cl_auth = authunix_create(au->aup_machname,
  		   au->aup_uid, au->aup_gid, au->aup_len, au->aup_gids);
    }
    a.rmt_port = (u_long)port;
    if ((res=clnt_call(client, a.rmt_proc, xdr_opaque_parms, &a,
  	    xdr_len_opaque_parms, &a, timeout)) == RPC_SUCCESS) {
  	if (!svc_sendreply(xprt, xdr_rmtcall_result, (caddr_t)&a)) {
          logerr("proc_call: svc_sendreply failure");
        }
    } else
      logerr("%s",clnt_sperror(client, "clnt_call"));

    AUTH_DESTROY(client->cl_auth);
    clnt_destroy(client);
  } else
    logerr("%s",clnt_spcreateerror("clntudp_create"));
  (void)close(socket);
  exit(0);
}

int
reap_rpc_callit(pid_t pid) {
  int slot;

  for (slot = MAX_CALLITS-1; slot >=0 ; slot--) {
	if (callit_childs[slot] == pid) {
          callit_childs[slot] = 0;
	  callit_cnt--;
	  return 1;
        }
  }
  return 0;
}
