/* Get public or secret key from key server.
   Copyright (C) 1996 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Ulrich Drepper <drepper@cygnus.com>, 1996.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <rpc.govno/netdb.h>
/*
   NOTE!!! rpc.govno contains rpc/ include files distributed
   with glibc-1.99. We do not use them, but rpc/netdb.h is still
   necessary.

   Please, make cd /usr/include and mv rpc rpc.govno; mv rpcsvc rpcsvc.govno
   now. You will need them anymore. BTW "govno" is shit in Russian 8)
 */
#include <rpc/key_prot.h>
#include <nsswitch.h>


/* Type of the lookup function for the public key.  */
typedef int (*publicsecret_function) (const char *, char **, char **);

/* The lookup function for the first entry of this service.  */
extern int __nss_publickey_lookup (service_user **nip, const char *name,
				   void **fctp);

int
getpublicsecretkey (const char *name, char **pkeyp, char **skeyp)
{
  static service_user *startp = NULL;
  static publicsecret_function start_fct;
  service_user *nip;
  publicsecret_function fct;
  enum nss_status status = NSS_STATUS_UNAVAIL;
  int no_more;

  if (startp == NULL)
    {
      no_more = __nss_publickey_lookup (&nip, "getpublicsecretkey", (void **) &fct);
      if (no_more)
	startp = (service_user *) -1;
      else
	{
	  startp = nip;
	  start_fct = fct;
	}
    }
  else
    {
      fct = start_fct;
      no_more = (nip = startp) == (service_user *) -1;
    }

  while (! no_more)
    {
      status = (*fct) (name, pkeyp, skeyp);

      no_more = __nss_next (&nip, "getpublicsecretkey", (void **) &fct, status, 0);
    }

  return status == NSS_STATUS_SUCCESS;
}

int
getpublickey_long (const char *name, char **keyp)
{
  return getpublicsecretkey(name, keyp, NULL);
}


#if 0

/* This function is used only by key adminstration programs;
   it is non-sense to put it into public place. --ANK
 */

int
getsecretkey(const char * netname, char * secretkey, const char * passwd)
{
	char * skey = NULL;

	if (!getsecretkey_long(netname, &skey, passwd))
		return 0;

	if (!skey) {
		secretkey[0] = 0;
		return 1;
	}

	if (strlen(skey) != DEFAULT_HEXKEYBYTES) {
		free(skey);
		secretkey[0] = 0;
		return 1;
	}

	strcpy(secretkey, skey);
	free(skey);
	return 1;
}
#endif

int
getpublickey(const char * netname, char * publickey)
{
	char * pkey = NULL;

	if (!getpublickey_long(netname, &pkey))
		return 0;

	if (!pkey)
		return 0;

	strncpy(publickey, pkey, DEFAULT_HEXKEYBYTES);
	publickey[DEFAULT_HEXKEYBYTES] = 0;
	free(pkey);
	return 1;
}


int
secretkey_decrypt(const char *netname, char *p, const char * passwd)
{
  int klen;

  klen = strlen(p) - KEYCHECKSUMSIZE;
  if (klen <= 0)
    return(0);
  if (!xxdecrypt(p, passwd, netname))
    return(0);

  if (bcmp(p, p + klen, KEYCHECKSUMSIZE) != 0) {
    p[0] = 0;
    return(1);
  }
  p[klen] = 0;
  return(1);
}
