/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
/*
 * Copyright (c) 1986-1991 by Sun Microsystems Inc. 
 */

#ident	"@(#)authsys_prot.c	1.10	93/07/05 SMI"

#if !defined(lint) && defined(SCCSIDS)
static char sccsid[] = "@(#)authsys_prot.c 1.24 89/02/07 Copyr 1984 Sun Micro";
#endif

/*
 * authsys_prot.c
 * XDR for UNIX style authentication parameters for RPC
 *
 * Copyright (C) 1984, Sun Microsystems, Inc.
 */

#ifdef KERNEL
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/user.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#endif

#include <rpc/types.h>
#include <rpc/trace.h>
#include <rpc/xdr.h>
#include <rpc/auth.h>
#include <rpc/auth_sys.h>

/*
 * XDR for unix authentication parameters.
 */
bool_t
xdr_authsys_parms(xdrs, p)
	register XDR *xdrs;
	register struct authsys_parms *p;
{
	trace1(TR_xdr_authsys_parms, 0);
	if (xdr_u_long(xdrs, &(p->aup_time)) &&
	    xdr_string(xdrs, &(p->aup_machname), MAX_MACHINE_NAME) &&
	    xdr_int(xdrs, &(p->aup_uid)) &&
	    xdr_int(xdrs, &(p->aup_gid)) &&
	    xdr_array(xdrs, (caddr_t *)&(p->aup_gids),
			&(p->aup_len), NGRPS, sizeof (int),
			(xdrproc_t) xdr_int)) {
		trace1(TR_xdr_authsys_parms, 1);
		return (TRUE);
	}
	trace1(TR_xdr_authsys_parms, 1);
	return (FALSE);
}

#if 0
/*
 * XDR user id types (uid_t)
 */
bool_t
xdr_uid_t(xdrs, ip)
	XDR *xdrs;
	int *ip;
{
	bool_t dummy;

	trace1(TR_xdr_uid_t, 0);
	if (sizeof (uid_t) == sizeof (long)) {
		dummy = xdr_long(xdrs, (long *)ip);
		trace1(TR_xdr_uid_t, 1);
		return (dummy);
	} else {
		dummy = xdr_short(xdrs, (short *)ip);
		trace1(TR_xdr_uid_t, 1);
		return (dummy);
	}
}

/*
 * XDR group id types (gid_t)
 */
bool_t
xdr_gid_t(xdrs, ip)
	XDR *xdrs;
	int *ip;
{
	bool_t dummy;

	trace1(TR_xdr_gid_t, 0);
	if (sizeof (gid_t) == sizeof (long)) {
		dummy = xdr_long(xdrs, (long *)ip);
		trace1(TR_xdr_gid_t, 1);
		return (dummy);
	} else {
		dummy = xdr_short(xdrs, (short *)ip);
		trace1(TR_xdr_gid_t, 1);
		return (dummy);
	}
}
#endif

#ifdef KERNEL
/*
 * XDR kernel unix auth parameters.
 * Goes out of the u struct directly.
 * NOTE: this is an XDR_ENCODE only routine.
 */
xdr_authkern(xdrs)
	register XDR *xdrs;
{
	int	*gp;
	uid_t	uid = getuid();
	gid_t	gid = getgid();
	int	len;
	caddr_t	groups;
	char	*name = hostname;

	trace1(TR_xdr_authkern, 0);
	if (xdrs->x_op != XDR_ENCODE) {
		trace1(TR_xdr_authkern, 1);
		return (FALSE);
	}

	for (gp = &u.u_groups[NGROUPS]; gp > u.u_groups; gp--) {
		if (gp[-1] >= 0) {
			break;
		}
	}
	len = gp - u.u_groups;
	groups = (caddr_t)u.u_groups;
	if (xdr_u_long(xdrs, (u_long *)&time.tv_sec) &&
	    xdr_string(xdrs, &name, MAX_MACHINE_NAME) &&
	    xdr_uid_t(xdrs, &uid) &&
	    xdr_gid_t(xdrs, &gid) &&
	    xdr_array(xdrs, &groups, (u_int *)&len, NGRPS,
			sizeof (gid_t), xdr_gid_t)) {
		trace1(TR_xdr_authkern, 1);
		return (TRUE);
	}
	trace1(TR_xdr_authkern, 1);
	return (FALSE);
}
#endif
