/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
/*
 * Copyright (c) 1986-1991 by Sun Microsystems Inc. 
 */

#ident	"@(#)netnamer.c	1.19	94/04/24 SMI"
/*
 * Copyright (c) 1986-1992 by Sun Microsystems Inc.
 */

#if !defined(lint) && defined(SCCSIDS)
static char sccsid[] = "@(#)netnamer.c 1.4 89/03/20 Copyr 1986 Sun Micro";
#endif

/*
 * netname utility routines convert from netnames to unix names (uid, gid)
 *
 * This module is operating system dependent!
 * What we define here will work with any unix system that has adopted
 * the Sun NIS domain architecture.
 */

#undef NIS
#include "rpc_mt.h"
#include <stdio.h>
#include <sys/types.h>
#include <rpc/trace.h>
#include <ctype.h>
#include <grp.h>
#include <pwd.h>
#include <string.h>
#include <syslog.h>
#include <sys/param.h>
#include <nsswitch.h>
#include <rpc/rpc.h>


static const char    OPSYS[]	= "unix";
#define	OPSYS_LEN 4

struct netid_userdata {
	uid_t	*uidp;
	gid_t	*gidp;
	int	*gidlenp;
	gid_t	*gidlist;
};

/* parse a comma separated gid list */
static int
parse_gidlist(char *p, struct netid_userdata *argp)
{
	int len;
	unsigned long g;
	char *ptr;

	if (!p)
	  return -1;

	g = strtoul(p, &ptr, 10);
	if (g==0 && ptr == p)
	  return -1;
	if (ptr == NULL || (*ptr && *ptr != ','))
	  return -1;
	*(argp->gidp) = g;

	len = 0;
	while (*ptr != 0) {
	  g = strtoul(ptr, &ptr, 10);
	  if (g==0 && ptr == p)
	    return -1;
	  if (ptr == NULL || (*ptr && *ptr != ','))
	    return -1;
	  argp->gidlist[len++] = g;
	}
	*(argp->gidlenp) = len;
	return 0;
}



typedef int (*netid_function) (const char *, uid_t *uidp, char **grplist);

/* The lookup function for the first entry of this service.  */
extern int __nss_publickey_lookup (service_user **nip, const char *name, void **fctp);

static int
getnetid (const char *name, uid_t *uidp, char **grplist)
{
  static service_user *startp = NULL;
  static netid_function start_fct;
  service_user *nip;
  netid_function fct;
  enum nss_status status = NSS_STATUS_UNAVAIL;
  int no_more;

  if (startp == NULL)
    {
      no_more = __nss_publickey_lookup (&nip, "getnetid", (void **) &fct);
      if (no_more)
	startp = (service_user *) -1;
      else
	{
	  startp = nip;
	  start_fct = fct;
	}
    }
  else
    {
      fct = start_fct;
      no_more = (nip = startp) == (service_user *) -1;
    }

  while (! no_more)
    {
      status = (*fct) (name, uidp, grplist);

      no_more = __nss_next (&nip, "getnetid", (void **) &fct, status, 0);
    }

  return status == NSS_STATUS_SUCCESS;
}


/*
 * Convert network-name into unix credential
 */
netname2user(const char *netname,
	     uid_t *uidp,
	     gid_t *gidp,
	     int *gidlenp,
	     gid_t *gidlist)
{
	int needfree = 1, res;
	struct netid_userdata argp;
	int err;
	char *grplist;

	/*
	 * First we do some generic sanity checks on the name we were
	 * passed. This lets us assume they are correct in the backends.
	 *
	 * NOTE: this code only recognizes names of the form :
	 *		unix.UID@domainname
	 */
	if (netname == NULL) {
	  syslog(LOG_ERR, "zero netname");
	  return 0;
	}
	if (strncmp(netname, OPSYS, OPSYS_LEN) != 0) {
		trace1(TR_netname2user, 1);
		return (0);
	}
	if (! isdigit(netname[OPSYS_LEN+1])) { /* check for uid string */
		trace1(TR_netname2user, 1);
		return (0);
	}

	argp.uidp = uidp;
	argp.gidp = gidp;
	argp.gidlenp = gidlenp;
	argp.gidlist = gidlist;

	err = getnetid(netname, argp.uidp, &grplist);
	if (err == 0)
	  return 0;

	err = parse_gidlist(grplist, &argp);

	if (grplist)
	  free(grplist);

	return err == 0;
}

/*
 * Convert network-name to hostname (fully qualified)
 * NOTE: this code only recognizes names of the form :
 *		unix.HOST@domainname
 *
 * This is very simple.  Since the netname is of the form:
 *	unix.host@domainname
 * We just construct the hostname using information from the domainname.
 */
netname2host(const char *netname, char *hostname, int hostlen)
{
	char *p, *domainname;
	int len, dlen;

	trace1(TR_netname2host, 0);

	if (!netname) {
		syslog(LOG_ERR, "netname2host: null netname");
		goto bad_exit;
	}

	if (strncmp(netname, OPSYS, OPSYS_LEN) != 0)
		goto bad_netname;
	p = (char *) netname + OPSYS_LEN;	/* skip OPSYS part */
	if (*p != '.')
		goto bad_netname;
	++p;				/* skip '.' */

	domainname = strchr(p, '@');	/* get domain name */
	if (domainname == 0)
		goto bad_netname;

	len = domainname - p;		/* host sits between '.' and '@' */
	domainname++;			/* skip '@' sign */

	if (len <= 0)
		goto bad_netname;

	if (hostlen < len) {
		syslog(LOG_ERR,
			"netname2host: insufficient space for hostname");
		goto bad_exit;
	}

	if (isdigit(*p))		/* don't want uid here */
		goto bad_netname;

	if (*p == '\0')			/* check for null hostname */
		goto bad_netname;

	strncpy(hostname, p, len);

#ifdef NO_ANK_FIX	/* cannot believe .... These guys #$#&* */
	/* make into fully qualified hostname by concatenating domain part */
	dlen = strlen(domainname);
	if (hostlen < (len + dlen + 2)) {
		syslog(LOG_ERR,
			"netname2host: insufficient space for hostname");
		goto bad_exit;
	}

	hostname[len] = '.';
	strncpy(hostname+len+1, domainname, dlen);
	hostname[len+dlen+1] = '\0';
#else
	hostname[len] = '\0';
#endif

	trace1(TR_netname2host, 1);
	return (1);


bad_netname:
	syslog(LOG_ERR, "netname2host: invalid host netname %s", netname);

bad_exit:
	hostname[0] = '\0';
	trace1(TR_netname2host, 1);
	return (0);
}
