#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <tiuser.h>
#include <netdir.h>
#include <syslog.h>

static char thishostname[64];

struct nd_addrlist*
_netdir_getbyname(const struct netconfig* nconf,
		  const struct nd_hostserv* hserv)
{
	struct nd_addrlist *nd;
	struct netbuf *nb;
	int len;
	char * host;

	host = hserv->h_host;
	if (!thishostname[0])
		gethostname(thishostname, sizeof(thishostname));

	if (host == NULL ||
	    strcmp(host, HOST_SELF)==0 ||
	    strcmp(host, HOST_ANY)==0 ||
	    strcmp(host, HOST_BROADCAST)==0 ||
	    strcmp(host, HOST_SELF_CONNECT)==0)
		host = thishostname;
	else if (strcmp(host, thishostname)) {
		_nderror = ND_NOHOST;
		return NULL;
	}

	len = strlen(host) + 1 + strlen(hserv->h_serv);
	nd = malloc(sizeof(*nd));
	nd->n_cnt = 1;
	nb = nd->n_addrs = malloc(sizeof(*nb));
	nb->buf = malloc(len + 1);
	nb->maxlen = len + 1;
	nb->len = len;
	strcpy(nb->buf, host);
	strcat(nb->buf, ".");
	strcat(nb->buf, hserv->h_serv);
	return nd;
}

struct nd_hostservlist*
_netdir_getbyaddr(const struct netconfig* nconf,
		  const struct netbuf *addr)
{
	struct nd_hostservlist* nd;
	char *p;
	int slen, hlen;

	nd = malloc(sizeof(*nd));
	nd->h_cnt = 1;
	nd->h_hostservs = malloc(sizeof(*(nd->h_hostservs)));

	p = addr->buf+addr->len;
	while (p > addr->buf) {
		p--;
		if (*p == '.') {
			hlen = p - addr->buf;
			slen = addr->len - hlen - 1;
			goto ok;
		}
	}
	return NULL;

ok:
	nd->h_hostservs->h_host = malloc(hlen+1);
	nd->h_hostservs->h_host[hlen] = 0;
	memcpy(nd->h_hostservs->h_host, addr->buf, hlen);
	nd->h_hostservs->h_serv = malloc(slen+1);
	nd->h_hostservs->h_serv[slen] = 0;
	memcpy(nd->h_hostservs->h_serv, addr->buf+hlen+1, slen);
	return nd;
}

int
_netdir_options(const struct netconfig* nconf,
		int option, int fd, const char* par)
{
	struct nd_mergearg *ma = (struct nd_mergearg *)par;

	switch (option) {
	case ND_SET_BROADCAST:
        case ND_SET_RESERVEDPORT:
        case ND_CHECK_RESERVEDPORT:
		_nderror = ND_FAILCTRL;
		return -1;
	case ND_MERGEADDR:
		break;
	default:
		_nderror = ND_NOCTRL;
		return -1;
	}
	
	ma->m_uaddr = strdup(ma->s_uaddr);
	return 0;
}

char *
_taddr2uaddr(const struct netconfig* nconf,
		    const struct netbuf *taddr)
{
	char *p;
	p = malloc(taddr->len + 1);
	memcpy(p, taddr->buf, taddr->len);
	p[taddr->len] = 0;
	return p;
}

struct netbuf*
_uaddr2taddr(const struct netconfig* nconf,
		    const char *uaddr)
{
	struct netbuf *nb;

	nb = malloc(sizeof(struct netbuf));
	nb->maxlen = nb->len = strlen(uaddr);
	nb->buf = malloc(nb->maxlen);
	memcpy(nb->buf, uaddr, nb->len);
	return nb;
}
