#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>

#include <sys/poll.h>

/*
 *     	poll with checking for invalid descriptor.
 *	Called from poll, if select fails.
 */

static int poll_inval(struct pollfd *p, unsigned long nfds, int timeo)
{
	int i, k;
	int res;

	fd_set fds;
	fd_set fdr;
	fd_set fdw;

	struct timeval tv;

	int maxfd = -1;

	memset(&tv, 0, sizeof(tv));

	FD_ZERO(&fds);
	FD_ZERO(&fdr);
	FD_ZERO(&fdw);

	for (i=0; i<nfds; i++) {
		if (p[i].fd < 0)
			continue;

retry:
		FD_SET(p[i].fd, &fdr);
		res = select(p[i].fd+1, &fds, NULL, NULL, &tv);
		if (res < 0) {
			if (errno == EINTR)
				goto retry;
			p[i].revents = POLLNVAL;
			FD_CLR(p[i].fd, &fds);
			continue;
		}
		FD_CLR(p[i].fd, &fds);

		p[i].revents = 0;
		if (p[i].events & (POLLIN|POLLRDNORM|POLLPRI|POLLRDBAND)) {
			FD_SET(p[i].fd, &fdr);
			if (p[i].fd > maxfd)
				maxfd = p[i].fd;
		}
		if (p[i].events & (POLLOUT|POLLWRNORM|POLLWRBAND)) {
			FD_SET(p[i].fd, &fdw);
			if (p[i].fd > maxfd)
				maxfd = p[i].fd;
		}
	}

	if (timeo != INFTIM)
	{
		tv.tv_sec = timeo/1000;
		tv.tv_usec = (timeo%1000)*1000;
	}

	i = select(maxfd+1, &fdr, &fdw, NULL, timeo != INFTIM ? &tv : NULL);
	if (i <= 0)
		return i;

	k = 0;
	for (i=0; i<nfds; i++) {
		if (p[i].fd < 0)
			continue;
		if (p[i].revents & POLLNVAL) {
			k++;
			continue;
		}
		if (FD_ISSET(p[i].fd, &fdr))
			p[i].revents |= POLLIN|POLLRDNORM;
		if (FD_ISSET(p[i].fd, &fdw))
			p[i].revents |= POLLOUT|POLLWRNORM;
		if (p[i].revents)
			k++;
	}
	return k;
}


int poll(struct pollfd *p, unsigned long nfds, int timeo)
{
	int i, k;
	int maxfd;

	fd_set fdr;
	fd_set fdw;
	struct timeval tv;

	if (timeo != INFTIM)
	{
		tv.tv_sec = timeo/1000;
		tv.tv_usec = (timeo%1000)*1000;
	}

	maxfd=-1;
	FD_ZERO(&fdr);
	FD_ZERO(&fdw);

	for (i=0; i<nfds; i++) {
		if (p[i].fd < 0)
			continue;
		p[i].revents = 0;
		if (p[i].events & (POLLIN|POLLRDNORM|POLLPRI|POLLRDBAND)) {
			FD_SET(p[i].fd, &fdr);
			if (p[i].fd > maxfd)
				maxfd = p[i].fd;
		}
		if (p[i].events & (POLLOUT|POLLWRNORM|POLLWRBAND)) {
			FD_SET(p[i].fd, &fdw);
			if (p[i].fd > maxfd)
				maxfd = p[i].fd;
		}
	}
	i = select(maxfd+1, &fdr, &fdw, NULL, timeo != INFTIM ? &tv : NULL);
	if (i == 0)
		return 0;

	if (i < 0) {
		if (errno == EBADF || errno == EINVAL)
			return poll_inval(p, nfds, timeo);
		return i;
	}

	k = 0;
	for (i=0; i<nfds; i++) {
		if (p[i].fd < 0)
			continue;
		if (FD_ISSET(p[i].fd, &fdr))
			p[i].revents |= POLLIN|POLLRDNORM;
		if (FD_ISSET(p[i].fd, &fdw))
			p[i].revents |= POLLOUT|POLLWRNORM;
		if (p[i].revents)
			k++;
	}
	return k;
}
