#ifndef _TI_EXT_H_
#define _TI_EXT_H_

#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>

extern int module_init(void);
extern int __lt_register_dev(char *modname, char *name, int (*open)(int));
extern int __lt_register_module(char *name, int (*sync)(int));


struct lt_desc
{
	int	fd;
	int	tstate;
	int	tevent;
	unsigned flags;
	struct  ucred peercred;
	struct  t_info info;
	struct  t_bind bnd;
	struct  t_call peer;
	int	acc_q_len;
	struct  acc_elt *acc_q;
	struct  proto_ops *ops;
	char	*dev;
	void    *priv;
};

#define LT_RECV_CRED	1
#define LT_RECV_SPECDST	2

struct proto_ops
{
	int (*open)(int oflags);
	int (*sync)(int fd);
	int (*unsync)(struct lt_desc *lt);
	int (*bind)(struct lt_desc *lt, struct t_bind *bnd);
	int (*autobind)(struct lt_desc *lt, struct t_bind *bnd);
	struct acc_elt *(*accept)(struct lt_desc *lt, struct t_call*);
	int (*connect)(struct lt_desc *lt, struct t_call*, struct t_call*);
	int (*rcvconnect)(struct lt_desc *lt, struct t_call*);
	int (*recv)(struct lt_desc *lt, struct t_unitdata*, int *);
	int (*send)(struct lt_desc *lt, struct t_unitdata*, int);
	int (*shutdown)(struct lt_desc *lt);
	int (*optmgmt)(struct lt_desc *lt, struct t_optmgmt *opt,
		       struct t_optmgmt *ret);

	int (*opt_laddr)(struct lt_desc *lt, int on);
	int (*opt_rcvladdr)(struct lt_desc *lt, struct netbuf *opt, struct netbuf *laddr);
	int (*opt_sndladdr)(struct lt_desc *lt, struct netbuf *opt, struct netbuf *laddr);

	int (*opt_auth)(struct lt_desc *lt, int on);
	int (*opt_rcvauth)(struct lt_desc *lt, struct netbuf *opt, struct ucred *cred);
	int (*opt_sndauth)(struct lt_desc *lt, struct netbuf *opt, struct ucred *cred);
};

extern struct lt_desc *__lt_table[256];

extern __inline__ struct lt_desc * lt_lookup(int fd)
{
	struct lt_desc *lt;

	if (fd >= 0 && fd < 256 && (lt=__lt_table[fd]) != NULL)
		return lt;
	return NULL;
}

#define __lt_lookup(fd) \
({ struct lt_desc *__tmp_lt = lt_lookup((fd)); \
	   if (!__tmp_lt) { t_errno = TBADF; return -1; } \
	   __tmp_lt; \
})

void netbuf_copy(struct netbuf *dst, struct netbuf *src);

#if 0
#define DEBUGGING 1
#define TRACING 1
#endif

#ifdef DEBUGGING
#define DEBUG(x...) syslog(LOG_DEBUG, x)
#else
#define DEBUG(x...)
#endif
#ifdef TRACING
#define TRACE(x...) syslog(LOG_DEBUG, x)
#else
#define TRACE(x...)
#endif
#define NOTIMPLEMENTED(x...) syslog(LOG_DEBUG, "not implemented: " x)
#define ERROR(x...) syslog(LOG_ERR, x)

#endif
