#include <stdio.h>
#include <tiuser.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <string.h>

#include "ti_int.h"
#include "ti_sock.h"


static int
inet_sockaddr2netbuf(struct sockaddr_un *sun, int alen, struct netbuf *addr)
{
	int i;
	struct sockaddr_in *sin = (struct sockaddr_in*)sun;
	if (addr->maxlen < sizeof(*sin))
		return -1;
	addr->len = sizeof(*sin);
	/* Fix for possible kernel bug */
	memset(sin->sin_zero, 0, sizeof(sin->sin_zero));
	memcpy(addr->buf, sun, sizeof(*sin));
	return 0;
}

static int
inet_netbuf2sockaddr(struct sockaddr_un *sun, int *alenp, struct netbuf *addr)
{
	struct sockaddr_in *sin = (struct sockaddr_in*)sun;
	if (addr->len != sizeof(*sin))
		return -1;
	*alenp = sizeof(*sin);
	memcpy(sun, addr->buf, sizeof(*sin));
	return 0;
}

#if 0
static int
inet_cmsg2netbuf(struct lt_desc *lt, struct msghdr *msg, struct netbuf *opt)
{
	opt->len = 0;
	if (!msg->msg_controllen || (msg->msg_flags & MSG_CTRUNC))
		return 0;
	if (opt->maxlen < msg->msg_controllen)
		return -1;
	memcpy(opt->buf, msg->msg_control, msg->msgcontrollen);
	opt->len = msg->msg_controllen;
	return 0;
}

struct opthdr	/* Idiotic SunOS-5 hack; Sun guys were apparently insane
		 * when invented it.
		 */
{
	long	level;		/* protocol level affected */
	long	name;		/* option to modify */
	long	len;		/* length of option value */
	char	data[0];
};

#define	OPTALIGN(x) (((x)+sizeof(long)-1) & ~(sizeof(long)-1))

int inet_optmgmt(struct lt_desc *lt, struct t_optmgmt *req,
		 struct t_optmgmt *ret)
{
	struct opthdr *in, *out;
	int ilen = req->opt.len;
	int omaxlen = 0;
	int olen;

	if (ret) {
		out = (struct opthdr*)ret->opt.buf;
		omaxlen = ret->opt.maxlen;
		olen = 0;
	}

	if (req->flags != T_NEGOTIATE)
	{
		t_errno = TNOTSUPPORT;
		return -1;
	}

	for (in = (struct opthdr*)req->opt.buf;
	     ilen >= sizeof(struct opthdr);
	     ilen -= sizeof(*in) + OPTALIGN(in->len),
	     in = (struct opthdr*)(in->data + OPTALIGN(in->len))) {
		if (setsockopt(lt->fd, in->level, in->name, in->data, in->len) < 0) {
			t_errno = TSYSERR;
			return -1;
		}
		if (olen + sizeof(*out) + in->len < omaxlen) {
			int retlen = in->len;

			out->level = in->level;
			out->name = in->name;
			if (getsockopt(lt->fd, out->level, out->name, out->data,
				   &retlen) < 0)
				continue;
			out->len = retlen;
			olen += sizeof(*out) + OPTALIGN(retlen);
			out = (struct opthdr*)(out->data+OPTALIGN(retlen));
		}
	}
	if (ret)
		ret->opt.len = olen;
	
	return 0;
}

#else

int inet_optmgmt(struct lt_desc *lt, struct t_optmgmt *req,
		 struct t_optmgmt *ret)
{
	struct msghdr msg;
	struct cmsghdr *cmsg, *out;
	int omaxlen = 0;
	int olen;

	msg.msg_control = req->opt.buf;
	msg.msg_controllen = req->opt.len;

	if (ret) {
		out = (struct cmsghdr*)ret->opt.buf;
		omaxlen = ret->opt.maxlen;
		olen = 0;
	}

	if (req->flags != T_NEGOTIATE)
	{
		t_errno = TNOTSUPPORT;
		return -1;
	}

	for (cmsg=CMSG_FIRSTHDR(&msg); cmsg;
	     cmsg=CMSG_NXTHDR(&msg, cmsg)) {
		int len = cmsg->cmsg_len - sizeof(struct cmsghdr);

		if (setsockopt(lt->fd, cmsg->cmsg_level, cmsg->cmsg_type,
			       CMSG_DATA(cmsg), len) < 0) {
			t_errno = TSYSERR;
			return -1;
		}
		if (olen + cmsg->cmsg_len < omaxlen) {
			*out = *cmsg;
			if (getsockopt(lt->fd, out->cmsg_level,
				       out->cmsg_type, CMSG_DATA(out),
				       &len) < 0)
				continue;
			out->cmsg_len = len + sizeof(*out);
			olen += CMSG_ALIGN(out->cmsg_len);
			out = (struct cmsghdr*)((char*)out+CMSG_ALIGN(out->cmsg_len));
		}
	}
	if (ret)
		ret->opt.len = olen;
	
	return 0;
}

int inet_laddr(struct lt_desc *lt, int on)
{
	struct sockaddr_in *sin = (struct sockaddr_in*)lt->bnd.addr.buf;
	if (sin->sin_family == AF_INET && sin->sin_addr.s_addr)
		return -1;
	if (setsockopt(lt->fd, SOL_IP, IP_PKTINFO, &on, sizeof(on)) < 0) {
		t_errno = TBADOPT;
		return -1;
	}

	return 0;
}

int inet_rcvladdr(struct lt_desc *lt, struct netbuf *opt, struct netbuf *laddr)
{
	struct msghdr msg;
	struct cmsghdr *cmsg;

	msg.msg_control = opt->buf;
	msg.msg_controllen = opt->len;

	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg;
	     cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level != SOL_IP)
			continue;
		if (cmsg->cmsg_type == IP_RECVOPTS) {
			syslog(LOG_ERR, "message with IP options!!!\n");
			return -1;
		}
		if (cmsg->cmsg_type == IP_PKTINFO &&
		    cmsg->cmsg_len >= sizeof(struct cmsghdr) + sizeof(struct in_pktinfo)) {
			((struct sockaddr_in*)laddr->buf)->sin_addr
				= ((struct in_pktinfo*)CMSG_DATA(cmsg))->ipi_spec_dst;
			return 0;
		}
	}
	return -1;
}

int inet_sndladdr(struct lt_desc *lt, struct netbuf *opt, struct netbuf *laddr)
{
	int optptr = CMSG_ALIGN(opt->len);
	struct cmsghdr *cmsg=(struct cmsghdr*)(opt->buf+optptr);
	struct sockaddr_in *sin = (struct sockaddr_in*)laddr->buf;

	if (sin->sin_addr.s_addr &&
	    opt->maxlen >=  optptr + sizeof(struct cmsghdr)+sizeof(struct in_pktinfo)) {
		cmsg->cmsg_len = sizeof(struct cmsghdr)+sizeof(struct in_pktinfo);
		opt->len += cmsg->cmsg_len;
		cmsg->cmsg_level = SOL_IP;
		cmsg->cmsg_type = IP_PKTINFO;
		memset(CMSG_DATA(cmsg), 0, sizeof(struct in_pktinfo));
		((struct in_pktinfo*)CMSG_DATA(cmsg))->ipi_spec_dst =
		((struct in_pktinfo*)CMSG_DATA(cmsg))->ipi_addr =
			sin->sin_addr;
		return 0;
	}
	return -1;
}

#endif

int inet_autobind(struct lt_desc *lt, struct t_bind *bnd)
{
	struct sockaddr_in sin;
	struct sockaddr_un sun;
	int	addr_len = sizeof(sin);
	struct sock_ops *sops = (struct sock_ops *)lt->priv;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
 
	if (bind(lt->fd, (struct sockaddr*)&sin, sizeof(sin)) < 0)
	{
		t_errno = TSYSERR;
		return -1;
	}
	if (getsockname(lt->fd, (struct sockaddr*)&sun, &addr_len) < 0)
	{
		t_errno = TSYSERR;
		return -1;
	}
	if (sops->sockaddr2netbuf(&sun, addr_len, &lt->bnd.addr) < 0)
	{
		t_errno = TSYSERR;
		return -1;
	}
#ifdef DEBUGGING
	DEBUG("inet autobind %d", lt->fd);
	sops->printaddr(&lt->bnd.addr);
#endif	
	if (bnd && bnd->qlen) {
		listen(lt->fd, bnd->qlen);
		lt->bnd.qlen = bnd->qlen;
	} else
		lt->bnd.qlen = 0;
	return 0;
}

static int inet_printaddr(struct netbuf *addr)
{
	struct sockaddr_in *sin = (struct sockaddr_in*)addr->buf;
	if (addr->len != sizeof(*sin) ||
	    sin->sin_family != AF_INET) {
		syslog(LOG_DEBUG, "bad inet address\n");
	}
	syslog(LOG_DEBUG, "%08x:%04x\n", ntohl(sin->sin_addr.s_addr),
	       ntohs(sin->sin_port));
	return 0;
}

struct sock_ops tcp_sops = {
	AF_INET,
	SOCK_STREAM,
	IPPROTO_TCP,
	inet_sockaddr2netbuf,
	inet_netbuf2sockaddr,
	NULL,
	inet_printaddr
};

struct sock_ops udp_sops = {
	AF_INET,
	SOCK_DGRAM,
	IPPROTO_UDP,
	inet_sockaddr2netbuf,
	inet_netbuf2sockaddr,
	NULL,
	inet_printaddr
};
