#include <stdio.h>
#include <malloc.h>
#include <tiuser.h>

#include "ti_ext.h"

/*
 *	TLI memory "manager" 8)
 */

#define MAX_SALLOC 5

#define SALLOC(x) \
({ struct x * __tmp = malloc(sizeof(struct x)); \
   if (__tmp == NULL) goto outofmem; \
   stack[sp++]=(void*)__tmp; \
   memset(__tmp, 0, sizeof(struct x)); \
   __tmp; \
})

#define NBALLOC(nb, size) \
({ int __sz = (size); \
   char * __tmp; \
   if (__sz < 0 && fields != T_ALL) goto outofmem; \
   if (__sz > 0) { \
       __tmp = malloc((size)); \
       if (__tmp == NULL) goto outofmem; \
       stack[sp++]=(void*)__tmp; \
       (nb).buf = __tmp; \
       (nb).maxlen = (size); \
   } \
})

#define SCLEAR() ({ while (sp > 0) free(stack[--sp]); })

char *t_alloc(int fd, int stype, int fields)
{
	int i;
	void *t;
	struct lt_desc *lt = lt_lookup(fd);
	void * stack[MAX_SALLOC];
	int    sp = 0;

	if (stype != T_INFO && !lt) {
		t_errno = TBADF;
		return NULL;
	}

	switch (stype) {
	case T_BIND:
	{
		struct t_bind *tt = SALLOC(t_bind);
		if (fields & T_ADDR)
			NBALLOC(tt->addr, lt->info.addr);
		return (char*)tt;
	}
	case T_CALL:
	{
		struct t_call *tt = SALLOC(t_call);
		if (fields & T_ADDR)
			NBALLOC(tt->addr, lt->info.addr);
		if (fields & T_UDATA)
			NBALLOC(tt->udata, lt->info.connect);
		if (fields & T_OPT)
			NBALLOC(tt->opt, lt->info.options);
		return (char*)tt;
	}
	case T_UNITDATA:
	{
		struct t_unitdata *tt = SALLOC(t_unitdata);
		if (fields & T_ADDR)
			NBALLOC(tt->addr, lt->info.addr);
		if (fields & T_UDATA)
			NBALLOC(tt->udata, lt->info.tsdu);
		if (fields & T_OPT)
			NBALLOC(tt->opt, lt->info.options);
		return (char*)tt;
	}
	case T_OPTMGMT:
	{
		struct t_optmgmt *tt = SALLOC(t_optmgmt);
		if (fields & T_OPT)
			NBALLOC(tt->opt, lt->info.options);
		return (char*)tt;
	}
	case T_DIS:
	{
		struct t_discon *tt = SALLOC(t_discon);
		if (fields & T_UDATA)
			NBALLOC(tt->udata, lt->info.discon);
		return (char*)tt;
	}
	case T_UDERROR:
	{
		struct t_uderr *tt = SALLOC(t_uderr);
		if (fields & T_ADDR)
			NBALLOC(tt->addr, lt->info.addr);
		if (fields & T_OPT)
			NBALLOC(tt->opt, lt->info.options);
		return (char*)tt;
	}
	case T_INFO:
	{
		struct t_info *tt = SALLOC(t_info);
		return (char*)tt;
	}
	default:
		t_errno = TNOTSUPPORT /* TNOSTRUCTYPE*/;
		return NULL;
	}

outofmem:
	SCLEAR();
	t_errno = TSYSERR;
	return NULL;
}


int t_free(char *p, int stype)
{
	void * ptr = p;

	if (!ptr)
		return 0;

	switch(stype) {
	case T_BIND:
	{
		struct t_bind *tt = ptr;
		if (tt->addr.buf)
			free(tt->addr.buf);
		break;
        }
	case T_CALL:
	{
		struct t_call *tt = ptr;
		if (tt->addr.buf)
			free(tt->addr.buf);
		if (tt->opt.buf)
			free(tt->opt.buf);
		if (tt->udata.buf)
			free(tt->udata.buf);
		break;
	}
	case T_UNITDATA:
	{
		struct t_unitdata *tt = ptr;
		if (tt->addr.buf)
			free(tt->addr.buf);
		if (tt->opt.buf)
			free(tt->opt.buf);
		if (tt->udata.buf)
			free(tt->udata.buf);
		break;
	}
	case T_OPTMGMT:
	{
		struct t_optmgmt *tt = ptr;
		if (tt->opt.buf)
			free(tt->opt.buf);
		break;
	}
	case T_DIS:
	{
		struct t_discon *tt = ptr;
		if (tt->udata.buf)
			free(tt->udata.buf);
		break;
	}
	case T_UDERROR:
	{
		struct t_uderr *tt = ptr;
		if (tt->addr.buf)
			free(tt->addr.buf);
		if (tt->opt.buf)
			free(tt->opt.buf);
		break;
	}
	case T_INFO:
		break;
	default:
		t_errno = TNOTSUPPORT /* TNOSTRUCTYPE */;
		return -1;
	}
	free(ptr);
	return 0;
}

