#include <stdio.h>
#include <tiuser.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <string.h>

#include <sys/ticots.h>

#include "ti_int.h"
#include "ti_sock.h"


static int
unix_sockaddr2netbuf(struct sockaddr_un *sun, int alen, struct netbuf *addr)
{
	addr->len = 0;
	if (alen < sizeof(short))
		return -1;
	if (alen == sizeof(short))
	{
		addr->len = 0;
		return 0;
	}
	if (alen > sizeof(short)+1 && sun->sun_path[0]==0)
	{
		if (alen-sizeof(short)-1 > addr->maxlen)
			return -1;
		addr->len = alen-sizeof(short)-1;
		memcpy(addr->buf, sun->sun_path+1, addr->len);
		return 0;
	}
	if (sun->sun_path[0] != '/')
		return -1;
	alen = strlen(sun->sun_path)+1;
	if (addr->maxlen < alen || sun)
		return -1;
	addr->len = alen;
	memcpy(addr->buf, sun->sun_path, alen);
	return 0;
}

static int
unix_netbuf2sockaddr(struct sockaddr_un *sun, int *alenp, struct netbuf *addr)
{
	if (addr->len < 1 || !addr->buf)
		return -1;
	if (addr->buf[0] == '/')
	{
		if (addr->len > sizeof(*sun) - sizeof(short))
			return -1;
		*alenp = addr->len+sizeof(short);
		sun->sun_family = AF_UNIX;
		memcpy(sun->sun_path, addr->buf, addr->len);
		return 0;
	}
	if (addr->len > sizeof(*sun) - sizeof(short) - 1)
		return -1;
	*alenp = addr->len+sizeof(short)+1;
	sun->sun_family = AF_UNIX;
	sun->sun_path[0] = 0;
	memcpy(sun->sun_path+1, addr->buf, addr->len);
	return 0;
}

#if 0
/*
 *	Hacky SunOS-5.3 loopback transport options.
 *	I'd like to look at this idiot, who invented them.
 */

struct ti_opt_hdr
{
	long	hdr_thisopt_off;	/* offset to current opt */
	long	hdr_nexthdr_off;	/* offset to next hdr */
};

struct ti_opt_body
{
	long    type;
	union {
		long	flag;
		uid_t	uid;
		gid_t	gid;
	} val;
};

int unix_makeoptions(struct lt_desc *lt, struct netbuf *opt)
{
	struct ti_opt_hdr *p;
	struct ti_opt_body *o;

	if (!(lt->flags & LT_RECV_UID) || !lt->peercred.pid)
		return 0;

	if (opt->maxlen < opt->len + sizeof(*p) + sizeof(*o))
		return 0;

	p = (struct ti_opt_hdr*)opt->buf;
	p->hdr_thisopt_off = (void*)&p[1] - (void*)p;
	p->hdr_nexthdr_off = 0;
	o = (struct ti_opt_body*)&p[1];
	o->type = TCO_OPT_UID;
	o->val.uid = lt->peercred.uid;
	opt->len += sizeof(*p) + sizeof(*o);
	return 0;
}

int unix_optmgmt(struct lt_desc *lt, struct t_optmgmt *req, struct t_optmgmt *ret)
{
	int next_hdr, optbody;
	struct ti_opt_hdr	*p;
	struct ti_opt_body 	*o;

	if (req->flags != T_NEGOTIATE || (ret && req != ret)) {
		t_errno = TNOTSUPPORT;
		return -1;
	}

	if (req->opt.len <= 0)
		return 0;

	next_hdr = 0;
	do {
		p = (struct ti_opt_hdr *)(req->opt.buf+next_hdr);
		o = (struct ti_opt_body *)(req->opt.buf+p->hdr_thisopt_off);
		next_hdr = p->hdr_nexthdr_off;
		if (o->type == TCO_OPT_NOOP)
			continue;
		switch (o->type) {
		case TCO_OPT_SETID:
			if (o->val.flag&~(TCO_IDFLG_UID|TCO_IDFLG_GID)) {
				t_errno = TBADOPT;
				return -1;
			}
			if (o->val.flag&TCO_IDFLG_UID)
				lt->flags |= LT_RECV_UID;
			if (o->val.flag&TCO_IDFLG_GID)
				lt->flags |= LT_RECV_GID;
			if (lt->flags & (LT_RECV_UID|LT_RECV_GID)) {
				int on = 1;
				if (setsockopt(lt->fd, SOL_SOCKET, SO_PASSCRED,
					       &on, sizeof(on)) < 0) {
					t_errno = TSYSERR;
					return -1;
				}
			}
			break;
		case TCO_OPT_GETID:
		case TCO_OPT_UID:
		case TCO_OPT_GID:
		case TCO_OPT_RUID:
		case TCO_OPT_RGID:
		default:
			t_errno = TBADOPT;
			return -1;
		}
		if (next_hdr >= req->opt.len) {
			t_errno = TBADOPT;
			return -1;
		}
	} while (next_hdr != TCO_OPT_NOHDR);
	return 0;
}

static int
unix_cmsg2netbuf(struct lt_desc *lt, struct msghdr *msg, struct netbuf *opt)
{
	struct cmsghdr *cmsg;

	opt->len = 0;
	if (!msg->msg_controllen || (msg->msg_flags & MSG_CTRUNC))
		return 0;

	for (cmsg=CMSG_FIRSTHDR(msg); cmsg; cmsg=CMSG_NXTHDR(msg, cmsg)) {
		if (cmsg->cmsg_level == SOL_SOCKET) {
			if (cmsg->cmsg_type == SCM_CREDENTIALS) {
				struct ucred *cred = (struct ucred*)cmsg->cmsg_data;
				struct ti_opt_hdr *p;
				struct ti_opt_body *o;
				p = (struct ti_opt_hdr*)opt->buf;
				if (lt->flags & LT_RECV_UID) {
					p->hdr_thisopt_off = (void*)&p[1] - (void*)p;
					p->hdr_nexthdr_off = 0;
					o = (struct ti_opt_body*)&p[1];
					o->type = TCO_OPT_UID;
					o->val.uid = cred->uid;
					opt->len += sizeof(*p) + sizeof(*o);
					return 0;
				}
			}
		}
	}
	return 0;
}

#else

int unix_makeoptions(struct lt_desc *lt, struct netbuf *opt)
{
	struct cmsghdr *cmsg = (struct cmsghdr*)(opt->buf+opt->len);
	struct ucred *cred;

	if (!(lt->flags & LT_RECV_CRED) || !lt->peercred.pid)
		return 0;

	if (opt->maxlen < opt->len + sizeof(*cmsg) + sizeof(struct ucred))
		return 0;

	memcpy(CMSG_DATA(cmsg), &lt->peercred, sizeof(lt->peercred));
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_CREDENTIALS;
	cmsg->cmsg_len = sizeof(*cmsg) + sizeof(lt->peercred);
	opt->len += CMSG_ALIGN(cmsg->cmsg_len);
	return 0;
}

int unix_optmgmt(struct lt_desc *lt, struct t_optmgmt *req,
		 struct t_optmgmt *ret)
{
	int on = 1;
	struct msghdr msg;
	struct cmsghdr *cmsg;

	if (req->flags != T_NEGOTIATE || (ret && req != ret)) {
		t_errno = TNOTSUPPORT;
		return -1;
	}

	if (req->opt.len <= 0)
		return 0;

	msg.msg_control = req->opt.buf;
	msg.msg_controllen = req->opt.len;

	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg;
	     cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level != SOL_SOCKET)
			continue;
		if (cmsg->cmsg_type != SO_PASSCRED)
			continue;
		lt->flags |= LT_RECV_CRED;
		if (setsockopt(lt->fd, SOL_SOCKET, SO_PASSCRED,
			       &on, sizeof(on)) < 0) {
			t_errno = TSYSERR;
			return -1;
		}
	}
	return 0;
}

#endif

int unix_autobind(struct lt_desc *lt, struct t_bind *bnd)
{
	struct sockaddr_un sun;
	int addr_len = sizeof(sun);
	int on = 1;
	struct sock_ops *sops = (struct sock_ops *)lt->priv;

	setsockopt(lt->fd, SOL_SOCKET, SO_PASSCRED, &on, sizeof(on));

	sun.sun_family = AF_UNIX;
	sun.sun_path[0] = 0;
 
	if (bind(lt->fd, (struct sockaddr*)&sun, sizeof(short)) < 0)
	{
		if (sops->type == SOCK_DGRAM) {
			ERROR("autobind on fd %d : %m", lt->fd);
		        t_errno = TSYSERR;
			return -1;
		}
	}
	if (getsockname(lt->fd, (struct sockaddr*)&sun, &addr_len) < 0)
	{
		ERROR("getsockname on fd %d : %m", lt->fd);
		t_errno = TSYSERR;
		return -1;
	}
	if (sops->sockaddr2netbuf(&sun, addr_len, &lt->bnd.addr) < 0)
	{
		ERROR("sockaddr2netbuf failure");
		t_errno = TSYSERR;
		return -1;
	}
#ifdef DEBUGGING
	DEBUG("unix autobind %d", lt->fd);
	sops->printaddr(&lt->bnd.addr);
#endif	
	if (bnd && bnd->qlen) {
		listen(lt->fd, bnd->qlen);
		lt->bnd.qlen = bnd->qlen;
	} else
		lt->bnd.qlen = 0;
	return 0;
}

static int unix_printaddr(struct netbuf *addr)
{
	if (addr->len == 0)
		syslog(LOG_DEBUG, "not bound");
	else
		syslog(LOG_DEBUG, "%*.*s", addr->len,
		       addr->len, addr->buf);
	return 0;
}


struct sock_ops unix_stream_sops = {
	AF_UNIX,
	SOCK_STREAM,
	0,
	unix_sockaddr2netbuf,
	unix_netbuf2sockaddr,
	unix_makeoptions,
	unix_printaddr
};

struct sock_ops unix_dgram_sops = {
	AF_UNIX,
	SOCK_DGRAM,
	0,
	unix_sockaddr2netbuf,
	unix_netbuf2sockaddr,
	NULL,
	unix_printaddr
};
