/*	Copyright (c) 1984, 1986, 1987, 1988, 1989 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"@(#)rpc_subr.c	1.13	94/04/25 SMI"	/* SVr4.0 1.1	*/
/*
 * Miscellaneous support routines for kernel implementation of RPC.
 *
 */

#include	<sys/param.h>
#include	<sys/t_lock.h>
#include	<rpc/types.h>
#include	<sys/user.h>
#include	<sys/vnode.h>
#include	<sys/stream.h>
#include	<sys/stropts.h>
#include	<sys/strsubr.h>
#include	<sys/socket.h>
#include	<sys/tihdr.h>
#include	<sys/timod.h>
#include	<sys/tiuser.h>
#include	<sys/systm.h>
#include	<rpc/auth.h>
#include	<rpc/clnt.h>
#include	<sys/cmn_err.h>

#ifdef _KERNEL
/*
 * Kernel level debugging aid. The global variable "rpclog" is a bit
 * mask which allows various types of debugging messages to be printed
 * out.
 *
 *	rpclog & 1 	will cause actual failures to be printed.
 *	rpclog & 2	will cause informational messages to be
 *			printed on the client side of rpc.
 *	rpclog & 4	will cause informational messages to be
 *			printed on the server side of rpc.
 */

int rpclog = 0;

int
rpc_log(u_long level, register char *str, register int a1)
{

	if (level & rpclog)
		printf(str, a1);
	return (0);
}

/*
 *  Revoke the authentication key in the given AUTH handle by setting
 *  it to NULL.  If newkey is true, then generate a new key instead of
 *  nulling out the old one.  This is necessary for AUTH_DES because
 *  the new key will be used next time the user does a keylogin.  If
 *  the zero'd key is used as actual key, then it cannot be revoked
 *  again!  This is not a problem for AUTH_KERB because the key is
 *  embedded in the ticket, which is cached.
 */

void
rpc_revoke_key(register AUTH *auth, int newkey)
{

	if (auth) {
		if (newkey) {
			if (key_gendes(&auth->ah_key) != RPC_SUCCESS) {
				/* failed to get new key, munge the old one */
				auth->ah_key.key.high ^= auth->ah_key.key.low;
				auth->ah_key.key.low  += auth->ah_key.key.high;
			}
		} else {
			/* null out old key */
			auth->ah_key.key.high = 0;
			auth->ah_key.key.low  = 0;
		}
	}
}

void
rpc_poptimod(vnode_t *vp)
{
	int error, isfound, ret;

	error = strioctl(vp, I_FIND, (int) "timod", 0, K_TO_K, CRED(),
			&isfound);
	if (error) {
		RPCLOG(1, "rpc_poptimod: I_FIND strioctl error %d\n", error);
		return;
	}
	if (isfound) {
		error = strioctl(vp, I_POP, (int) "timod", 0, K_TO_K, CRED(),
				&ret);
		if (error) {
			RPCLOG(1, "rpc_poptimod: I_POP strioctl error %d\n",
				error);
			return;
		}
	}
}

void
rpc_adjustque(vnode_t *vp, ulong hiwat, ulong lowat)
{
	struct T_optmgmt_req *req;
	struct opthdr *opt;
	int bufsize;
	struct strioctl strioc;
	int error;
	int retval;
	int isfound;

	bufsize = sizeof (*req) + sizeof (*opt) + sizeof (ulong);
	req = (struct T_optmgmt_req *)kmem_alloc(bufsize, KM_SLEEP);

	/*
	 * see if SOCKMOD is already pushed
	 */
	error = strioctl(vp, I_FIND, (int) "sockmod", 0, K_TO_K, CRED(),
			&retval);
	if (error) {
		kmem_free((caddr_t)req, bufsize);
		return;
	}

	if (retval == 0) {
tryagain:
		error = strioctl(vp, I_PUSH, (int) "sockmod", 0, K_TO_K, CRED(),
								&retval);
		if (error) {
			switch (error) {
			case ENOSPC:
			case EAGAIN:
			case ENOSR:
				/*
				 * This probably means the master file
				 * should be tuned.
				 */
				cmn_err(CE_WARN,
			"rpc_adjustque: I_PUSH of sockmod failed, error %d\n",
					error);
				(void) delay(HZ);
				error = 0;
				goto tryagain;

			default:
				kmem_free((caddr_t)req, bufsize);
				cmn_err(CE_WARN,
			"rpc_adjustque: I_PUSH of sockmod failed, error %d\n",
					error);
				return;
			}
		}
	}

	req->PRIM_type = T_OPTMGMT_REQ;
	req->OPT_length = sizeof (*opt) + sizeof (int);
	req->OPT_offset = sizeof (*req);
	req->MGMT_flags = T_NEGOTIATE;

	opt = (struct opthdr *)((char *)req + sizeof (*req));
	opt->level = SOL_SOCKET;
	opt->name = SO_RCVBUF;
	opt->len = sizeof (ulong);

	*((ulong *)((char *)opt + sizeof (*opt))) = hiwat;

	strioc.ic_cmd = TI_OPTMGMT;
	strioc.ic_timout = -1;
	strioc.ic_len = bufsize;
	strioc.ic_dp = (char *)req;

	error = strioctl(vp, I_STR, (int) &strioc, 0, K_TO_K, CRED(), &retval);
	if (error || retval) {
		cmn_err(CE_WARN, "rpc_adjustque: error %d setting hiwat\n",
			error);
		kmem_free((caddr_t)req, bufsize);
		return;
	}

	req->PRIM_type = T_OPTMGMT_REQ;
	req->OPT_length = sizeof (*opt) + sizeof (int);
	req->OPT_offset = sizeof (*req);
	req->MGMT_flags = T_NEGOTIATE;

	opt = (struct opthdr *)((char *)req + sizeof (*req));
	opt->level = SOL_SOCKET;
	opt->name = SO_RCVLOWAT;
	opt->len = sizeof (ulong);

	*((ulong *)((char *)opt + sizeof (*opt))) = lowat;

	strioc.ic_cmd = TI_OPTMGMT;
	strioc.ic_timout = -1;
	strioc.ic_len = bufsize;
	strioc.ic_dp = (char *)req;

	error = strioctl(vp, I_STR, (int) &strioc, 0, K_TO_K, CRED(), &retval);
	if (error || retval) {
		cmn_err(CE_WARN, "rpc_adjustque: error %d setting lowat\n",
			error);
		kmem_free((caddr_t)req, bufsize);
		return;
	}

	kmem_free((caddr_t)req, bufsize);

	error = strioctl(vp, I_FIND, (int) "sockmod", 0, K_TO_K, CRED(),
			&isfound);
	if (error) {
		cmn_err(CE_WARN, "rpc_adjustque: I_FIND strioctl error %d\n",
			error);
		return;
	}
	if (isfound) {
		error = strioctl(vp, I_POP, (int) "sockmod", 0, K_TO_K, CRED(),
				&retval);
		if (error)
			cmn_err(CE_WARN,
				"rpc_adjustque: I_POP strioctl error %d\n",
				error);
	}
}
#endif /* _KERNEL */
