/*	Copyright (c) 1984, 1986, 1987, 1988, 1989 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#pragma ident	"@(#)rpcmod.c	1.12	93/07/11 SMI"

/*
 * Kernel RPC filtering module
 */

#include "sys/param.h"
#include "sys/types.h"
#include "sys/stream.h"
#include "sys/stropts.h"
#include "sys/tihdr.h"
#include "sys/timod.h"
#include "sys/tiuser.h"
#include "sys/debug.h"
#include "sys/signal.h"
#include "sys/psw.h"
#include "sys/pcb.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/cred.h"
#include "sys/inline.h"
#include "sys/cmn_err.h"
#include "sys/kmem.h"
#include "sys/file.h"
#include "sys/sysmacros.h"
#include "sys/systm.h"
#include "sys/t_lock.h"
#include "sys/ddi.h"
#include "rpc/rpc.h"

/*
 * This is the loadable module wrapper.
 */
#include <sys/conf.h>
#include <sys/modctl.h>

extern struct streamtab rpcinfo;

static struct fmodsw fsw = {
	"rpcmod",
	&rpcinfo,
	D_NEW|D_MP,
};

/*
 * Module linkage information for the kernel.
 */
extern struct mod_ops mod_strmodops;

static struct modlstrmod modlstrmod = {
	&mod_strmodops, "rpc interface str mod", &fsw
};

static struct modlinkage modlinkage = {
	MODREV_1, (void *)&modlstrmod, NULL
};

_init()
{
	int			error;

	if (error = mod_install(&modlinkage))
		return (error);

	return (0);
}

/*
 * The unload entry point fails, because we advertise entry points into
 * rpcmod from the rest of kRPC: rpcmod_send() and rpcmod_release().
 */
_fini()
{
	return (EBUSY);
}

_info(modinfop)
	struct modinfo *modinfop;
{
	return (mod_info(&modlinkage, modinfop));
}


extern nulldev();

#define	RPCMOD_ID	2049

int rpcmodopen(), rpcmodclose();
int rpcmodrput(), rpcmodwput();
int rpcmodrsrv(), rpcmodwsrv();

static struct module_info rpcmod_info =
	{RPCMOD_ID, "rpcmod", 0, INFPSZ, 256*1024, 1024};
/*
 * Read side has no service procedure.
 */
static struct qinit rpcmodrinit =
	{ rpcmodrput, NULL, rpcmodopen, rpcmodclose, nulldev,
	    &rpcmod_info, NULL};
/*
 * putnext is a subroutine in SVR4.
 */
extern int putnext(queue_t *, mblk_t *);

/*
 * The write put procedure is simply putnext to conserve stack space.
 * The write service procedure is not used to queue data, but instead to
 * synchronize with flow control.
 */
static struct qinit rpcmodwinit =
	{ putnext, rpcmodwsrv, rpcmodopen, rpcmodclose, nulldev,
	    &rpcmod_info, NULL};
struct streamtab rpcinfo = { &rpcmodrinit, &rpcmodwinit, NULL, NULL };

extern void add_rpcm(struct rpcm *);
extern void del_rpcm(struct rpcm *);
extern void svc_queuereq(queue_t *, mblk_t *);

extern int (*rpc_send)(queue_t *, mblk_t *);
static int rpcmod_send(queue_t *, mblk_t *);

extern void (*rpc_rele)(queue_t *, mblk_t *);
static void rpcmod_release(queue_t *, mblk_t *);

/*
 * rpcmodopen -	open routine gets called when the module gets pushed
 *		onto the stream.
 */
/*ARGSUSED*/
rpcmodopen(q, devp, flag, sflag, crp)
	register queue_t *q;
	dev_t *devp;
	int flag;
	int sflag;
	cred_t *crp;
{
	struct rpcm *rmp;
	char buf[20];

	TRACE_0(TR_FAC_KRPC, TR_RPCMODOPEN_START,
		"rpcmodopen_start:");

	/*
	 * Initialize entry points to release a rpcmod slot (and an input
	 * message if supplied) and to send an output message to the module
	 * below rpcmod.
	 */
	if (rpc_send == NULL) {
		rpc_rele = rpcmod_release;
		rpc_send = rpcmod_send;
	}

	ASSERT(q != NULL);

	/*
	 * Check for re-opens.
	 */
	if (q->q_ptr) {
		TRACE_1(TR_FAC_KRPC, TR_RPCMODOPEN_END,
			"rpcmodopen_end:(%s)", "q->qptr");
		return (0);
	}

	/*
	 * The write put procedure is just putnext(), and thus does
	 * not call canput() to check for downstream flow control.
	 * We could implement a rpcmod_wput to do this, but if we
	 * did, we would have to queue data on canput() failures, making
	 * rpcmod_wsrv() more complex (slower). If we allow any user
	 * to push rpcmod, it would be a trivial thing for the user to
	 * run the system out of memory by abusing the module.
	 *
	 * Only root can use this module, and it is assumed that root
	 * will use this module properly, and NOT send bulk data from
	 * downstream.
	 */
	if (suser(crp) == 0)
		return (EPERM);

	/*
	 * Allocate slot data structure.
	 */
	rmp = kmem_zalloc(sizeof (*rmp), KM_SLEEP);

	(void) sprintf(buf, "rpcm slot %lx", rmp);
	mutex_init(&rmp->rm_lock, buf, MUTEX_DEFAULT, NULL);
	cv_init(&rmp->rm_fwait, "rpcmod slot flow wait", CV_DEFAULT, NULL);
	cv_init(&rmp->rm_cwait, "rpcmod slot close wait", CV_DEFAULT, NULL);

	(void) sprintf(buf, "rpcm que %lx", rmp);
	mutex_init(&rmp->rm_qlock, buf, MUTEX_DEFAULT, NULL);
	cv_init(&rmp->rm_qcv, "rpcmod queue wait", CV_DEFAULT, NULL);

	q->q_ptr = (void *)rmp;
	WR(q)->q_ptr = (void *)rmp;

	/*
	 * Add private data structures to list for svc_run,
	 */
	add_rpcm(rmp);

	/*
	 * Normally, a per slot mutex need not be entered in the open()
	 * entry point. However, we need to de-reference q_next, and
	 * q_next is not available until qprocson() is invoked. Invoking
	 * qprocson() puts the queue onto the stream, which can let our
	 * put procedures run, which without rm_stream set, would be a
	 * bad thing. Thus we acquire a lock to synchronize with the
	 * put procedures.
	 */
	mutex_enter(&rmp->rm_lock);
	qprocson(q);

	/*
	 * Record the stream head pointer. Really, rpcmod does not know
	 * or care that the stream head's q_ptr field points to an
	 * stdata_t structure, we just need the pointer to be used as a
	 * key.
	 */
	rmp->rm_stream = q->q_next->q_ptr;
	rmp->rm_wq = WR(q);
	mutex_exit(&rmp->rm_lock);

	TRACE_1(TR_FAC_KRPC, TR_RPCMODOPEN_END,
		"rpcmodopen_end:(%s)", "end");
	return (0);
}

/*
 * rpcmodclose - This routine gets called when the module gets popped
 * off of the stream.
 */
/*ARGSUSED*/
rpcmodclose(q, flag, crp)
	register queue_t *q;
	int flag;
	cred_t *crp;
{
	struct rpcm *rmp;

	ASSERT(q != NULL);
	rmp = (struct rpcm *)q->q_ptr;

	/*
	 * Mark our state as closing.
	 */
	mutex_enter(&rmp->rm_lock);
	rmp->rm_state |= RM_CLOSING;

	/*
	 * If any threads are blocked on flow control, force them
	 * wake up, and send their messages, regardless whether the downstream
	 * module is ready to accept data.
	 */
	if (rmp->rm_flowed)
		qenable(WR(q));

	/*
	 * Block while there are kRPC threads with a reference to
	 * this message.
	 *
	 * Wake up all svc_run threads
	 */
	cv_broadcast(&rmp->rm_qcv);
	while (rmp->rm_ref) {
		cv_wait(&rmp->rm_cwait, &rmp->rm_lock);
	}
	mutex_exit(&rmp->rm_lock);

	/*
	 * It now safe to remove this queue from the stream. No kRPC threads
	 * have a reference to the stream, and none ever will, because
	 * RM_CLOSING is set.
	 */
	del_rpcm(rmp);
	qprocsoff(q);
	q->q_ptr = NULL;
	WR(q)->q_ptr = NULL;
	mutex_destroy(&rmp->rm_lock);
	cv_destroy(&rmp->rm_fwait);
	cv_destroy(&rmp->rm_cwait);
	cv_destroy(&rmp->rm_qcv);
	kmem_free((char *)rmp, sizeof (*rmp));
	return (0);
}

#ifdef	DEBUG
int	rpcmod_send_msg_up = 0;
int	rpcmod_send_uderr = 0;
int	rpcmod_send_dup = 0;
int	rpcmod_send_dup_cnt = 0;
#endif
/*
 * rpcmodrput -	Module read put procedure.  This is called from
 *		the module, driver, or stream head downstream.
 */
rpcmodrput(q, mp)
register queue_t *q;
register mblk_t *mp;
{
	struct rpcm *rmp;
	union T_primitives *pptr;

	TRACE_0(TR_FAC_KRPC, TR_RPCMODRPUT_START,
		"rpcmodrput_start:");

	ASSERT(q != NULL);
	rmp = (struct rpcm *)q->q_ptr;

#ifdef	DEBUG
	if (rpcmod_send_msg_up > 0) {
		mblk_t *nmp = copymsg(mp);
		if (nmp) {
			putnext(q, nmp);
			rpcmod_send_msg_up--;
		}
	}
	if ((rpcmod_send_uderr > 0) && mp->b_datap->db_type == M_PROTO) {
		mblk_t *nmp;
		struct T_unitdata_ind *data;
		struct T_uderror_ind *ud;
		int d;
		data = (struct T_unitdata_ind *)mp->b_rptr;
		if (data->PRIM_type == T_UNITDATA_IND) {
			d = sizeof (*ud) - sizeof (*data);
			nmp = allocb(mp->b_wptr - mp->b_rptr + d, BPRI_HI);
			if (nmp) {
				ud = (struct T_uderror_ind *)nmp->b_rptr;
				ud->PRIM_type = T_UDERROR_IND;
				ud->DEST_length = data->SRC_length;
				ud->DEST_offset = data->SRC_offset + d;
				ud->OPT_length = data->OPT_length;
				ud->OPT_offset = data->OPT_offset + d;
				ud->ERROR_type = ENETDOWN;
				if (data->SRC_length) {
					bcopy((caddr_t)mp->b_rptr
						+ data->SRC_offset,
						(caddr_t)nmp->b_rptr
						+ ud->DEST_offset,
						(size_t)data->SRC_length);
				}
				if (data->OPT_length) {
					bcopy((caddr_t)mp->b_rptr
						+ data->OPT_offset,
						(caddr_t)nmp->b_rptr
						+ ud->OPT_offset,
						(size_t)data->OPT_length);
				}
				nmp->b_wptr += d;
				nmp->b_wptr += (mp->b_wptr - mp->b_rptr);
				nmp->b_datap->db_type = M_PROTO;
				putnext(q, nmp);
				rpcmod_send_uderr--;
			}
		}
	}
#endif
	switch (mp->b_datap->db_type) {
	default:
		putnext(q, mp);
		break;

	case M_PROTO:
	case M_PCPROTO:
		ASSERT((mp->b_wptr - mp->b_rptr) >= sizeof (long));
		pptr = (union T_primitives *)mp->b_rptr;

		/*
		 * Forward this message to krpc if it is data.
		 */
		if (pptr->type == T_UNITDATA_IND) {
#ifdef	DEBUG
			mblk_t	*nmp;
#endif

			/*
			 * Check if the module is being popped.
			 */
			mutex_enter(&rmp->rm_lock);
			if (rmp->rm_state & RM_CLOSING) {
				mutex_exit(&rmp->rm_lock);
				putnext(q, mp);
				break;
			}
#ifdef	DEBUG
			/*
			 * Test duplicate request cache and rm_ref count
			 * handling by sending a duplicate every so often,
			 * if desired.
			 */
			if (rpcmod_send_dup &&
			    rpcmod_send_dup_cnt++ % rpcmod_send_dup)
				nmp = copymsg(mp);
			else
				nmp = NULL;
#endif

			/*
			 * Raise the reference count on this module to
			 * prevent it from being popped before krpc generates
			 * the reply.
			 */
			rmp->rm_ref++;
			mutex_exit(&rmp->rm_lock);

			/*
			 * Submit the message to krpc.
			 */
			svc_queuereq(q, mp);
#ifdef	DEBUG
			/*
			 * Send duplicate if we created one.
			 */
			if (nmp) {
				mutex_enter(&rmp->rm_lock);
				rmp->rm_ref++;
				mutex_exit(&rmp->rm_lock);
				svc_queuereq(q, nmp);
			}
#endif
			break;
		}
		putnext(q, mp);
		break;
	}
	TRACE_0(TR_FAC_KRPC, TR_RPCMODRPUT_END,
		"rpcmodrput_end:");
#ifdef	lint
	/*
	 * Return codes are not looked at by the STREAMS framework.
	 */
	return (0);
#endif
}

/*
 * Module write service procedure. This is called by downstream modules
 * for back enabling during flow control.
 */
rpcmodwsrv(q)
	queue_t *q;
{
	struct rpcm *rmp;

	rmp = (struct rpcm *)q->q_ptr;
	mutex_enter(&rmp->rm_lock);

	/*
	 * If any kRPC threads are blocked in flow control,
	 * wake them up.
	 */
	if (rmp->rm_flowed) {
		cv_broadcast(&rmp->rm_fwait);
	}

	mutex_exit(&rmp->rm_lock);

#ifdef	lint
	/*
	 * Return codes are not looked at by the STREAMS framework.
	 */
	return (0);
#endif
}

/*
 * Entry point for krpc into rpcmod to send response.
 */
static int
rpcmod_send(q, bp)
	register queue_t *q;
	register mblk_t *bp;
{
	register struct rpcm *rmp;
	register int locked;

	rmp = (struct rpcm *)q->q_ptr;

	locked = 0;

	/*
	 * Block untill downstream flow control has ebbed.
	 */
	/*CONSTANTCONDITION*/
	while (1) {

		/*
		 * The probability is that canputnext() will succeed,
		 * so on the first iteration we do the check without a lock.
		 */
		if (canputnext(q))
			break;

		/*
		 * canputnext() failed. Acquire the lock if not
		 * held.
		 */
		if (!locked) {
			locked = 1;
			mutex_enter(&rmp->rm_lock);

			/*
			 * We have to recheck the flow control
			 * because the back enable could have happenned
			 * before we go to cv_wait().
			 */
			if (canputnext(q))
				break;
		}

		/*
		 * If somebody is trying to pop rpcmod, don't bother
		 * blocking in flow control. Send the response now.
		 * We don't need to do this under a lock, and so could
		 * do this check sooner. However, nothing is gained in
		 * optimizing the infrequent kRPC handle tear down state.
		 */
		if (rmp->rm_state & RM_CLOSING)
			break;

		/*
		 * Mark this thread as blocked on flow control.
		 */
		rmp->rm_flowed++;
		cv_wait(&rmp->rm_fwait, &rmp->rm_lock);

		/*
		 * We've been signaled, so flow control has apparently
		 * ebbed.
		 */
		rmp->rm_flowed--;
	}

	/*
	 * Release the mutex if we entered it.
	 */
	if (locked) {
		mutex_exit(&rmp->rm_lock);
	}

	/*
	 * Send the response.
	 */
	putnext(q, bp);

	/*
	 * Return success to the caller.
	 */
	return (0);
}

static void
rpcmod_release(q, bp)
	register queue_t *q;
	register mblk_t *bp;
{
	struct rpcm *rmp;

	/*
	 * For now, just free the message.
	 */
	if (bp)
		freemsg(bp);
	rmp = (struct rpcm *)q->q_ptr;

	mutex_enter(&rmp->rm_lock);
	rmp->rm_ref--;

	if (rmp->rm_ref == 0 && (rmp->rm_state & RM_CLOSING)) {
		cv_broadcast(&rmp->rm_cwait);
	}

	mutex_exit(&rmp->rm_lock);
}
