/*	Copyright (c) 1984, 1986, 1987, 1988, 1989 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"@(#)svc.c	1.31	94/04/25 SMI"	/* SVr4.0 1.9	*/

/*
 *  		PROPRIETARY NOTICE (Combined)
 *
 *  This source code is unpublished proprietary information
 *  constituting, or derived under license from AT&T's Unix(r) System V.
 *  In addition, portions of such source code were derived from Berkeley
 *  4.3 BSD under license from the Regents of the University of
 *  California.
 *
 *
 *
 *  		Copyright Notice
 *
 *  Notice of copyright on this source code product does not indicate
 *  publication.
 *
 *  	(c) 1986, 1987, 1988, 1989  Sun Microsystems, Inc.
 *  	(c) 1983, 1984, 1985, 1986, 1987, 1988, 1989  AT&T.
 *		All rights reserved.
 */

/*
 * svc.c, Server-side remote procedure call interface.
 *
 * There are two sets of procedures here.  The xprt routines are
 * for handling transport handles.  The svc routines handle the
 * list of service routines.
 *
 */

#include <sys/param.h>
#include <sys/types.h>
#include <rpc/types.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <rpc/xdr.h>
#include <rpc/auth.h>
#include <rpc/clnt.h>
#include <rpc/rpc_msg.h>
#include <sys/tiuser.h>
#include <sys/t_kuser.h>
#include <rpc/svc.h>
#include <rpc/svc_auth.h>

caddr_t rqcred_head;  /* head of cached, free authentication parameters */
extern kmutex_t	rqcred_lock;

#include <sys/proc.h>
#include <sys/user.h>
#include <sys/stream.h>
#include <sys/strsubr.h>
#include <sys/tihdr.h>
#include <sys/debug.h>
#include <sys/cmn_err.h>
#include <sys/file.h>
#include <sys/systm.h>
#include <sys/callb.h>

#include <sys/vtrace.h>


#define	NULL_SVC ((struct svc_callout *)0)
#define	RQCRED_SIZE	400		/* this size is excessive */

#define	SVC_VERSQUIET 0x0001		/* keept quiet about vers mismatch */
#define	version_keepquiet(xp)  ((u_long)(xp)->xp_p3 & SVC_VERSQUIET)

static SVCXPRT *xprt_head = NULL;
extern krwlock_t xprt_lock;
extern krwlock_t rpc_list_lock;

#define	RPC_THREAD_TIMEOUT	(5 * HZ)
long rpc_thread_timo;

/*
 * The services list
 * Each entry represents a set of procedures (an rpc program).
 * The dispatch routine takes request structs and runs the
 * apropriate procedure.
 */
static struct svc_callout {
	struct svc_callout *sc_next;
	u_long		    sc_prog;
	u_long		    sc_vers;
	void		    (*sc_dispatch)();
} *svc_head;
extern krwlock_t	svc_lock;

int (*rpc_send)(queue_t *, mblk_t *);
void (*rpc_rele)(queue_t *, mblk_t *);


static struct svc_callout *svc_find(u_long, u_long, struct svc_callout **);
static void svc_run(struct stdata *);


#ifdef TRACE
/*
 * Counter to track queue depth (overall).
 */
long rpc_queued_req;

/*
 * Only for NFS now.
 */
static char *rfscallnames_v2[] = {
	"RFS2_NULL",
	"RFS2_GETATTR",
	"RFS2_SETATTR",
	"RFS2_ROOT",
	"RFS2_LOOKUP",
	"RFS2_READLINK",
	"RFS2_READ",
	"RFS2_WRITECACHE",
	"RFS2_WRITE",
	"RFS2_CREATE",
	"RFS2_REMOVE",
	"RFS2_RENAME",
	"RFS2_LINK",
	"RFS2_SYMLINK",
	"RFS2_MKDIR",
	"RFS2_RMDIR",
	"RFS2_READDIR",
	"RFS2_STATFS"
};

#endif

static struct rpcm *rpcm_list = NULL;

void
add_rpcm(rm)
	struct rpcm *rm;
{
	struct rpcm *trm;

	rw_enter(&rpc_list_lock, RW_WRITER);
	for (trm = rpcm_list; trm != NULL; trm = trm->rm_next) {
		if (trm == rm) {
			cmn_err(CE_WARN, "add_rpcm: rpcmod pushed twice");
			rw_exit(&rpc_list_lock);
			return;
		}
	}
	rm->rm_next = rpcm_list;
	rpcm_list = rm;
	rw_exit(&rpc_list_lock);
	return;
}

void
del_rpcm(rm)
	struct rpcm *rm;
{
	struct rpcm *trm;
	struct rpcm *prev;

	rw_enter(&rpc_list_lock, RW_WRITER);
	prev = rpcm_list;
	for (trm = rpcm_list; trm != NULL; prev = trm, trm = trm->rm_next) {
		if (trm == rm) {
			if (trm == rpcm_list)
				rpcm_list = trm->rm_next;
			else
				prev->rm_next = trm->rm_next;
			break;
		}
	}
	rw_exit(&rpc_list_lock);
	return;
}

struct rpcm *
find_rpcm(str)
	struct stdata *str;
{
	struct rpcm *rm;

	rw_enter(&rpc_list_lock, RW_READER);
	for (rm = rpcm_list; rm != NULL; rm = rm->rm_next) {
		if (rm->rm_stream == str) {
			mutex_enter(&rm->rm_lock);
			rw_exit(&rpc_list_lock);
			return (rm);
		}
	}
	rw_exit(&rpc_list_lock);
	return (NULL);
}

/* ***************  SVCXPRT related stuff **************** */

/*
 * Activate a transport handle.
 * Called by svc_clts_kcreate() to add server xprt handle to
 * global list.
 */
void
xprt_register(xprt)
	const SVCXPRT *xprt;
{
	struct rpcm *rmp;

	rw_enter(&xprt_lock, RW_WRITER);

	/*
	 * Add xprt handle to list, last one added is at the head.
	 */
	((SVCXPRT *)xprt)->xp_next = xprt_head;
	xprt_head = (SVCXPRT *) xprt;

	if ((rmp = find_rpcm(xprt->xp_tiptr->fp->f_vnode->v_stream)) == NULL) {
		cmn_err(CE_WARN, "xprt_register: rpcmod not pushed");
		rw_exit(&xprt_lock);
		return;
	}

	/*
	 * If this is the first xprt handle, create an initial server thread
	 * which will be responsible for creating extra threads as needed.
	 */
	if (rmp->rm_xprt_num++ == 0) {
		rpc_thread_timo = RPC_THREAD_TIMEOUT;
		if (thread_create(NULL, NULL, svc_run,
				(caddr_t)xprt->xp_tiptr->fp->f_vnode->v_stream,
				0, &p0, TS_RUN, 60) != NULL) {
			rmp->rm_svc_threads++;
		} else {
			cmn_err(CE_WARN, "xprt_register: thread_create failed");
		}
	}
	mutex_exit(&rmp->rm_lock);
	rw_exit(&xprt_lock);
}

/*
 * De-activate a transport handle.
 * Called from SVC_DESTROY() to remove server's handle from global list.
 */
void
xprt_unregister(xprt)
	const SVCXPRT *xprt;
{
	register SVCXPRT *xp, *xprev;
	struct rpcm *rmp;

	xprev = NULL;
	rw_enter(&xprt_lock, RW_WRITER);
	for (xp = xprt_head; xp != NULL; xp = xp->xp_next) {
		if (xp == xprt) {
			if (xprev != NULL) {
				xprev->xp_next = xp->xp_next;
			} else {
				xprt_head = xp->xp_next;
			}
			if ((rmp = find_rpcm(
				    xprt->xp_tiptr->fp->f_vnode->v_stream))
					== NULL) {
				cmn_err(CE_WARN,
					"xprt_unregister: rpcmod not pushed");
				rw_exit(&xprt_lock);
				return;
			}

			rmp->rm_xprt_num--;
			mutex_exit(&rmp->rm_lock);
			rw_exit(&xprt_lock);
			return;
		}
		xprev = xp;
	}
	rw_exit(&xprt_lock);
	cmn_err(CE_PANIC, "xprt_unregister: service handle not found");
}

/*
 * Called by svc_run to match the stream head pointer in the request
 * message to an available xprt handle.
 * Searches for one that matches and is not in use (i.e. not
 * locked).
 */
static SVCXPRT *
findxprt(str)
	struct stdata *str;
{
	SVCXPRT *xprt;

	rw_enter(&xprt_lock, RW_READER);
	for (xprt = xprt_head; xprt != NULL; xprt = xprt->xp_next) {
		if (xprt->xp_tiptr->fp->f_vnode->v_stream == str &&
		    mutex_tryenter(&xprt->xp_lock)) {
			rw_exit(&xprt_lock);
			return (xprt);
		}
	}
	rw_exit(&xprt_lock);
	return (NULL);
}

/* ********************** CALLOUT list related stuff ************* */

/*
 * Add a service program to the callout list.
 * The dispatch routine will be called when a rpc request for this
 * program number comes in.
 */
/*ARGSUSED*/
bool_t
svc_register(xprt, prog, vers, dispatch, protocol)
	SVCXPRT *xprt;
	u_long prog;
	u_long vers;
	void (*dispatch)();
	int protocol;
{
	struct svc_callout *prev;
	register struct svc_callout *s;

	rw_enter(&svc_lock, RW_WRITER);
	if ((s = svc_find(prog, vers, &prev)) != NULL_SVC) {
		if (s->sc_dispatch == dispatch)
			goto pmap_it;  /* he is registering another xptr */
		rw_exit(&svc_lock);
		return (FALSE);
	}
	s = (struct svc_callout *)mem_alloc(sizeof (struct svc_callout));
	s->sc_prog = prog;
	s->sc_vers = vers;
	s->sc_dispatch = dispatch;
	s->sc_next = svc_head;
	svc_head = s;
pmap_it:
	rw_exit(&svc_lock);
	return (TRUE);
}

/*
 * Remove a service program from the callout list.
 */
void
svc_unregister(prog, vers)
	u_long prog;
	u_long vers;
{
	struct svc_callout *prev;
	register struct svc_callout *s;

	rw_enter(&svc_lock, RW_WRITER);
	if ((s = svc_find(prog, vers, &prev)) == NULL_SVC) {
		rw_exit(&svc_lock);
		return;
	}
	if (prev == NULL_SVC) {
		svc_head = s->sc_next;
	} else {
		prev->sc_next = s->sc_next;
	}
	s->sc_next = NULL_SVC;
	rw_exit(&svc_lock);

	mem_free((char *) s, (u_int) sizeof (struct svc_callout));
}

/*
 * Search the callout list for a program number, return the callout
 * struct.
 */
static struct svc_callout *
svc_find(prog, vers, prev)
	u_long prog;
	u_long vers;
	struct svc_callout **prev;
{
	register struct svc_callout *s, *p;

	ASSERT(RW_WRITE_HELD(&svc_lock));
	p = NULL_SVC;
	for (s = svc_head; s != NULL_SVC; s = s->sc_next) {
		if ((s->sc_prog == prog) && (s->sc_vers == vers))
			goto done;
		p = s;
	}
done:
	*prev = p;
	return (s);
}

/* ******************* REPLY GENERATION ROUTINES  ************ */

/*
 * Send a reply to an rpc request
 */
bool_t
svc_sendreply(xprt, xdr_results, xdr_location)
	const  SVCXPRT *xprt;
	xdrproc_t xdr_results;
	caddr_t xdr_location;
{
	struct rpc_msg rply;

	rply.rm_direction = REPLY;
	rply.rm_reply.rp_stat = MSG_ACCEPTED;
	rply.acpted_rply.ar_verf = xprt->xp_verf;
	rply.acpted_rply.ar_stat = SUCCESS;
	rply.acpted_rply.ar_results.where = xdr_location;
	rply.acpted_rply.ar_results.proc = xdr_results;

	return (SVC_REPLY((SVCXPRT *) xprt, &rply));
}

/*
 * No procedure error reply
 */
void
svcerr_noproc(xprt)
	const  SVCXPRT *xprt;
{
	struct rpc_msg rply;

	rply.rm_direction = REPLY;
	rply.rm_reply.rp_stat = MSG_ACCEPTED;
	rply.acpted_rply.ar_verf = xprt->xp_verf;
	rply.acpted_rply.ar_stat = PROC_UNAVAIL;
	SVC_FREERES((SVCXPRT *) xprt);
	SVC_REPLY((SVCXPRT *) xprt, &rply);
}

/*
 * Can't decode args error reply
 */
void
svcerr_decode(xprt)
	const SVCXPRT *xprt;
{
	struct rpc_msg rply;

	rply.rm_direction = REPLY;
	rply.rm_reply.rp_stat = MSG_ACCEPTED;
	rply.acpted_rply.ar_verf = xprt->xp_verf;
	rply.acpted_rply.ar_stat = GARBAGE_ARGS;
	SVC_FREERES((SVCXPRT *) xprt);
	SVC_REPLY((SVCXPRT *) xprt, &rply);
}

/*
 * Some system error
 */
void
svcerr_systemerr(xprt)
	const SVCXPRT *xprt;
{
	struct rpc_msg rply;

	rply.rm_direction = REPLY;
	rply.rm_reply.rp_stat = MSG_ACCEPTED;
	rply.acpted_rply.ar_verf = xprt->xp_verf;
	rply.acpted_rply.ar_stat = SYSTEM_ERR;
	SVC_FREERES((SVCXPRT *) xprt);
	SVC_REPLY((SVCXPRT *) xprt, &rply);
}

/*
 * Authentication error reply
 */
void
svcerr_auth(xprt, why)
	const SVCXPRT *xprt;
	enum auth_stat why;
{
	struct rpc_msg rply;

	rply.rm_direction = REPLY;
	rply.rm_reply.rp_stat = MSG_DENIED;
	rply.rjcted_rply.rj_stat = AUTH_ERROR;
	rply.rjcted_rply.rj_why = why;
	SVC_FREERES((SVCXPRT *) xprt);
	SVC_REPLY((SVCXPRT *) xprt, &rply);
}

/*
 * Auth too weak error reply
 */
void
svcerr_weakauth(xprt)
	const SVCXPRT *xprt;
{

	svcerr_auth((SVCXPRT *) xprt, AUTH_TOOWEAK);
}

/*
 * Program unavailable error reply
 */
void
svcerr_noprog(xprt)
	const SVCXPRT *xprt;
{
	struct rpc_msg rply;

	rply.rm_direction = REPLY;
	rply.rm_reply.rp_stat = MSG_ACCEPTED;
	rply.acpted_rply.ar_verf = xprt->xp_verf;
	rply.acpted_rply.ar_stat = PROG_UNAVAIL;
	SVC_FREERES((SVCXPRT *) xprt);
	SVC_REPLY((SVCXPRT *) xprt, &rply);
}

/*
 * Program version mismatch error reply
 */
void
svcerr_progvers(xprt, low_vers, high_vers)
	const SVCXPRT *xprt;
	u_long low_vers;
	u_long high_vers;
{
	struct rpc_msg rply;

	rply.rm_direction = REPLY;
	rply.rm_reply.rp_stat = MSG_ACCEPTED;
	rply.acpted_rply.ar_verf = xprt->xp_verf;
	rply.acpted_rply.ar_stat = PROG_MISMATCH;
	rply.acpted_rply.ar_vers.low = low_vers;
	rply.acpted_rply.ar_vers.high = high_vers;
	SVC_FREERES((SVCXPRT *) xprt);
	SVC_REPLY((SVCXPRT *) xprt, &rply);
}

/* ******************* SERVER INPUT STUFF ******************* */

/*
 * Get server side input from some transport.
 *
 * Statement of authentication parameters management:
 * This function owns and manages all authentication parameters, specifically
 * the "raw" parameters (msg.rm_call.cb_cred and msg.rm_call.cb_verf) and
 * the "cooked" credentials (rqst->rq_clntcred).
 * However, this function does not know the structure of the cooked
 * credentials, so it make the following assumptions:
 *   a) the structure is contiguous (no pointers), and
 *   b) the cred structure size does not exceed RQCRED_SIZE bytes.
 * In all events, all three parameters are freed upon exit from this routine.
 * The storage is trivially managed on the call stack in user land, but
 * is malloced in kernel land.
 */

static void
svc_getreq(xprt, mp)
	SVCXPRT *xprt;
	mblk_t *mp;
{
	struct rpc_msg msg;
	int prog_found;
	u_long low_vers;
	u_long high_vers;
	struct svc_req r;
	char *cred_area;  /* too big to allocate on call stack */

	TRACE_0(TR_FAC_KRPC, TR_SVC_GETREQ_START,
		"svc_getreq_start:");

	/*
	 * Firstly, allocate the authentication parameters' storage
	 */
	mutex_enter(&rqcred_lock);
	if (rqcred_head) {
		cred_area = rqcred_head;
		/* LINTED pointer alignment */
		rqcred_head = *(caddr_t *)rqcred_head;
		mutex_exit(&rqcred_lock);
	} else {
		mutex_exit(&rqcred_lock);
		cred_area = (char *)mem_alloc(2*MAX_AUTH_BYTES + RQCRED_SIZE);
	}
	msg.rm_call.cb_cred.oa_base = cred_area;
	msg.rm_call.cb_verf.oa_base = &(cred_area[MAX_AUTH_BYTES]);
	r.rq_clntcred = &(cred_area[2*MAX_AUTH_BYTES]);

	/* now receive msgs from xprtprt (support batch calls) */
	do {
		TRACE_0(TR_FAC_KRPC, TR_SVC_GETREQ_LOOP_START,
			"svc_getreq_loop_start:");

		if (SVC_RECV(xprt, mp, &msg)) {

			/* now find the exported program and call it */
			register struct svc_callout *s;
			enum auth_stat why;

			r.rq_xprt = xprt;
			r.rq_prog = msg.rm_call.cb_prog;
			r.rq_vers = msg.rm_call.cb_vers;
			r.rq_proc = msg.rm_call.cb_proc;
			r.rq_cred = msg.rm_call.cb_cred;
			/* first authenticate the message */
			TRACE_0(TR_FAC_KRPC, TR_SVC_GETREQ_AUTH_START,
				"svc_getreq_auth_start:");
			if ((why = _authenticate(&r, &msg)) != AUTH_OK) {
				TRACE_1(TR_FAC_KRPC, TR_SVC_GETREQ_AUTH_END,
					"svc_getreq_auth_end:(%S)", "failed");
				svcerr_auth(xprt, why);
				/*
				 * Free the arguments
				 */
				(void) SVC_FREEARGS(xprt, (xdrproc_t)0,
						    (caddr_t)0);
				continue;
			}
			TRACE_1(TR_FAC_KRPC, TR_SVC_GETREQ_AUTH_END,
				"svc_getreq_auth_end:(%S)", "good");
			/* now match message with a registered service */
			prog_found = FALSE;
			low_vers = (u_long)-1;
			high_vers = 0;
			rw_enter(&svc_lock, RW_READER);
			for (s = svc_head; s != NULL_SVC; s = s->sc_next) {
				if (s->sc_prog == r.rq_prog) {
					if (s->sc_vers == r.rq_vers) {
						(*s->sc_dispatch)(&r, xprt);
						break;
					}  /* found correct version */
					prog_found = TRUE;
					if (s->sc_vers < low_vers)
						low_vers = s->sc_vers;
					if (s->sc_vers > high_vers)
						high_vers = s->sc_vers;
				}   /* found correct program */
			}
			rw_exit(&svc_lock);
			if (s == NULL_SVC) {
				/*
				 * if we got here, the program or version
				 * is not served ...
				 */
				if (prog_found && !version_keepquiet(xprt)) {
					svcerr_progvers(xprt, low_vers,
							high_vers);
				} else {
					svcerr_noprog(xprt);
				}
				/*
				 * Free the arguments.  This is done
				 * by the dispatch routine for successful
				 * calls.
				 */
				(void) SVC_FREEARGS(xprt, (xdrproc_t)0,
						    (caddr_t)0);
				/* Fall through to ... */
			}
		}
		TRACE_0(TR_FAC_KRPC, TR_SVC_GETREQ_LOOP_END,
			"svc_getreq_loop_end:");
	} while (SVC_STAT(xprt) == XPRT_MOREREQS);

	/*
	 * free authentication parameters' storage
	 */
	mutex_enter(&rqcred_lock);
	/* LINTED pointer alignment */
	*(caddr_t *)cred_area = rqcred_head;
	rqcred_head = cred_area;
	mutex_exit(&rqcred_lock);
}

extern int ncpus;

int Rpccnt;

/*
 * This is the rpc server side idle loop
 * Wait for input, call server program.
 */
static void
svc_run(str)
	struct stdata *str;
{
	SVCXPRT *xprt;
	struct rpcm *rmp;
	struct msgb *mp;
	int timeleft;
	callb_cpr_t cprinfo;
	void svc_getreq(SVCXPRT *, mblk_t *);

	if ((xprt = findxprt(str)) == NULL)  {
		/*
		 * If we can't get a transport then we have to back off the
		 * rm_svc_threads count.  If we can't get the rpcm then the
		 * rpcmod count is gone anyways.
		 */
		if ((rmp = find_rpcm(str)) != NULL) {
			rmp->rm_svc_threads--;
			mutex_exit(&rmp->rm_lock);
		} else {
			cmn_err(CE_WARN, "svc_run: rpcmod not pushed");
		}
		thread_exit();
		/* NOTREACHED */
	}

	ASSERT(MUTEX_HELD(&xprt->xp_lock));

	if ((rmp = find_rpcm(str)) == NULL) {
		/*
		 * If we can't get a rpcm then either rpcmod has not been
		 * pushed or we are in bad shape.
		 */
		mutex_exit(&xprt->xp_lock);
		cmn_err(CE_WARN, "svc_run: rpcmod not pushed");
		thread_exit();
		/* NOTREACHED */
	}

	ASSERT(MUTEX_HELD(&rmp->rm_lock));

	rmp->rm_ref++;

	xprt->xp_wq = rmp->rm_wq;

	CALLB_CPR_INIT(&cprinfo, &rmp->rm_qlock, callb_generic_cpr);

	do {
		ASSERT(MUTEX_HELD(&rmp->rm_lock));
		mutex_exit(&rmp->rm_lock);
		/*
		 * for loop iterates until the thread becomes
		 * idle too long.
		 */
		for (;;) {
			TRACE_0(TR_FAC_KRPC, TR_SVC_RUN,
				"svc_run");
			/*
			 * Inner while loop iterates until we
			 * get a request.
			 */
			timeleft = 1;
			mutex_enter(&rmp->rm_qlock);
			while (rmp->rm_qfirst == NULL) {
				/*
				 * If time expired the last time we
				 * blocked, or if all services have
				 * terminated, exit outer for loop.
				 */
				if (timeleft <= 0 || rmp->rm_xprt_num == 0) {
					ASSERT(MUTEX_HELD(&rmp->rm_qlock));
					mutex_exit(&rmp->rm_qlock);
					goto again;
				}

				/*
				 * No work to do, count us as asleep,
				 * and wait for a request.
				 */
				ASSERT(rmp->rm_asleep >= 0);
				rmp->rm_asleep++;
				CALLB_CPR_SAFE_BEGIN(&cprinfo);
				timeleft = cv_timedwait(&rmp->rm_qcv,
						&rmp->rm_qlock,
						rpc_thread_timo + lbolt);
				CALLB_CPR_SAFE_END(&cprinfo, &rmp->rm_qlock);
				ASSERT(MUTEX_HELD(&rmp->rm_qlock));

				/*
				 * If drowsy == 0 then we have timed out
				 * before a cv_signal.
				 */
				if (rmp->rm_drowsy == 0) {
					rmp->rm_asleep--;
				} else {
					rmp->rm_drowsy--;
				}
				ASSERT(rmp->rm_asleep >= 0);
				ASSERT(rmp->rm_drowsy >= 0);

				/*
				 * If RM_CLOSING is set rpcmod is being
				 * closed.  We have a reference to to
				 * rpcmod so we need to exit.
				 */
				if (rmp->rm_state & RM_CLOSING) {
					mutex_exit(&rmp->rm_qlock);
					goto again;
				}
			}

			/*
			 * De-queue request.
			 */
			mp = rmp->rm_qfirst;
			rmp->rm_qfirst = mp->b_next;
			if (rmp->rm_qfirst == NULL)
				rmp->rm_qlast = NULL;
			else
				rmp->rm_qfirst->b_prev = NULL;
			mp->b_next = NULL;
			mp->b_prev = NULL;
#ifdef TRACE
			/*
			 * HACK: I assume NFS v2, and good packet.
			 */
			rpc_queued_req--;
			TRACE_4(TR_FAC_KRPC, TR_NFSFP_QUE_REQ_DEQ,
		"rpc_que_req_deq:entries on que %d (%S) proc_num %d xid %x",
			    rpc_queued_req,
			    rfscallnames_v2[((struct rpc_msg *)
				mp->b_cont->b_rptr)->rm_call.cb_proc],
				((struct rpc_msg *)mp->b_cont->b_rptr)
				->rm_call.cb_proc,
			    ((struct rpc_msg *)mp->b_cont->b_rptr)->rm_xid);
#endif
			/*
			 * If I was the only drowsy thread and there are
			 * more requests on the queue, then wake one up.
			 */
			ASSERT(rmp->rm_drowsy >= 0);
			ASSERT(rmp->rm_asleep >= 0);
			if ((rmp->rm_qfirst != NULL) &&
			    (rmp->rm_asleep != 0 && rmp->rm_drowsy < 1)) {
				cv_signal(&rmp->rm_qcv);
				ASSERT(rmp->rm_asleep > 0);
				rmp->rm_asleep--;
				rmp->rm_drowsy++;
			}
			mutex_exit(&rmp->rm_qlock);

			/*
			 * If there are no idle threads, we need more if
			 * we haven't reached the fixed limit.
			 */
			if ((rmp->rm_asleep + rmp->rm_drowsy) == 0 &&
			    rmp->rm_svc_threads < rmp->rm_xprt_num) {
				mutex_enter(&rmp->rm_lock);
				if ((rmp->rm_svc_threads < rmp->rm_xprt_num) &&
				    thread_create(NULL, NULL, svc_run,
						(caddr_t)rmp->rm_stream, 0, &p0,
						TS_RUN, 60) != NULL) {
					rmp->rm_svc_threads++;
				}
				mutex_exit(&rmp->rm_lock);
			}

			/*
			 * Process the request.
			 */
			Rpccnt++;
			ASSERT(MUTEX_HELD(&xprt->xp_lock));
			svc_getreq(xprt, mp);
			/*
			 * Release our reference on the rpcmod
			 * slot attached to xp_wq->q_ptr.
			 */
			(*rpc_rele)(xprt->xp_wq, NULL);
		}
	again:
		ASSERT(MUTEX_HELD(&xprt->xp_lock));
		if (SVC_STAT(xprt) == XPRT_DIED) {
			mutex_exit(&xprt->xp_lock);
			SVC_DESTROY(xprt);
			mutex_enter(&rmp->rm_lock);
			goto cleanup;
		}

		/*
		 * If RM_CLOSING is set rpcmod is being
		 * closed.  We have a reference to to
		 * rpcmod so we need to exit.
		 */
		if (rmp->rm_state & RM_CLOSING) {
			mutex_enter(&rmp->rm_lock);
			break;
		}

		/*
		 * If we get here, this thread has decided it has
		 * been idle too long, so it will normally exit this
		 * loop, and destroy itself. However, if there
		 * is only one thread left, it has to stay present
		 * so that it can fork off more threads in case the
		 * load goes back up. If there are no xprt handles,
		 * even the last thread must die.
		 */
		mutex_enter(&rmp->rm_lock);
	} while (rmp->rm_xprt_num > 0 && rmp->rm_svc_threads == 1);

	mutex_exit(&xprt->xp_lock);

cleanup:
	rmp->rm_svc_threads--;
	mutex_exit(&rmp->rm_lock);

	/*
	 * need to clean up the CPR info which is protected by rm_qlock.
	 */
	mutex_enter(&rmp->rm_qlock);
	CALLB_CPR_SAFE_BEGIN(&cprinfo);
	CALLB_CPR_EXIT(&cprinfo);
	mutex_exit(&rmp->rm_qlock);

	/*
	 * Release our reference on the rpcmod
	 * slot attached to rm_wq->q_ptr.
	 */
	(*rpc_rele)(rmp->rm_wq, NULL);

	thread_exit();
	/* NOTREACHED */
}

void
svc_queuereq(q, mp)
	queue_t *q;
	mblk_t *mp;
{
	struct rpcm *rmp;
	struct msgb *tmp;

	TRACE_0(TR_FAC_KRPC, TR_SVC_QUEUEREQ_START,
		"svc_queuereq_start");

	rmp = (struct rpcm *)q->q_ptr;

	mutex_enter(&rmp->rm_qlock);
	if (!rmp->rm_qfirst) {
		mp->b_next = NULL;
		mp->b_prev = NULL;
		rmp->rm_qfirst = mp;
		rmp->rm_qlast = mp;
	} else {
		tmp = rmp->rm_qlast;
		mp->b_next = NULL;
		mp->b_prev = tmp;
		tmp->b_next = mp;
		rmp->rm_qlast = mp;
	}

#ifdef TRACE
	rpc_queued_req++;
	TRACE_4(TR_FAC_KRPC, TR_NFSFP_QUE_REQ_ENQ,
		"rpc_que_req_enq:entries on que %d (%S) proc_num %d xid %x",
		rpc_queued_req,
		rfscallnames_v2[((struct rpc_msg *)mp->b_cont->b_rptr)
			->rm_call.cb_proc],
		((struct rpc_msg *)mp->b_cont->b_rptr)->rm_call.cb_proc,
		((struct rpc_msg *)mp->b_cont->b_rptr)->rm_xid);
#endif
	/*
	 * If we have sleeping threads and none are pending on the run
	 * queue wake one up.
	 */
	ASSERT(rmp->rm_asleep >= 0);
	ASSERT(rmp->rm_drowsy >= 0);
	if (rmp->rm_asleep != 0 && rmp->rm_drowsy < 1) {
		cv_signal(&rmp->rm_qcv);
		rmp->rm_asleep--;
		rmp->rm_drowsy++;
	}
	mutex_exit(&rmp->rm_qlock);
	TRACE_1(TR_FAC_KRPC, TR_SVC_QUEUEREQ_END,
		"svc_queuereq_end:(%S)", "end");
}
