/*	Copyright (c) 1984, 1986, 1987, 1988, 1989 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"@(#)svcauthdes.c	1.18	94/04/25 SMI"	/* SVr4.0 1.10	*/

/*
 *  		PROPRIETARY NOTICE (Combined)
 *
 *  This source code is unpublished proprietary information
 *  constituting, or derived under license from AT&T's Unix(r) System V.
 *  In addition, portions of such source code were derived from Berkeley
 *  4.3 BSD under license from the Regents of the University of
 *  California.
 *
 *
 *
 *  		Copyright Notice
 *
 *  Notice of copyright on this source code product does not indicate
 *  publication.
 *
 *  	(c) 1986, 1987, 1988, 1989  Sun Microsystems, Inc.
 *  	(c) 1983, 1984, 1985, 1986, 1987, 1988, 1989  AT&T.
 *		  All rights reserved.
 */

/*
 * svcauth_des.c, server-side des authentication
 *
 * We insure for the service the following:
 * (1) The timestamp microseconds do not exceed 1 million.
 * (2) The timestamp plus the window is less than the current time.
 * (3) The timestamp is not less than the one previously
 *	seen in the current session.
 *
 * It is up to the server to determine if the window size is
 * too small.
 *
 */


#include <rpc/des_crypt.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/systm.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <rpc/types.h>
#include <rpc/xdr.h>
#include <rpc/auth.h>
#include <rpc/auth_des.h>
#include <sys/tiuser.h>
#include <sys/tihdr.h>
#include <sys/t_kuser.h>
#include <rpc/rpc_msg.h>
#include <rpc/svc.h>
#include <rpc/svc_auth.h>
#include <rpc/clnt.h>	/* for client-side error codes */
#include <sys/t_lock.h>
#include <sys/debug.h>
#include <sys/time.h>

#ifdef _KERNEL
/*
#include <sys/kernel.h>
#define	gettimeofday(tvp, tzp)	(*(tvp) = hrestime.tv_sec)
*/
#else
#include <sys/syslog.h>
#endif

#define	USEC_PER_SEC ((u_long) 1000000L)
#define	BEFORE(t1, t2) timercmp(t1, t2, <)

/*
 * LRU cache of conversation keys and some other useful items.
 */
#define	AUTHDES_CACHESZ 64
struct cache_entry {
	des_block key;		/* conversation key */
	char *rname;		/* client's name */
	u_int window;		/* credential lifetime window */
	struct timeval laststamp;	/* detect replays of creds */
	char *localcred;	/* generic local credential */
};
static struct cache_entry *authdes_cache /* [AUTHDES_CACHESZ] */;
static short *authdes_lru /* [AUTHDES_CACHESZ] */;

static void	cache_init(void);
static short	cache_victim(void);
static void	cache_ref(short);
static short	cache_spot(des_block *, char *, struct timeval *);
static void	invalidate(char *);

/*
 * cache statistics
 */
static struct {
	u_long ncachehits;	/* times cache hit, and is not replay */
	u_long ncachereplays;	/* times cache hit, and is replay */
	u_long ncachemisses;	/* times cache missed */
} svcauthdes_stats;

/*
 * Service side authenticator for AUTH_DES
 */
enum auth_stat
_svcauth_des(register struct svc_req *rqst, register struct rpc_msg *msg)
{
	register long *ixdr;
	des_block cryptbuf[2];
	register struct authdes_cred *cred;
	struct authdes_verf verf;
	int status;
	register struct cache_entry *entry;
	short sid;
	des_block *sessionkey;
	des_block ivec;
	u_int window;
	struct timeval timestamp;
	u_long namelen;
	struct area {
		struct authdes_cred area_cred;
		char area_netname[MAXNETNAMELEN+1];
	} *area;

	mutex_enter(&authdes_lock);
	if (authdes_cache == NULL) {
		cache_init();
	}
	mutex_exit(&authdes_lock);

	/* LINTED pointer alignment */
	area = (struct area *)rqst->rq_clntcred;
	cred = (struct authdes_cred *)&area->area_cred;

	/*
	 * Get the credential
	 */
	/* LINTED pointer alignment */
	ixdr = (long *)msg->rm_call.cb_cred.oa_base;
	cred->adc_namekind = IXDR_GET_ENUM(ixdr, enum authdes_namekind);
	switch (cred->adc_namekind) {
	case ADN_FULLNAME:
		namelen = IXDR_GET_U_LONG(ixdr);
		if (namelen > MAXNETNAMELEN) {
			return (AUTH_BADCRED);
		}
		cred->adc_fullname.name = area->area_netname;
		bcopy((char *)ixdr, cred->adc_fullname.name,
			(u_int)namelen);
		cred->adc_fullname.name[namelen] = 0;
		ixdr += (RNDUP(namelen) / BYTES_PER_XDR_UNIT);
		cred->adc_fullname.key.key.high = (u_long)*ixdr++;
		cred->adc_fullname.key.key.low = (u_long)*ixdr++;
		cred->adc_fullname.window = (u_long)*ixdr++;
		break;
	case ADN_NICKNAME:
		cred->adc_nickname = (u_long)*ixdr++;
		break;
	default:
		return (AUTH_BADCRED);
	}

	/*
	 * Get the verifier
	 */
	/* LINTED pointer alignment */
	ixdr = (long *)msg->rm_call.cb_verf.oa_base;
	verf.adv_xtimestamp.key.high = (u_long)*ixdr++;
	verf.adv_xtimestamp.key.low =  (u_long)*ixdr++;
	verf.adv_int_u = (u_long)*ixdr++;

	mutex_enter(&authdes_lock);

	/*
	 * Get the conversation key
	 */
	if (cred->adc_namekind == ADN_FULLNAME) {
		sessionkey = &cred->adc_fullname.key;
		if (key_decryptsession(cred->adc_fullname.name, sessionkey) !=
		    RPC_SUCCESS) {
#ifdef	_KERNEL
			RPCLOG(1,
			    "_svcauth_des: key_decryptsessionkey failed\n", 0);
#else
			(void) syslog(LOG_DEBUG,
			    "_svcauth_des:  key_decryptsessionkey failed");
#endif
			mutex_exit(&authdes_lock);
			return (AUTH_BADCRED); /* key not found */
		}
	} else { /* ADN_NICKNAME */
		sid = cred->adc_nickname;
		if (sid >= AUTHDES_CACHESZ) {
#ifdef	_KERNEL
			RPCLOG(1, "_svcauth_des: bad nickname %d\n", sid);
#else
			(void) syslog(LOG_DEBUG, "_svcauth_des:  bad nickname");
#endif
			mutex_exit(&authdes_lock);
			return (AUTH_BADCRED);	/* garbled credential */
		}
		sessionkey = &authdes_cache[sid].key;
	}


	/*
	 * Decrypt the timestamp
	 */
	cryptbuf[0] = verf.adv_xtimestamp;
	if (cred->adc_namekind == ADN_FULLNAME) {
		cryptbuf[1].key.high = cred->adc_fullname.window;
		cryptbuf[1].key.low = verf.adv_winverf;
		ivec.key.high = ivec.key.low = 0;
		status = cbc_crypt((char *)sessionkey, (char *)cryptbuf,
			2*sizeof (des_block), DES_DECRYPT,
			(char *)&ivec);
	} else {
		status = ecb_crypt((char *)sessionkey, (char *)cryptbuf,
			sizeof (des_block), DES_DECRYPT);
	}
	if (DES_FAILED(status)) {
#ifdef	_KERNEL
		RPCLOG(1, "_svcauth_des: decryption failure\n", 0);
#else
		(void) syslog(LOG_DEBUG, "_svcauth_des:  decryption failure");
#endif
		mutex_exit(&authdes_lock);
		return (AUTH_FAILED);	/* system error */
	}

	/*
	 * XDR the decrypted timestamp
	 */
	ixdr = (long *)cryptbuf;
	timestamp.tv_sec = IXDR_GET_LONG(ixdr);
	timestamp.tv_usec = IXDR_GET_LONG(ixdr);

	/*
	 * Check for valid credentials and verifiers.
	 * They could be invalid because the key was flushed
	 * out of the cache, and so a new session should begin.
	 * Be sure and send AUTH_REJECTED{CRED, VERF} if this is the case.
	 */
	{
		struct timeval current;
		int nick;
		int winverf;

		if (cred->adc_namekind == ADN_FULLNAME) {
			window = IXDR_GET_U_LONG(ixdr);
			winverf = IXDR_GET_U_LONG(ixdr);
			if (winverf != window - 1) {
#ifdef	_KERNEL
				RPCLOG(1,
				"_svcauth_des: window verifier mismatch %d\n",
				    winverf);
#else
				(void) syslog(LOG_DEBUG,
				"_svcauth_des:  window verifier mismatch");
#endif
				mutex_exit(&authdes_lock);
				return (AUTH_BADCRED);	/* garbled credential */
			}
			sid = cache_spot(sessionkey, cred->adc_fullname.name,
			    &timestamp);
			if (sid < 0) {
#ifdef	_KERNEL
				RPCLOG(1,
				"_svcauth_des: replayed credential sid %d\n",
				    sid);
#else
				(void) syslog(LOG_DEBUG,
				"_svcauth_des:  replayed credential");
#endif
				mutex_exit(&authdes_lock);
				return (AUTH_REJECTEDCRED);	/* replay */
			}
			nick = 0;
		} else {	/* ADN_NICKNAME */
			window = authdes_cache[sid].window;
			nick = 1;
		}

		if ((u_long)timestamp.tv_usec >= USEC_PER_SEC) {
#ifdef	_KERNEL
			RPCLOG(1, "_svcauth_des: invalid usecs %d\n",
			    timestamp.tv_usec);
#else
			(void) syslog(LOG_DEBUG,
					"_svcauth_des:  invalid usecs");
#endif
			/* cached out (bad key), or garbled verifier */
			mutex_exit(&authdes_lock);
			return (nick ? AUTH_REJECTEDVERF : AUTH_BADVERF);
		}
		if (nick && BEFORE(&timestamp,
				    &authdes_cache[sid].laststamp)) {
#ifdef	_KERNEL
			RPCLOG(1, "_svcauth_des: timestamp before last seen\n",
			    0);

#else
			(void) syslog(LOG_DEBUG,
				"_svcauth_des:  timestamp before last seen");
#endif
			mutex_exit(&authdes_lock);
			return (AUTH_REJECTEDVERF);	/* replay */
		}
/*
		(void) gettimeofday(&current, (struct timezone *)NULL);
*/
		current.tv_sec = hrestime.tv_sec;
		current.tv_usec = hrestime.tv_nsec/1000;

		current.tv_sec -= window;	/* allow for expiration */
		if (!BEFORE(&current, &timestamp)) {
#ifdef	_KERNEL
			RPCLOG(1, "_svcauth_des: timestamp expired\n", 0);
#else
			(void) syslog(LOG_DEBUG,
					"_svcauth_des:  timestamp expired");
#endif
			/* replay, or garbled credential */
			mutex_exit(&authdes_lock);
			return (nick ? AUTH_REJECTEDVERF : AUTH_BADCRED);
		}
	}

	/*
	 * Set up the reply verifier
	 */
	verf.adv_nickname = sid;

	/*
	 * xdr the timestamp before encrypting
	 */
	ixdr = (long *)cryptbuf;
	IXDR_PUT_LONG(ixdr, timestamp.tv_sec - 1);
	IXDR_PUT_LONG(ixdr, timestamp.tv_usec);

	/*
	 * encrypt the timestamp
	 */
	status = ecb_crypt((char *)sessionkey, (char *)cryptbuf,
	    sizeof (des_block), DES_ENCRYPT);
	if (DES_FAILED(status)) {
#ifdef	_KERNEL
		RPCLOG(1, "_svcauth_des: encryption failure\n", 0);
#else
		(void) syslog(LOG_DEBUG, "_svcauth_des:  encryption failure");
#endif
		mutex_exit(&authdes_lock);
		return (AUTH_FAILED);	/* system error */
	}
	verf.adv_xtimestamp = cryptbuf[0];

	/*
	 * Serialize the reply verifier, and update rqst
	 */
	/* LINTED pointer alignment */
	ixdr = (long *)msg->rm_call.cb_verf.oa_base;
	*ixdr++ = (long)verf.adv_xtimestamp.key.high;
	*ixdr++ = (long)verf.adv_xtimestamp.key.low;
	*ixdr++ = (long)verf.adv_int_u;

	rqst->rq_xprt->xp_verf.oa_flavor = AUTH_DES;
	rqst->rq_xprt->xp_verf.oa_base = msg->rm_call.cb_verf.oa_base;
	rqst->rq_xprt->xp_verf.oa_length =
		(char *)ixdr - msg->rm_call.cb_verf.oa_base;

	/*
	 * We succeeded, commit the data to the cache now and
	 * finish cooking the credential.
	 */
	entry = &authdes_cache[sid];
	entry->laststamp = timestamp;
	cache_ref(sid);
	if (cred->adc_namekind == ADN_FULLNAME) {
		cred->adc_fullname.window = window;
		cred->adc_nickname = sid;	/* save nickname */
		if (entry->rname != NULL) {
			mem_free(entry->rname, strlen(entry->rname) + 1);
		}
		entry->rname =
		    (char *)mem_alloc((u_int)strlen(cred->adc_fullname.name)
					    + 1);
		if (entry->rname != NULL) {
			(void) strcpy(entry->rname, cred->adc_fullname.name);
		} else {
#ifdef	_KERNEL
			/* EMPTY */
			RPCLOG(1, "_svcauth_des: out of memory\n", 0);
#else
			(void) syslog(LOG_DEBUG,
					"_svcauth_des:  out of memory");
#endif
		}
		entry->key = *sessionkey;
		entry->window = window;
		invalidate(entry->localcred); /* mark any cached cred invalid */
	} else { /* ADN_NICKNAME */
		/*
		 * nicknames are cooked into fullnames
		 */
		cred->adc_namekind = ADN_FULLNAME;
		cred->adc_fullname.name = entry->rname;
		cred->adc_fullname.key = entry->key;
		cred->adc_fullname.window = entry->window;
	}
	mutex_exit(&authdes_lock);
	return (AUTH_OK);	/* we made it! */
}

/*
 * Initialize the cache
 */
static void
cache_init(void)
{
	register short i;

	ASSERT(MUTEX_HELD(&authdes_lock));
	authdes_cache = (struct cache_entry *)
		mem_alloc(sizeof (struct cache_entry) * AUTHDES_CACHESZ);
	bzero((char *)authdes_cache,
		sizeof (struct cache_entry) * AUTHDES_CACHESZ);

	authdes_lru = (short *)mem_alloc(sizeof (short) * AUTHDES_CACHESZ);
	/*
	 * Initialize the lru list
	 */
	for (i = 0; i < AUTHDES_CACHESZ; i++) {
		authdes_lru[i] = i;
	}
}

/*
 * Find the lru victim
 */
static short
cache_victim(void)
{

	ASSERT(MUTEX_HELD(&authdes_lock));
	return (authdes_lru[AUTHDES_CACHESZ-1]);
}

/*
 * Note that sid was referenced
 */
static void
cache_ref(register short sid)
{
	register int i;
	register short curr;
	register short prev;

	ASSERT(MUTEX_HELD(&authdes_lock));
	prev = authdes_lru[0];
	authdes_lru[0] = sid;
	for (i = 1; prev != sid; i++) {
		curr = authdes_lru[i];
		authdes_lru[i] = prev;
		prev = curr;
	}
}

/*
 * Find a spot in the cache for a credential containing
 * the items given.  Return -1 if a replay is detected, otherwise
 * return the spot in the cache.
 */
static short
cache_spot(register des_block *key, char *name, struct timeval *timestamp)
{
	register struct cache_entry *cp;
	register int i;
	register u_long hi;

	ASSERT(MUTEX_HELD(&authdes_lock));
	hi = key->key.high;
	for (cp = authdes_cache, i = 0; i < AUTHDES_CACHESZ; i++, cp++) {
		if (cp->key.key.high == hi &&
		    cp->key.key.low == key->key.low &&
		    cp->rname != NULL &&
		    bcmp(cp->rname, name, strlen(name) + 1) == 0) {
			if (BEFORE(timestamp, &cp->laststamp)) {
				mutex_enter(&svcauthdesstats_lock);
				svcauthdes_stats.ncachereplays++;
				mutex_exit(&svcauthdesstats_lock);
				return (-1); /* replay */
			}
			mutex_enter(&svcauthdesstats_lock);
			svcauthdes_stats.ncachehits++;
			mutex_exit(&svcauthdesstats_lock);
			return (i);	/* refresh */
		}
	}
	mutex_enter(&svcauthdesstats_lock);
	svcauthdes_stats.ncachemisses++;
	mutex_exit(&svcauthdesstats_lock);
	return (cache_victim());	/* new credential */
}

/*
 * Local credential handling stuff.
 * NOTE: bsd unix dependent.
 * Other operating systems should put something else here.
 */
#define	UNKNOWN 	-2	/* grouplen, if cached cred is unknown user */
#define	INVALID		-1 	/* grouplen, if cache entry is invalid */

struct bsdcred {
	uid_t uid;		/* cached uid */
	gid_t gid;		/* cached gid */
	short grouplen;	/* length of cached groups */
	short groups[NGROUPS_UMAX];	/* cached groups */
};

/*
 * Map a des credential into a unix cred.
 * We cache the credential here so the application does
 * not have to make an rpc call every time to interpret
 * the credential.
 */
int
authdes_getucred(const struct authdes_cred *adc, uid_t *uid, gid_t *gid,
	short *grouplen, register int *groups)
{
	unsigned sid;
	register int i;
	uid_t i_uid;
	gid_t i_gid;
	int i_grouplen;
	struct bsdcred *cred;

	sid = adc->adc_nickname;
	if (sid >= AUTHDES_CACHESZ) {
#ifdef	_KERNEL
		RPCLOG(1, "authdes_getucred:  invalid nickname\n", 0);
#else
		(void) syslog(LOG_DEBUG, "authdes_getucred:  invalid nickname");
#endif
		return (0);
	}
	mutex_enter(&authdes_lock);
	/* LINTED pointer alignment */
	cred = (struct bsdcred *)authdes_cache[sid].localcred;
	if (cred == NULL) {
		cred = (struct bsdcred *)mem_alloc(sizeof (struct bsdcred));
		authdes_cache[sid].localcred = (char *)cred;
		cred->grouplen = INVALID;
	}
	if (cred->grouplen == INVALID) {
		/*
		 * not in cache: lookup
		 */
		if (netname2user(adc->adc_fullname.name, &i_uid, &i_gid,
			&i_grouplen, groups) != RPC_SUCCESS)
		{
#ifdef	_KERNEL
			RPCLOG(1, "authdes_getucred:  unknown netname\n", 0);
#else
			(void) syslog(LOG_DEBUG,
					"authdes_getucred:  unknown netname");
#endif
			cred->grouplen = UNKNOWN;	/* mark as lookup up */
							/* but not found */
			mutex_exit(&authdes_lock);
			return (0);
		}
#ifdef	_KERNEL
		RPCLOG(1, "authdes_getucred:  missed ucred cache\n", 0);
#else
		(void) syslog(LOG_DEBUG,
				"authdes_getucred:  missed ucred cache");
#endif
		*uid = cred->uid = i_uid;
		*gid = cred->gid = i_gid;
		*grouplen = cred->grouplen = (short)i_grouplen;
		for (i = i_grouplen - 1; i >= 0; i--) {
			cred->groups[i] = groups[i]; /* int to short */
		}
		mutex_exit(&authdes_lock);
		return (1);
	} else if (cred->grouplen == UNKNOWN) {
		/*
		 * Already lookup up, but no match found
		 */
		mutex_exit(&authdes_lock);
		return (0);
	}

	/*
	 * cached credentials
	 */
	*uid = cred->uid;
	*gid = cred->gid;
	*grouplen = cred->grouplen;
	for (i = cred->grouplen - 1; i >= 0; i--) {
		groups[i] = cred->groups[i];	/* short to int */
	}
	mutex_exit(&authdes_lock);
	return (1);
}

static void
invalidate(char *cred)
{

	if (cred == NULL) {
		return;
	}
	/* LINTED pointer alignment */
	((struct bsdcred *)cred)->grouplen = INVALID;
}
