/*	Copyright (c) 1984, 1986, 1987, 1988, 1989 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#pragma ident   "@(#)xdr_mblk.c 1.19     94/04/25 SMI"

/*
 *		PROPRIETARY NOTICE (Combined)
 *
 *  This source code is unpublished proprietary information
 *  constituting, or derived under license from AT&T's Unix(r) System V.
 *  In addition, portions of such source code were derived from Berkeley
 *  4.3 BSD under license from the Regents of the University of
 *  California.
 *
 *
 *
 *		Copyright Notice
 *
 *  Notice of copyright on this source code product does not indicate
 *  publication.
 *
 *	(c) 1986, 1987, 1988, 1989  Sun Microsystems, Inc.
 *	(c) 1983, 1984, 1985, 1986, 1987, 1988, 1989  AT&T.
 *		All rights reserved.
 */

#ifdef _KERNEL
/*
 * xdr_mblk.c, XDR implementation on kernel streams mblks.
 */

#include <sys/param.h>
#include <sys/types.h>
#include <rpc/types.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/cmn_err.h>
#include <sys/strsubr.h>
#include <rpc/xdr.h>
#include <netinet/in.h>

bool_t	xdrmblk_getlong(), xdrmblk_putlong();
bool_t	xdrmblk_getbytes(), xdrmblk_putbytes();
u_int	xdrmblk_getpos();
bool_t	xdrmblk_setpos();
long *	xdrmblk_inline();
void	xdrmblk_destroy();
bool_t	xdrmblk_control();

static mblk_t *xdrmblk_alloc();

/*
 * Xdr on mblks operations vector.
 */
struct	xdr_ops xdrmblk_ops = {
	xdrmblk_getlong,
	xdrmblk_putlong,
	xdrmblk_getbytes,
	xdrmblk_putbytes,
	xdrmblk_getpos,
	xdrmblk_setpos,
	xdrmblk_inline,
	xdrmblk_destroy,
	xdrmblk_control
};

/*
 * Initialize xdr stream.
 */
void
xdrmblk_init(xdrs, m, op, sz)
	register XDR *xdrs;
	register mblk_t *m;
	enum xdr_op op;
	int sz;
{

	xdrs->x_op = op;
	xdrs->x_ops = &xdrmblk_ops;
	xdrs->x_base = (caddr_t)m;
	xdrs->x_public = NULL;
	xdrs->x_private = (caddr_t)sz;

	if (op == XDR_DECODE) {
		xdrs->x_handy = m->b_wptr - m->b_rptr;
	} else	{
		xdrs->x_handy = m->b_datap->db_lim - m->b_datap->db_base;
	}
}

void
/* ARGSUSED */
xdrmblk_destroy(xdrs)
	XDR *xdrs;
{
	/* do nothing */
}

bool_t
xdrmblk_getlong(xdrs, lp)
	register XDR *xdrs;
	long *lp;
{
	register mblk_t *m;

	/* LINTED pointer alignment */
	m = (mblk_t *)xdrs->x_base;
	if (m == NULL)
		return (FALSE);
	if ((xdrs->x_handy -= sizeof (long)) < 0) {
		m = m->b_cont;
		xdrs->x_base = (caddr_t)m;
		if (m == NULL) {
			xdrs->x_handy = 0;
			return (FALSE);
		}
		xdrs->x_handy = m->b_wptr - m->b_rptr - sizeof (long);
	}
	/* LINTED pointer alignment */
	*lp = ntohl(*((long *)(m->b_rptr)));
	m->b_rptr += sizeof (long);
	return (TRUE);
}

bool_t
xdrmblk_putlong(xdrs, lp)
	register XDR *xdrs;
	long *lp;
{
	register mblk_t *m;

	/* LINTED pointer alignment */
	m = (mblk_t *)xdrs->x_base;
	if (m == NULL)
		return (FALSE);
	if ((xdrs->x_handy -= sizeof (long)) < 0) {
		if (m->b_cont == NULL)
			m->b_cont = xdrmblk_alloc((int)xdrs->x_private);
		m = m->b_cont;
		xdrs->x_base = (caddr_t)m;
		if (m == NULL) {
			xdrs->x_handy = 0;
			return (FALSE);
		}
		xdrs->x_handy = m->b_datap->db_lim - m->b_datap->db_base -
				sizeof (long);
	}
	/* LINTED pointer alignment */
	*(long *)m->b_wptr = htonl(*lp);
	m->b_wptr += sizeof (long);
	return (TRUE);
}

bool_t
xdrmblk_getbytes(xdrs, addr, len)
	register XDR *xdrs;
	caddr_t addr;
	register u_int len;
{
	register mblk_t *m;

	/* LINTED pointer alignment */
	m = (mblk_t *)xdrs->x_base;
	if (m == NULL)
		return (FALSE);
	while ((xdrs->x_handy -= (int)len) < 0) {
		if ((xdrs->x_handy += (int)len) > 0) {
			bcopy((caddr_t)m->b_rptr, addr, (u_int)xdrs->x_handy);
			m->b_rptr += xdrs->x_handy;
			addr += xdrs->x_handy;
			len -= xdrs->x_handy;
		}
		m = m->b_cont;
		xdrs->x_base = (caddr_t)m;
		if (m == NULL) {
			xdrs->x_handy = 0;
			return (FALSE);
		}
		xdrs->x_handy = m->b_wptr - m->b_rptr;
	}
	bcopy((caddr_t)m->b_rptr, addr, (u_int)len);
	m->b_rptr += len;
	return (TRUE);
}

/*
 * Sort of like getbytes except that instead of getting
 * bytes we return the mblk chain which contains all the rest
 * of the bytes.
 */
bool_t
xdrmblk_getmblk(xdrs, mm, lenp)
	register XDR *xdrs;
	register mblk_t **mm;
	register u_int *lenp;
{
	register mblk_t *m;
	register int len;
	long llen;

	if (! xdrmblk_getlong(xdrs, &llen)) {
		return (FALSE);
	}
	*lenp = llen;
	/* LINTED pointer alignment */
	m = (mblk_t *)xdrs->x_base;
	*mm = m;
	/*
	 * Consistency check.
	 */
	len = 0;
	while (m) {
		len += (m->b_wptr - m->b_rptr);
		m = m->b_cont;
	}
	if (len < llen) {
		printf("xdrmblk_getmblk failed\n");
		return (FALSE);
	}
	xdrs->x_base = NULL;
	xdrs->x_handy = 0;
	return (TRUE);
}

bool_t
xdrmblk_putbytes(xdrs, addr, len)
	register XDR *xdrs;
	register caddr_t addr;
	register u_int len;
{
	register mblk_t *m;
	register u_int i;

	/* LINTED pointer alignment */
	m = (mblk_t *)xdrs->x_base;
	if (m == NULL)
		return (FALSE);
	/*
	 * Performance tweak: converted explicit bcopy()
	 * call to simple in-line. This function is called
	 * to process things like readdir reply filenames
	 * which are small strings--typically 12 bytes or less.
	 * Overhead of calling bcopy() is obnoxious for such
	 * small copies.
	 */
	while ((xdrs->x_handy -= (int)len) < 0) {
		if ((xdrs->x_handy += (int)len) > 0) {
			for (i = 0; i < (u_int)xdrs->x_handy; i++)
				*m->b_wptr++ = *addr++;
			len -= xdrs->x_handy;
		}
		if (m->b_cont == NULL)
			m->b_cont = xdrmblk_alloc((int)xdrs->x_private);
		m = m->b_cont;
		xdrs->x_base = (caddr_t)m;
		if (m == NULL) {
			xdrs->x_handy = 0;
			return (FALSE);
		}
		xdrs->x_handy = m->b_datap->db_lim - m->b_datap->db_base;
	}
	for (i = 0; i < len; i++)
		*m->b_wptr++ = *addr++;
	return (TRUE);
}

struct xdrmblkfree {
	frtn_t myfree;
	frtn_t rfree;
};

void
xdrmblk_putbuffree(frp)
	struct xdrmblkfree *frp;
{

	(*frp->rfree.free_func)(frp->rfree.free_arg);
	kmem_free((caddr_t)frp, sizeof (*frp));
}

/*
 * Like putbuf, only all we have to do is add this mblk to the list.
 * The caller is responsible for allocating the mblk.
 */
bool_t
xdrmblk_putmblk(xdrs, m, len)
	register XDR *xdrs;
	register mblk_t *m;
	u_int len;
{
	long llen = (long)len;

	if (len & 3) {	/* can't handle roundup problems */
		return (FALSE);
	}

	if (! xdrmblk_putlong(xdrs, &llen)) {
		return (FALSE);
	}
	/* LINTED pointer alignment */
	((mblk_t *)xdrs->x_base)->b_cont = m;
	xdrs->x_base = (caddr_t)m;
	xdrs->x_handy = 0;
	return (TRUE);
}

u_int
xdrmblk_getpos(xdrs)
	register XDR *xdrs;
{
	register u_int tmp;
	register mblk_t *m;

	/* LINTED pointer alignment */
	m = (mblk_t *)xdrs->x_base;

	if (xdrs->x_op == XDR_DECODE) {
		tmp = (u_int)(m->b_rptr - m->b_datap->db_base);
	} else {
		tmp = (u_int)(m->b_wptr - m->b_datap->db_base);
	}
	return (tmp);

}

bool_t
xdrmblk_setpos(xdrs, pos)
	register XDR *xdrs;
	u_int pos;
{
	register mblk_t *m;
	register unsigned char *newaddr;

	/* LINTED pointer alignment */
	m = (mblk_t *)xdrs->x_base;
	if (m == NULL)
		return (FALSE);

	/* calculate the new address from the base */
	newaddr = m->b_datap->db_base + pos;

	if (xdrs->x_op == XDR_DECODE) {
		if (newaddr > m->b_wptr)
			return (FALSE);
		m->b_rptr = newaddr;
		xdrs->x_handy = m->b_wptr - newaddr;
	} else {
		if (newaddr > m->b_datap->db_lim)
			return (FALSE);
		m->b_wptr = newaddr;
		xdrs->x_handy = m->b_datap->db_lim - newaddr;
	}
	return (TRUE);
}

int xdrmblk_inline_hits = 0;
int xdrmblk_inline_misses = 0;

long *
xdrmblk_inline(xdrs, len)
	register XDR *xdrs;
	int len;
{
	long *buf = 0;
	register mblk_t *m;

	if (xdrs->x_op == XDR_FREE || xdrs->x_base == NULL)
		return (buf);

	if (xdrs->x_handy >= len) {
		xdrs->x_handy -= len;
		/* LINTED pointer alignment */
		m = (mblk_t *)xdrs->x_base;
		if (xdrs->x_op == XDR_DECODE) {
			/* LINTED pointer alignment */
			buf = (long *)m->b_rptr;
			m->b_rptr += len;
		} else {
			/* LINTED pointer alignment */
			buf = (long *)m->b_wptr;
			m->b_wptr += len;
		}
		xdrmblk_inline_hits++;
	} else {
		xdrmblk_inline_misses++;
	}

	return (buf);
}

bool_t
xdrmblk_control(xdrs, request, info)
	register XDR *xdrs;
	void *info;
{

	switch (request) {

	/*
	 * Return the next 4 byte long in the XDR stream.
	 */
	case XDR_PEEK:
		{
			register mblk_t *m;
			long *lp;

			if (xdrs->x_handy < sizeof (long)) {
				return (FALSE);
			}

			/* LINTED pointer alignment */
			m = (mblk_t *)xdrs->x_base;
			lp = (long *)info;
			/* LINTED pointer alignment */
			*lp = ntohl(*((long *)(m->b_rptr)));
			return (TRUE);
		}
	default:
			return (FALSE);
	}
}

static mblk_t *
xdrmblk_alloc(sz)
	int sz;
{
	mblk_t *mp;

	if (sz == 0)
		return (NULL);

	while ((mp = allocb(sz, BPRI_LO)) == NULL) {
		if (strwaitbuf(sz, BPRI_LO))
			break;
	}
	return (mp);
}
#endif	/* _KERNEL */
